grammar silver:analysis:agnosis:core;
import silver:base:core;

synthesized attribute str :: String;
synthesized attribute type :: String;
synthesized attribute pName :: String;
synthesized attribute pInNames :: [AgData];
synthesized attribute pOutName :: String;
synthesized attribute ntName :: String;
synthesized attribute tName :: String;
synthesized attribute attrName :: String;
synthesized attribute attrType :: String;
synthesized attribute attrDirection :: String;
synthesized attribute attribution :: AgData;
nonterminal AgData with {str, type, pName, pInNames, pOutName, ntName, tName, attrName, attrType, attrDirection, typerep, attribution};

abstract production ntDclData
top::AgData ::= ntN::String
{
  top.type = "NT";
  top.ntName = ntN;
}

abstract production startDclData
top::AgData ::= ntN::String
{
  top.type = "Start";
  top.ntName = ntN;
}

abstract production tDclData
top::AgData ::= tN::String
{
  top.type = "T";
  top.tName = tN;
}

abstract production pDclData
top::AgData ::= pN::String  l::[AgData] out::String
{
  top.type = "Prod";
  top.pName = pN;
  top.pInNames = l;
  top.pOutName = out;
}

abstract production attrDclData
top::AgData ::= ntN::String atN::String atT::String atD::String
{
  top.type = "AttrDecl";
  top.ntName  = ntN;
  top.attrName = atN;
  top.attrType = atT;
  top.attrDirection = atD;
}

abstract production localAttrDclData
top::AgData ::= pN::String atN::String atT::String atD::String
{
  top.type = "LocalAttrDecl";
  top.pName  = pN;
  top.attrName = atN;
  top.attrType = atT;
  top.attrDirection = atD;
}

abstract production attributionData
top::AgData ::= pN::String ag::AgData
{
  top.type = "AttributionRule";
  top.pName  = pN;
  top.attribution = ag'';
}


function makeName
String ::= st::String
{
  local attribute i :: Integer;
  i = indexof( ":", st);

  return if i == -1
         then st
         else substring(0, i, st) ++ "_" ++ makeName(substring(i+1, length(st), st));
}


synthesized attribute lhsRule :: AgData;
synthesized attribute rhsRule :: AgData;

synthesized attribute lhsIndex :: Integer;
synthesized attribute rhsIndex :: Integer;

attribute lhsRule occurs on AgData;
attribute rhsRule occurs on AgData;
attribute lhsIndex occurs on AgData;
attribute rhsIndex occurs on AgData;

abstract production attrDefDclData1
top::AgData ::= l::AgData r::AgData
{
  top.type = "AttrDep";
  top.lhsRule = new(l);
  top.rhsRule = new(r);
}

abstract production attrDefDclData2
top::AgData ::= l::AgData i::Integer
{
    top.type = "AttrDepHigher";
    top.lhsRule = new(l);
    top.rhsIndex = i;
}

abstract production attrDefDclData3
top::AgData ::= l::AgData i::Integer
{
    top.type = "AttrDepRef";
    top.lhsRule = new(l);
    top.rhsIndex = i;
}
abstract production attrDefDclData4
top::AgData ::= l::AgData r::AgData
{
  top.type = "AttrDepRefTree";
  top.lhsRule = new(l);
  top.rhsRule = new(r);
}
abstract production attrDefDclData5
top::AgData ::= l::AgData r::AgData
{
  top.type = "AttrDepTreeRef";
  top.lhsRule = new(l);
  top.rhsRule = new(r);
}

abstract production attrDefDclData6
top::AgData ::= l::AgData
{
  top.type = "AttrDepConst";
  top.lhsRule = new(l);
}

synthesized attribute preDotIndex :: Integer;
synthesized attribute postDotName :: String;
synthesized attribute preDotName :: String;
synthesized attribute localName :: String;

attribute preDotIndex occurs on AgData;
attribute postDotName occurs on AgData;
attribute preDotName occurs on AgData;
attribute localName occurs on AgData;


abstract production attrAccessDclData1
top::AgData ::= i::Integer n::String t::Decorated TypeRep
{
  top.type = "AttrDepElt";
  top.typerep = t;
  top.preDotIndex = i;
  top.postDotName = n;
}
abstract production attrAccessDclData2
top::AgData ::= n::String t::Decorated TypeRep
{
  top.type = "AttrDepEltLocal";
  top.typerep = t;
  top.localName = n;
}
abstract production attrAccessDclData3
top::AgData ::= n1::String n2::String t::Decorated TypeRep
{
  top.type = "AttrDepEltRef";
  top.typerep = t;

  top.preDotName = n1;
  top.postDotName = n2;
}
abstract production attrAccessDclData4
top::AgData ::= n1::String n2::String t::Decorated TypeRep
{
  top.type = "AttrDepEltTree";
  top.typerep = t;

  top.preDotName = n1;
  top.postDotName = n2;
}

abstract production attrAccessDclData5
top::AgData ::=
{
  top.type = "CONSTANT";
}

synthesized attribute index ::Integer;
attribute index occurs on AgData;

abstract production attrAccessDclData6
top::AgData ::= i::Integer t::Decorated TypeRep
{
  top.type = "CHILD";
  top.index = i;
  top.typerep = t;
}

