grammar silver:analysis:agnosis:core;
import silver:base:core;

attribute agnosis occurs on ProductionBody;
attribute agnosis occurs on ProductionStmts;
attribute agnosis occurs on ProductionStmt;
attribute agnosis occurs on AttributeDef;
attribute agnosis occurs on LHSExpr;


aspect production emptyProductionBody
top::ProductionBody ::= 
{
  top.agnosis = [];
}

aspect production defaultProductionBody
top::ProductionBody ::= stmts::ProductionStmts
{
  top.agnosis = stmts.agnosis;
}

aspect production productionStmtsNone
top::ProductionStmts ::= 
{
  top.agnosis = [];
}

aspect production productionStmtsSingle
top::ProductionStmts ::= stmt::ProductionStmt
{
  top.agnosis = stmt.agnosis;
}

aspect production productionStmtsCons
top::ProductionStmts ::= h::ProductionStmt t::ProductionStmts
{
  top.agnosis = h.agnosis +++ t.agnosis;
}

aspect production productionStmtsAppend
top::ProductionStmts ::= h::ProductionStmts t::ProductionStmts
{
  top.agnosis = h.agnosis +++ t.agnosis;
}

aspect production forwardsDcl
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr s::Semi_t
{
  top.agnosis = [];
}

aspect production forwardsDclWith
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  top.agnosis = [];
}

aspect production forwardingDclWith
top::ProductionStmt ::= f::Forwarding_kwd w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  top.agnosis = [];
}

aspect production stmtToAttrDef
top::ProductionStmt ::= a::AttributeDef
{
  top.agnosis = a.agnosis;
}

aspect production attributeDclLocal
top::ProductionStmt ::= l::Local_kwd  at::Attribute_kwd a::Id_t 
                        ht::HasType_t te::Type  s::Semi_t
{
  local attribute pN :: String;
  pN = top.thisName;

  local attribute atN :: String;
  atN = makeName(a.lexeme);

  local attribute atT :: String;
  atT = if te.typerep.isNonTerminal 
	then "AttrHigher (NT \"" ++ makeName(te.typerep.typeName) ++ "\")"
	else if te.typerep.isDecorated
	then "AttrRef (NT \"" ++ makeName(te.typerep.typeName) ++ "\")"
	else "AttrScaler";

  local attribute atD :: String;
  atD = "Synthesized";

  top.agnosis = [localAttrDclData(pN, atN, atT, atD)];
}

aspect production attrDef
top::AttributeDef ::= lhs::LHSExpr eq::Equal_t e::Expr s::Semi_t
{
  local attribute allAg ::[AgData];
  allAg = e.agnosis +++ e.lastAccess;

  top.agnosis = if e.isConstant 
		then [attributionData(top.thisName, attrDefDclData6(head(lhs.agnosis)))]
		else makeAgData(top.thisName, head(lhs.agnosis), allAg);
}

function makeAgData
[AgData] ::= pN::String lhs::AgData ags::[AgData]
{
  local attribute ag ::AgData;
  ag = head(ags);

  local attribute result ::AgData;
  result = if ag.type == "LocalAttrDecl" || ag.type == "AttributionRule"
	   then ag''
	   else if ag.type == "CONSTANT"
	   then attributionData(pN, attrDefDclData6(lhs''))
	   else if ag.type == "CHILD" && ag.typerep.isNonTerminal
	   then attributionData(pN, attrDefDclData2(lhs'', ag.index))
	   else if ag.type == "CHILD" && ag.typerep.isDecorated
	   then attributionData(pN, attrDefDclData3(lhs'', ag.index))
	   else if ag.type == "CHILD"
	   then attributionData(pN, attrDefDclData6(lhs''))
	   else if lhs.typerep.isNonTerminal && ag.typerep.isDecorated
	   then attributionData(pN, attrDefDclData5(lhs'', ag''))
	   else if lhs.typerep.isDecorated && ag.typerep.isNonTerminal
 	   then attributionData(pN, attrDefDclData4(lhs'', ag''))
	   else attributionData(pN, attrDefDclData1(lhs'', ag''));

  return if null(ags) then [] else cons(result'', makeAgData(pN, lhs'', tail(ags)));
}

aspect production lhsExprOne
top::LHSExpr ::= id::Id_t
{
  top.agnosis = [attrAccessDclData2(fName, top.typerep)];
}

aspect production lhsExprTwo
top::LHSExpr ::= id::Id_t d::Dot_t q::QName
{
  top.agnosis = if in_sig 
		then [attrAccessDclData1(i, makeName(fName2), top.typerep)]
		else if head(vals2).typerep.isDecorated
		     then [attrAccessDclData3(makeName(fName1), makeName(fName2), top.typerep)]
		     else [attrAccessDclData4(makeName(fName1), makeName(fName2), top.typerep)];

  local attribute i :: Integer;
  i = getIndex(fName1, top.fullSignatureNames);

  local attribute in_sig ::Boolean;
  in_sig = !null(getFullNameDcl(id.lexeme, top.signatureEnv));
}
function getIndex
Integer ::= n::String l::[String]
{
  return getIndexHelp(0, n, l);
}

function getIndexHelp
Integer ::= i::Integer n::String l::[String]
{
  return if null(l) 
	 then -1 
	 else if n == head(l)
	 then i
         else getIndexHelp(i+1, n, tail(l));
	
}