grammar silver:analysis:termination:core;
import silver:base:core;

synthesized attribute terminationRuleRHS :: String occurs on Expr, Exprs;

--Base Expressions
aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.terminationRuleRHS = e.terminationRuleRHS;
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  local attribute vb :: String;
  vb = lookupTerminationRuleVarBinding (q.pp, top.terminationRuleVarBindingsEnv);

  top.terminationRuleRHS = if in_locals then "LocalVariable" else vb;--if vb == "UnknownVariable" then q.pp else vb;
}

aspect production decorateExpr
top::Expr ::= q::QName
{
  local attribute vb :: String;
  vb = lookupTerminationRuleVarBinding (q.pp, top.terminationRuleVarBindingsEnv);

  top.terminationRuleRHS = if in_locals then "LocalVariable" else vb;--if vb == "UnknownVariable" then q.pp else vb;
}

aspect production productionReference
top::Expr ::= q::QName
{
  local attribute vb :: String;
  vb = lookupTerminationRuleVarBinding (q.pp, top.terminationRuleVarBindingsEnv);

  top.terminationRuleRHS = if vb == "UnknownVariable" then q.pp else vb;
}

--no one can take the dot off of a production call there would need to be a decorate ...with {...} inserted.
aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  local attribute vb :: String;
  vb = lookupTerminationRuleVarBinding (e.pp, top.terminationRuleVarBindingsEnv);

  -- The generated rule for  e(e1, e2) is e'(e1', e2') if e is an explicit production name
  -- and dead_production_2 (e', e1, e2) if e is a production attribute or a production-valued child
  -- The number of arguments is appended to dead_production since each function has a constant arity
  top.terminationRuleRHS = case e of 
				-- Production valued attribute
				attributeAccess (_, _, _) => "dead_production" ++ (if es.terminationRuleRHS == "" 
											then "_0 (" ++ e.terminationRuleRHS ++ ")"
											else "_" ++ toString (es.exprCount) ++ " (" ++ e.terminationRuleRHS ++ ", " ++ es.terminationRuleRHS ++ ")") |

				_ => if vb == "UnknownVariable" 
					-- e is not a child (i.e., a production valued child)
					then (if e.pp == "parse" 
						-- parse is treated differently
						then "ParseFunction" 
						-- explicit production name
						else e.terminationRuleRHS) ++ 

					     (if es.terminationRuleRHS == "" then "" else " (" ++ es.terminationRuleRHS ++ ")")

					-- e is a production valued child
					else ("dead_production" ++ (if es.terminationRuleRHS == "" 
									then "_0 (" ++ e.terminationRuleRHS ++ ")"
									else "_" ++ toString (es.exprCount) ++ " (" ++ e.terminationRuleRHS ++ ", " ++ es.terminationRuleRHS ++ ")"))
			   end;
}

aspect production atAccess
top::Expr ::= e::Expr _ q::QName
{
  top.terminationRuleRHS = if isInherited (fName, top.env)
				then "InheritedAttribute"
				else e.terminationRuleRHS;
}

aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.terminationRuleRHS = "dead_primitive";
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.terminationRuleRHS = "dead_primitive";
}

aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  top.terminationRuleRHS =  e.terminationRuleRHS;
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}


aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}


aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}


aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}


aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}


aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.terminationRuleRHS = "dead_ternary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ", " ++ e3.terminationRuleRHS ++ ")";
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.terminationRuleRHS = "dead_primitive";
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.terminationRuleRHS = "dead_primitive";
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  top.terminationRuleRHS = e.terminationRuleRHS;
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.terminationRuleRHS = "dead_primitive";
}

aspect production plusPlus
top::Expr ::= e1::Expr p::PlusPlus_t e2::Expr
{
  top.terminationRuleRHS = "dead_binary (" ++ e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS ++ ")";
}

aspect production exprsEmpty
top::Exprs ::=
{
  top.terminationRuleRHS = "";
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.terminationRuleRHS = e.terminationRuleRHS;
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.terminationRuleRHS = e1.terminationRuleRHS ++ ", " ++ e2.terminationRuleRHS;
}

aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  top.terminationRuleRHS = e.terminationRuleRHS;
}

aspect production exprStmtsEmpty
top::ExprInh ::= 
{
}

aspect production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 
}

aspect production exprLhsExpr
top::ExprLHSExpr ::= q::QName
{
}

-- count the number of expressions, to generate the right function symbol in the generated rule
synthesized attribute exprCount :: Integer occurs on Exprs;
aspect production exprsEmpty
top::Exprs ::=
{
  top.exprCount = 0;
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.exprCount = 1;
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.exprCount = 1 + e2.exprCount;
}
