grammar silver:analysis:termination:core;
import silver:base:core;

-- the lhs of the rule, say p(x,y)
autocopy attribute terminationRuleLHS :: String;
attribute terminationRuleLHS occurs on ProductionSignature, ProductionLHS, ProductionLHSElem, ProductionBody, ProductionStmts, ProductionStmt, AttributeDef;

-- the variables in the lhs say "x,y"
synthesized attribute LHSVariableString :: String occurs on ProductionSignature, ProductionRHS, ProductionRHSElem;

-- the number of variables in the lhs, say p(x,y) means 2
synthesized attribute LHSVariableLastIndex :: Integer occurs on ProductionSignature, ProductionRHS;

-- list of bindings of variables that might appear in the forwards to rule
-- we bind child names to new variables (x_1, x_2, ...) and the lhs variable to the lhs string
-- so that top forwards to f(top) is translated to p(x) -> f(p(x))
synthesized attribute terminationRuleVarBindings :: [ StringPair ] occurs on ProductionSignature, ProductionLHS, ProductionLHSElem, ProductionRHS, ProductionRHSElem;

autocopy attribute terminationRuleVarBindingsEnv :: [ StringPair ];
attribute terminationRuleVarBindingsEnv occurs on ProductionBody, ProductionStmts, ProductionStmt, Expr, Exprs, ExprInh, AttributeDef;

inherited attribute LHSVariableIndex :: Integer occurs on ProductionRHS, ProductionRHSElem;

aspect production productionDcl
 top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{
  top.terminationRules = body.terminationRules;
  top.LHSVariableLastIndex = ns.LHSVariableLastIndex;

  local attribute ruleLHS :: String;
  ruleLHS = id.lexeme ++ (if ns.LHSVariableString == "" then "" else "(" ++ ns.LHSVariableString ++ ")");

  body.terminationRuleLHS = ruleLHS;
  body.terminationRuleVarBindingsEnv = ns.terminationRuleVarBindings;

  ns.terminationRuleLHS = ruleLHS;
}

aspect production productionSignatureEmptyRHS
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t
{
  top.LHSVariableString = "";
  top.LHSVariableLastIndex = 0;
  top.terminationRuleVarBindings = lhs.terminationRuleVarBindings;
}

aspect production productionSignature
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t rhs::ProductionRHS 
{
  top.LHSVariableString = rhs.LHSVariableString;
  top.LHSVariableLastIndex = rhs.LHSVariableLastIndex;
  top.terminationRuleVarBindings = lhs.terminationRuleVarBindings ++ rhs.terminationRuleVarBindings;

  rhs.LHSVariableIndex = 1;
}

aspect production productionLHS
top::ProductionLHS ::= lhs::ProductionLHSElem
{
  top.terminationRuleVarBindings = lhs.terminationRuleVarBindings;
}

aspect production productionLHSElem
top::ProductionLHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.terminationRuleVarBindings = [ string_pair (id.lexeme, top.terminationRuleLHS) ];
}

aspect production productionRHSEmpty
top::ProductionRHS ::=
{
  top.LHSVariableString = "";
  top.LHSVariableLastIndex = top.LHSVariableIndex - 1;
  top.terminationRuleVarBindings = [];
}

aspect production productionRHSSingle
top::ProductionRHS ::= rhs::ProductionRHSElem
{
  top.LHSVariableString = rhs.LHSVariableString;
  top.LHSVariableLastIndex = top.LHSVariableIndex;
  top.terminationRuleVarBindings = rhs.terminationRuleVarBindings;

  rhs.LHSVariableIndex = top.LHSVariableIndex;
}

aspect production productionRHSCons
top::ProductionRHS ::= h::ProductionRHSElem t::ProductionRHS
{
  top.LHSVariableString = h.LHSVariableString ++ ", " ++ t.LHSVariableString;
  top.LHSVariableLastIndex = t.LHSVariableLastIndex;
  top.terminationRuleVarBindings = h.terminationRuleVarBindings ++ t.terminationRuleVarBindings;

  h.LHSVariableIndex = top.LHSVariableIndex;
  t.LHSVariableIndex = top.LHSVariableIndex + 1;
}

aspect production productionRHSElem
top::ProductionRHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  local attribute varName :: String;
  varName = "x_" ++ toString (top.LHSVariableIndex);

  top.LHSVariableString = varName;
  top.terminationRuleVarBindings = [ string_pair (id.lexeme, varName) ];
}
