grammar silver:analysis:typechecking:core;
import silver:base;
import silver:utils;

import core only bValue;

attribute typeErrors occurs on ProductionBody, ProductionStmts, ProductionStmt, AttributeDef, LHSExpr, ForwardInh, ForwardLHSExpr;

aspect production emptyProductionBody
top::ProductionBody ::= 
{
  top.typeErrors = "";
}

aspect production defaultProductionBody
top::ProductionBody ::= stmts::ProductionStmts
{
  top.typeErrors = stmts.typeErrors;
}

aspect production productionStmtsNone
top::ProductionStmts ::= 
{
  top.typeErrors = "";
}

aspect production productionStmtsSingle
top::ProductionStmts ::= stmt::ProductionStmt
{
  top.typeErrors = stmt.typeErrors;
}

aspect production productionStmtsAppend
top::ProductionStmts ::= h::ProductionStmts t::ProductionStmts
{
  top.typeErrors = h.typeErrors ++ t.typeErrors;
}
aspect production productionStmtsCons
top::ProductionStmts ::= h::ProductionStmt t::ProductionStmts
{
  top.typeErrors = h.typeErrors ++ t.typeErrors;
}
aspect production forwardsDcl
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr s::Semi_t
{
  local attribute er1 :: String;
  er1 = if e.typerep.isNonTerminal then "" else top.location ++ "Error: You must forward to a NonTerminal.\n";

  local attribute lhsT :: Decorated TypeRep;
  lhsT = head(top.signatureTypes);

  local attribute er2 :: String;
  er2 = if lhsT.typeEquals(lhsT, e.typerep).bValue
	then ""
	else top.location ++ "Error: You must forward to a NonTerminal of type '" ++ lhsT.typeName ++ "'.\n";


  top.typeErrors = e.typeErrors ++ er1 ++ er2;
}

aspect production forwardsDclWith
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  local attribute er1 :: String;
  er1 = if e.typerep.isNonTerminal then "" else top.location ++ "Error: You must forward to a NonTerminal.\n";

  local attribute lhsT :: Decorated TypeRep;
  lhsT = head(top.signatureTypes);

  local attribute er2 :: String;
  er2 = if lhsT.typeEquals(lhsT, e.typerep).bValue
	then ""
	else top.location ++ "Error: You must forward to a NonTerminal of type '" ++ lhsT.typeName ++ "'.\n";


  top.typeErrors = e.typeErrors ++ inh.typeErrors ++ er1 ++ er2;
}

aspect production forwardingDclWith
top::ProductionStmt ::= f::Forwarding_kwd w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  top.typeErrors = inh.typeErrors;
}

aspect production forwardStmtsEmpty
top::ForwardInh ::= 
{
  top.typeErrors = "";
}

aspect production forwardStmtsCons
top::ForwardInh ::= lhs::ForwardLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ForwardInh
{
  local attribute valid :: Boolean;
  valid = lhs.typerep.typeEquals(lhs.typerep, e.typerep).bValue;

  local attribute er :: String;
  er = if valid then "" else top.location ++ "Error: The LHS and RHS of the assignment must be the same type.\n" ++
           "\tInstead they are '" ++ lhs.typerep.pp ++ "' and '" ++ e.typerep.pp ++ "'\n" ; 


  top.typeErrors = er ++ e.typeErrors ++ lhs.typeErrors; 
}


aspect production forwardLhsExpr
top::ForwardLHSExpr ::= q::QName
{
  top.typeErrors = "";
}

aspect production attributeDclLocal
top::ProductionStmt ::= l::Local_kwd  at::Attribute_kwd a::Id_t 
                        ht::HasType_t te::Type  s::Semi_t
{
  top.typeErrors = "";
}
aspect production stmtToAttrDef
top::ProductionStmt ::= a::AttributeDef
{
  top.typeErrors = a.typeErrors;
}

aspect production attrDef
top::AttributeDef ::= lhs::LHSExpr e::Expr
{
  local attribute valid :: Boolean;
  valid = lhs.typerep.typeEquals(lhs.typerep, e.typerep).bValue;

  local attribute er :: String;
  er = if valid then "" 
       else top.location ++ "Error: The LHS and RHS of the assignment must be the same type.\n" ++
                                  "\tInstead they are '" ++ lhs.typerep.pp ++ "' and '" ++ e.typerep.pp ++ "'\n" ; 

  top.typeErrors = er ++ lhs.typeErrors ++ e.typeErrors;
}

aspect production lhsExprOne
top::LHSExpr ::= id::Id_t
{
  top.typeErrors = "";
}

aspect production lhsExprTwo
top::LHSExpr ::= id::Id_t d::Dot_t q::QName
{
  local attribute occursOn :: Boolean;
  occursOn = if  !null(fNames1) && !null(vals1) && !null(fNames2)
             then doesOccurOn(fName2, head(vals1).typerep.typeName, top.env)
             else false;	
  
  local attribute e1 :: String;
  e1 = if occursOn then ""
       else top.location ++ "Error: Attribute '" ++ q.name ++ "' does not decorate value '" ++ id.lexeme ++ "'.\n";

  local attribute e2 :: String;
  e2 = if !null(fNames2) && id.lexeme == head(top.signatureNames) && isInherited(fName2, top.env)
       then top.location ++ "Error: Cannot assign to the lhs's inherited attributes.\n"
       else "";

  local attribute e3 :: String;
  e3 = if !null(fNames2) && containsString(id.lexeme, tail(top.signatureNames)) && isSynthesized(fName2, top.env)
       then top.location ++ "Error: Cannot assign to a child's synthesized attributes.\n"
       else "";

  local attribute e4 :: String;
  e4 = if !null(fNames1) && !null(fNames2) && !null(getValueDclOne(fName1, top.localsEnv)) && isSynthesized(fName2, top.env)
       then top.location ++ "Error: Cannot assign to a locals synthesized attributes.\n"
       else "";

  top.typeErrors = e1 ++ e2 ++ e3 ++ e4;
}

