grammar silver:base:core;

terminal Attribute_kwd   /attribute/   lexer precedence = 5 ;
terminal Synthesized_kwd /synthesized/ lexer precedence = 5 ;
terminal Inherited_kwd   /inherited/   lexer precedence = 5 ;

concrete production attributeDclInh
top::AGDcl ::= i::Inherited_kwd at::Attribute_kwd a::Id_t ht::HasType_t te::Type s::Semi_t
{
  top.pp = "inherited attribute " ++ a.lexeme ++ " :: " ++ te.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  production attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ a.lexeme;

  top.defs = addAttributeDcl(fName, te.typerep, 
             addInheritedDcl(fName, 
	     addFullNameDcl(a.lexeme, fName,  emptyDefs())));


  top.globalDefs = addAttributeDcl(fName, te.typerep, 
                   addInheritedDcl(fName, emptyDefs()));

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(a.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ a.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getAttributeDcl(fName, top.env)) > 1
       then top.location ++ "Error: Attribute '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ te.errors;
}


concrete production attributeDclSyn
top::AGDcl ::= sn::Synthesized_kwd  at::Attribute_kwd a::Id_t ht::HasType_t te::Type  s::Semi_t
{
  top.location = "[line: " ++ toString(sn.line) ++ ", column: " ++ toString(sn.column) ++ "] ";
  top.pp = "synthesized attribute " ++ a.lexeme ++ " :: " ++ te.pp ++ " ;" ;

  production attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ a.lexeme;

  top.defs = addAttributeDcl(fName, te.typerep, 
             addSynthesizedDcl(fName, 
	     addFullNameDcl(a.lexeme, fName,  emptyDefs())));

  top.globalDefs = addAttributeDcl(fName, te.typerep, 
             	   addSynthesizedDcl(fName, emptyDefs()));

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(a.lexeme, top.env)) > 1
       then top.location ++ "Error: Name '" ++ a.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getAttributeDcl(fName, top.env)) > 1
       then top.location ++ "Error: Attribute '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ te.errors;
}




