grammar silver:base:core;
import core;
import silver:utils;
import silver:utils:command with parse as commandParse, grammarName as gName;

synthesized attribute code :: Integer;
synthesized attribute order :: Integer;
nonterminal Unit with {ioIn, io, code, order};

synthesized attribute rootSpec :: Decorated RootSpec;
inherited attribute parser :: Production(Root ::= String);

nonterminal RunUnit with {io};

abstract production run
top::RunUnit ::= iIn::IO args::String p::Production(Root ::= String)
{
  production attribute a :: Command;
  a = commandParse(args);

  production attribute preOps :: [Unit] with unitPP;
  preOps := if a.displayVersion then [printVersion()] else [];

  local attribute preIO :: IOInteger;
  preIO = runAll(iIn, unitMergeSort(preOps));

  local attribute envArg :: IOString;
  envArg = envVar("GRAMMAR_PATH", preIO.io);

  local attribute sPath :: String;
  sPath = a.searchPath ++ ":" ++ envArg.sValue;

  local attribute spath :: [String];
  spath = getSearchPath(sPath);

  --the grammar path ':' replaced by '/'
  local attribute gpath :: String;
  gpath = getGrammarPath(a.gName) ++ "/";

  local attribute loc :: MaybeIOStr;
  loc = findGrammarLocation(envArg.io, gpath, spath);

  local attribute i :: IO;
  i = system("mkdir -p " ++ a.generatedPath, loc.io).io;
 
  production attribute extraGrammars :: [String] with spp;
  extraGrammars := [];

  local attribute extraUnit :: CompilationUnit;
  extraUnit = compileAllExtra(extraGrammars, [::String], [::String]);
  extraUnit.parser = p;
  extraUnit.globalEnv = gEnv;
  extraUnit.compiledGrammars = grammars;


  production attribute unit :: CompilationUnit;
  unit = compileAll(i,
		    spath,
		    cons(a.gName, extraUnit.needGrammars), 
		    extraUnit.seenGrammars);

  unit.parser = p;
  unit.globalEnv = gEnv;
  unit.compiledGrammars = grammars;

  production attribute gEnv :: Decorated Env;
  gEnv = toEnv(appendDefs(makeDefaultGlobalDefs(), appendDefs(unit.globalDefs, extraUnit.globalDefs)));

  production attribute postOps :: [Unit] with unitPP;
  postOps := if a.noBindingChecking then [] else [printAllErrors(grammars)]; 

  production attribute grammars :: [GrammarSpec];
  grammars =  extraUnit.compiledList +++ unit.compiledList;

  top.io = if preIO.iValue != 0
           then preIO.io
           else
           if a.okay && loc.found
	   then runAll(unit.io, unitMergeSort(postOps)).io
	   else if a.okay && !loc.found
	        then print("\nGrammar '" ++ a.gName ++ "' could not be located, make sure that the grammar name is correct and it's location is on $GRAMMAR_PATH.\n\n", loc.io)
		else print(a.usage, iIn);
}
abstract production printAllErrors
top::Unit ::= specs::[GrammarSpec]
{
  forwards to printAllErrorsHelp(specs)
	with {
		ioIn = print("Checking Binding Errors.\n", top.ioIn);
	};
}

abstract production printVersion
top::Unit ::= 
{
  top.order = 0;
  top.io = print("Silver Version 0.2\n", top.ioIn);
  top.code = -1;
}

abstract production printAllErrorsHelp
top::Unit ::= specs::[GrammarSpec]
{
  local attribute es :: String;
  es = head(specs).rootSpec.errors;

  local attribute i :: IO;
  i = if es == "" then top.ioIn else print("Binding Errors for : " ++ head(specs).name ++ "\n" ++ es ++ "\n\n", top.ioIn);

  local attribute recurse :: Unit;
  recurse = printAllErrorsHelp(tail(specs));
  recurse.ioIn = i;

  top.io = if null(specs) then top.ioIn else recurse.io;

  top.code = if null(specs) || (es == ""  && recurse.code == 0)
	     then 0 
	     else -1;

  top.order = 0;
}

function runAll
IOInteger ::= i::IO l::[Unit]
{
  local attribute now :: Unit;
  now = head(l);
  now.ioIn = i;

  return  if null(l) 
	  then ioInteger(i, 0)
	  else if now.code != 0
	       then ioInteger(now.io, now.code)
	       else runAll(now.io, tail(l));
}

function unitPP
[Unit] ::= l1::[Unit] l2::[Unit]
{
  return l1 +++ l2;
}



function spp
[String] ::= c1::[String] c2::[String]
{
  return c1 +++ c2;
}


synthesized attribute compiledList :: [GrammarSpec];
inherited attribute compiledGrammars :: [GrammarSpec];

synthesized attribute needGrammars :: [String];
synthesized attribute seenGrammars :: [String];

attribute needGrammars occurs on CompilationUnit;
attribute seenGrammars occurs on CompilationUnit;

abstract production compileAllExtra
top::CompilationUnit ::= grams::[String] need::[String] seen::[String]
{
  local attribute r :: Root;
  r = top.parser(head(grams));

  r.compiledGrammars = top.compiledGrammars;
  r.env = toEnv(appendDefs(r.defs, makeDefaultDefs()));
  r.globalEnv = top.globalEnv;
  r.fileName = "Generated File";

  local attribute rSpec :: Decorated RootSpec;
  rSpec = decorate consRootSpec(r, decorate emptyRootSpec() with {}) with {};

  local attribute new_seen :: [String];
  new_seen = cons(r.declaredName.name, seen);

  local attribute new_need :: [String];
  -- This line first removes from the the new grammarIncludes the ones we have already
  -- compiled (new_seen) and then appends them to the list of grammars we still need to
  -- parse.  It then makes a set of the strings so we do not compile the same grammar
  -- twice.
  new_need = makeSet(rSpec.includedGrammars +++ need);

  local attribute recurse :: CompilationUnit;
  recurse = compileAllExtra(tail(grams), new_need, new_seen);
  recurse.parser = top.parser;
  recurse.globalEnv = top.globalEnv;
  recurse.compiledGrammars = top.compiledGrammars;
 
  top.compiledList = if null(grams) then [] else cons(grammarSpec(r.declaredName.name, rSpec), recurse.compiledList);
  top.globalDefs = if null(grams) then emptyDefs() else appendDefs(rSpec.globalDefs, recurse.globalDefs);

  top.needGrammars = if null(grams) then rem(need, seen) else recurse.needGrammars;
  top.seenGrammars = if null(grams) then seen else recurse.seenGrammars;
}

nonterminal CompilationUnit with {io, compiledList, parser, globalDefs, globalEnv, compiledGrammars};
abstract production compileAll
top::CompilationUnit ::= iIn::IO sPath::[String] need::[String] seen::[String]
{
  production attribute now :: Grammar;
  now = compileGrammar(iIn, head(need), sPath);
  now.parser = top.parser;
  now.globalEnv = top.globalEnv;
  now.compiledGrammars = top.compiledGrammars;

  local attribute new_seen :: [String];
  new_seen = cons(head(need), seen);

  -- This line first removes from the the new grammarIncludes the ones we have already
  -- compiled (new_seen) and then appends them to the list of grammars we still need to
  -- parse.  It then makes a set of the strings so we do not compile the same grammar
  -- twice.
  local attribute new_need :: [String];
  new_need = makeSet(rem(now.rootSpec.includedGrammars, new_seen) +++ tail(need));

  production attribute recurse :: CompilationUnit;
  recurse = compileAll(now.io, sPath, new_need, new_seen);
  recurse.parser = top.parser;
  recurse.globalEnv = top.globalEnv;
  recurse.compiledGrammars = top.compiledGrammars;
 
  top.io = if null(need) then iIn else recurse.io;
  top.compiledList = if null(need)  
		     then [] 
		     else if !now.found 
			  then recurse.compiledList
			  else cons(grammarSpec(head(need), now.rootSpec), recurse.compiledList);

  top.globalDefs = if null(need)
		   then emptyDefs()
		   else if !now.found
		        then recurse.globalDefs
			else appendDefs(now.globalDefs, recurse.globalDefs);
}

nonterminal Grammar with {io, rootSpec, parser, globalDefs, globalEnv, compiledGrammars, name, found};
abstract production compileGrammar
top::Grammar ::= iIn::IO gn::String sPath::[String]
{
  --the grammar path ':' replaced by '/'
  local attribute gPath :: String;
  gPath = getGrammarPath(gn) ++ "/";

  -- the location (if found) of the grammar
  local attribute loc :: MaybeIOStr;
  loc = findGrammarLocation(iIn, gPath, sPath);

  -- the list of files from the grammar directory
  local attribute temp_files :: IOStringList;
  temp_files = listContents(loc.sValue, loc.io);

  -- the list of silver files for the grammar
  local attribute files :: [String];
  files = filterFiles(convert(temp_files.stringList));

  production attribute cu :: Roots;
  cu = compileFiles(print("Compiling Grammar: " ++ gn ++ "\n", temp_files.io), files, loc.sValue);
  cu.parser = top.parser;
  cu.env = toEnv(appendDefs(makeDefaultDefs(), cu.defs));
  cu.globalEnv = top.globalEnv;
  cu.compiledGrammars = top.compiledGrammars;

  top.found = loc.found;
  top.io =  if loc.found then cu.io else loc.io;
  top.rootSpec = if loc.found then cu.rootSpec else (decorate emptyRootSpec() with {});
  top.globalDefs = if loc.found then cu.globalDefs else emptyDefs();
  top.name = gn;
}

nonterminal Roots with {env, io, rootSpec, parser, defs, globalDefs, globalEnv, compiledGrammars};
abstract production compileFiles
top::Roots ::= iIn::IO files::[String] gpath::String
{
  local attribute text :: IOString;
  text = readFile(gpath ++ head(files), print("\tParsing file: " ++ gpath ++ head(files) ++ "\n", iIn));

  production attribute r :: Root;
  r = top.parser(text.sValue);
  r.env = top.env;
  r.globalEnv = top.globalEnv;
  r.fileName = head(files);
  r.compiledGrammars = top.compiledGrammars;

  top.rootSpec = if null(files) 
		 then decorate emptyRootSpec() with {} 
		 else decorate consRootSpec(r, recurse.rootSpec) with {};

  production attribute recurse :: Roots ;
  recurse = compileFiles(text.io, tail(files), gpath) ;
  recurse.parser = top.parser;
  recurse.env = top.env;
  recurse.globalEnv = top.globalEnv;
  recurse.compiledGrammars = top.compiledGrammars;

  top.io = if null(files) then iIn else recurse.io;
  top.defs = if null(files) then emptyDefs() else appendDefs(r.defs, recurse.defs);
  top.globalDefs = if null(files) then emptyDefs() else appendDefs(r.globalDefs, recurse.globalDefs);
}



--removes the strings of the first list that appear in the second list.  returns the filtered list.
function rem
[String] ::= n::[String] seen::[String]
{
  return if null(n) then [::String] 
         else if contains(head(n), seen)
	      then rem(tail(n), seen)
	      else cons(head(n), rem(tail(n), seen));
}

--takes in a list of strings and returns a set of strings.
function makeSet
[String] ::= list::[String]
{
  local attribute recurse :: [String];
  recurse = makeSet(tail(list));

  return if null(list) then list
         else if contains(head(list), recurse)
	      then recurse
	      else cons(head(list), recurse);
}

function contains
Boolean ::=  str::String sl::[String]
{
  return (!null(sl)) && (str == head(sl) || contains(str, tail(sl)));
}

--takes a asd;asdasd;adasd; string and returns a list.
function getSearchPath
[String] ::= sp::String
{
  local attribute path :: String;
  path = if indexof(":", sp) == 0 then substring(1, length(sp), sp) else sp;

  local attribute i :: Integer;
  i = indexof(":", path);

  local attribute h:: String;
  h = if i == -1 then path else substring(0, i, path);

  return if path == "" then [::String] else cons(he, ta);

--  top.empty = (path == "");

  local attribute he :: String;
  he = if (substring(length(h)-1, length(h), h) == "/") then h else h ++ "/";

  local attribute ta :: [String];
  ta = if i == -1 || i == (length(path)-1)
       then [::String]
       else getSearchPath(substring(i+1, length(path), path));
}

-- returns only valid silver files.
function filterFiles
[String] ::= files::[String]
{
  return if null(files) then files
         else if isValidSilverFile(head(files))
	      then cons(head(files), filterFiles(tail(files)))
	      else filterFiles(tail(files));
}

-- a file is valid if it has a .sv extension
function isValidSilverFile
Boolean ::= file::String
{
  local attribute l :: Integer;
  l = length(file);

  return substring(l-3, l, file) == ".sv";
}

--takes in a grammar name and returns a grammar path
function getGrammarPath
String ::= n::String
{
  return if n == "" 
	 then "" 
	 else if substring(0, 1, n) == ":"
	      then "/" ++ getGrammarPath(substring(1, length(n), n))
	      else substring(0, 1, n) ++ getGrammarPath(substring(1, length(n), n));
}


--takes in a grammar path and a list of possible locations and returns the correct location if any.
nonterminal MaybeIOStr with {sValue, found, io};
abstract production findGrammarLocation
top::MaybeIOStr ::= iIn::IO path::String paths::[String]
{
  local attribute exists :: IOBoolean;
  exists = isDirectory(head(paths) ++ path, iIn);

  top.found = if null(paths) then false
	      else exists.bValue || recurse.found;

  top.sValue = if null(paths) then ""
	       else if exists.bValue then head(paths) ++ path else recurse.sValue;

  top.io = if null(paths) then iIn else recurse.io;

  local attribute recurse :: MaybeIOStr;
  recurse = findGrammarLocation(exists.io, path, tail(paths));
}


nonterminal GrammarSpec with {rootSpec, name};

abstract production grammarSpec
top::GrammarSpec ::= n::String s::Decorated RootSpec
{
  top.name = n;
  top.rootSpec = s;
}




