grammar silver:base:core;
import core;

closed nonterminal Env with {typeTree, valueTree, attrTree, nameTree, restTree, productionTree, synthesizedTree, inheritedTree, occursTree};
synthesized attribute size :: Integer;

synthesized attribute typeTree :: Decorated ScopeEnvTree;
synthesized attribute valueTree :: Decorated ScopeEnvTree;
synthesized attribute nameTree :: Decorated ScopeEnvTree;
synthesized attribute attrTree :: Decorated ScopeEnvTree;
synthesized attribute productionTree :: Decorated ScopeEnvTree;

synthesized attribute synthesizedTree :: Decorated ScopeEnvTree;
synthesized attribute inheritedTree :: Decorated ScopeEnvTree;
synthesized attribute occursTree :: Decorated ScopeEnvTree;

synthesized attribute restTree :: Decorated ScopeEnvTree;


----------------------------------------------------------------------------------------------------
--Environment creation functions--------------------------------------------------------------------
----------------------------------------------------------------------------------------------------
abstract production internal_emptyEnv 
top::Env ::= 
{
  top.typeTree =  emptyScopeEnvTree();
  top.valueTree = emptyScopeEnvTree();
  top.nameTree =  emptyScopeEnvTree();
  top.attrTree =  emptyScopeEnvTree();
  top.productionTree = emptyScopeEnvTree();
  
  top.synthesizedTree = emptyScopeEnvTree();
  top.inheritedTree = emptyScopeEnvTree();
  top.occursTree = emptyScopeEnvTree(); 

  top.restTree =  emptyScopeEnvTree();
}

function emptyEnv
Decorated Env ::=
{
  return decorate internal_emptyEnv() with {};
}


function toEnv
Decorated Env ::= d::Decorated Defs {
  return decorate internal_toEnv(d) with {};
}

abstract production internal_toEnv
Env ::= d::Decorated Defs {

  top.typeTree =  oneScopeEnvTree(buildTree(d.typeList));
  top.valueTree = oneScopeEnvTree(buildTree(d.valueList));
  top.nameTree =  oneScopeEnvTree(buildTree(d.nameList));
  top.attrTree =  oneScopeEnvTree(buildTree(d.attrList));
  top.productionTree = oneScopeEnvTree(buildTree(d.productionList));

  top.synthesizedTree = oneScopeEnvTree(buildTree(d.synthesizedList)); 
  top.inheritedTree = oneScopeEnvTree(buildTree(d.inheritedList)); 
  top.occursTree = oneScopeEnvTree(buildTree(d.occursList)); 

  top.restTree = oneScopeEnvTree(buildTree(d.restList));
}

function appendEnv
Decorated Env ::= e1::Decorated Env  e2::Decorated Env {
  return decorate internal_appendEnv(e1, e2) with {};
}

abstract production internal_appendEnv
Env ::= e1::Decorated Env  e2::Decorated Env {

  top.typeTree =  appendScopeEnvTree(e1.typeTree, e2.typeTree);
  top.valueTree = appendScopeEnvTree(e1.valueTree, e2.valueTree);
  top.nameTree =  appendScopeEnvTree(e1.nameTree, e2.nameTree);
  top.attrTree =  appendScopeEnvTree(e1.attrTree, e2.attrTree);
  top.productionTree = appendScopeEnvTree(e1.productionTree, e2.productionTree);

  top.synthesizedTree = appendScopeEnvTree(e1.synthesizedTree, e2.synthesizedTree); 
  top.inheritedTree = appendScopeEnvTree(e1.inheritedTree, e2.inheritedTree); 
  top.occursTree = appendScopeEnvTree(e1.occursTree, e2.occursTree);

  top.restTree = appendScopeEnvTree(e1.restTree, e2.restTree);
}

function appendDefsEnv
Decorated Env ::= e1::Decorated Defs  e2::Decorated Env {
  return decorate internal_appendDefsEnv(e1, e2) with {};
}

abstract production internal_appendDefsEnv
Env ::= d::Decorated Defs e::Decorated Env {

  top.typeTree = appendScopeEnvTree(oneScopeEnvTree(buildTree(d.typeList)), e.typeTree);
  top.valueTree = appendScopeEnvTree(oneScopeEnvTree(buildTree(d.valueList)), e.valueTree);
  top.nameTree =  appendScopeEnvTree(oneScopeEnvTree(buildTree(d.nameList)), e.nameTree);
  top.attrTree =  appendScopeEnvTree(oneScopeEnvTree(buildTree(d.attrList)), e.attrTree);
  top.productionTree = appendScopeEnvTree(oneScopeEnvTree(buildTree(d.productionList)), e.productionTree);

  top.synthesizedTree = appendScopeEnvTree(oneScopeEnvTree(buildTree(d.synthesizedList)), e.synthesizedTree);
  top.inheritedTree = appendScopeEnvTree(oneScopeEnvTree(buildTree(d.inheritedList)), e.inheritedTree);
  top.occursTree = appendScopeEnvTree(oneScopeEnvTree(buildTree(d.occursList)), e.occursTree);

  top.restTree = appendScopeEnvTree(oneScopeEnvTree(buildTree(d.restList)), e.restTree);
}



----------------------------------------------------------------------------------------------------
--Environment query functions-----------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

function getDclOne
[Decorated EnvItem] ::= search::String e::Decorated ScopeEnvTree
{
  return getDclsScope(search, e);
}

function getDcl
[Decorated EnvItem] ::= search::String e::Decorated ScopeEnvTree
{
  return getDclsScope(search, e);
}

function getValueDclOne
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDclOne(search, e.valueTree);
}

function getValueDcl
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDcl(search, e.valueTree);
}

function getTypeDclOne
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDclOne(search, e.typeTree);
}

function getTypeDcl
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDcl(search, e.typeTree);
}

function getAttributeDclOne
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDclOne(search, e.attrTree);
}

function getAttributeDcl
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDcl(search, e.attrTree);
}

function getProductionDclOne
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDclOne(search, e.productionTree);
}

function getProductionDcl
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDcl(search, e.productionTree);
}

function getProductionDclSig
[Decorated EnvItem] ::= search::String tr::Decorated TypeRep e::Decorated Env
{
  return getDcl(search, e.productionTree);
}

function getProductionDclSigHelp
[Decorated EnvItem] ::= tr::Decorated TypeRep e::[Decorated EnvItem]
{
  local attribute i :: Decorated EnvItem;
  i = head(e);

  return if null(e) 
	 then [] 
	 else if i.typerep.typeEquals(i.typerep, tr).bValue
	      then cons(i, getProductionDclSigHelp(tr, tail(e)))
	      else getProductionDclSigHelp(tr, tail(e));
}

function getFullNameDclOne
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDclOne(search, e.nameTree);
}

function getFullNameDcl
[Decorated EnvItem] ::= search::String e::Decorated Env
{
  return getDcl(search, e.nameTree);
}

function doesOccurOn
--attribute, nt, env
Boolean ::= search::String dn::String e::Decorated Env
{
  return doesOccurOnHelp(dn, getDcl(search, e.occursTree));
}

function doesOccurOnHelp
Boolean ::= dn::String e::[Decorated EnvItem]
{
  return  (!null(e) && head(e).isOccursDeclaration && head(e).decoratesName == dn) 
       || (!null(e) && doesOccurOnHelp(dn, tail(e)));
}


function isSynthesized
Boolean ::= search::String e::Decorated Env
{
  return isSynthesizedHelp(getDcl(search, e.synthesizedTree));
}

function isSynthesizedHelp
Boolean ::= e::[Decorated EnvItem]
{
  return  (!null(e) && (head(e).isSynthesizedDeclaration))
       || (!null(e) && isSynthesizedHelp(tail(e)));
}

function isInherited
Boolean ::= search::String e::Decorated Env
{
  return isInheritedHelp(getDcl(search, e.inheritedTree));
}

function isInheritedHelp
Boolean ::= e::[Decorated EnvItem]
{
  return  (!null(e) && (head(e).isInheritedDeclaration))
       || (!null(e) && isInheritedHelp(tail(e)));
}

function isClosed
Boolean ::= search::String e::Decorated Env
{
  return !null(searchScopeEnvTree(search, searchIsClosed(), e.restTree));
}

abstract production searchIsClosed
top::EnvSearch ::=
{
  top.found = isClosedHelp(top.inEnvItems);
}

function isClosedHelp
Boolean ::= e::[Decorated EnvItem]
{
  return  (!null(e) && (head(e).isCloseDeclaration))
       || (!null(e) && isClosedHelp(tail(e)));
}


function getThisDcl
[Decorated EnvItem] ::= e::Decorated Env
{
  return getThisDclHelp(flattenScope(e.restTree));
}

function getThisDclHelp
[Decorated EnvItem] ::= e::[Decorated EnvItem]
{
  return if null(e)
	 then []
	 else if (head(e).isThisDeclaration) 
	      then cons(head(e), recurse)
              else recurse;

  local attribute recurse :: [Decorated EnvItem];
  recurse = getThisDclHelp(tail(e));
}


function getAspects
[StmtRep] ::= search::String e::Decorated Env
{
  return getAspectsHelp(search, flattenScope(e.restTree));
}

function getAspectsHelp
[StmtRep] ::= search::String e::[Decorated EnvItem]
{
  local attribute recurse :: [StmtRep];
  recurse = getAspectsHelp(search, tail(e));

  return if null(e) then []
         else if (head(e).isAspectDeclaration && head(e).itemName == search)
	      then head(e).stmtList +++ recurse''
	      else recurse'';
}

function getProdAtts
Decorated Defs ::= search::String e::Decorated Env
{
  return getProdAttsHelp(search, flattenScope(e.restTree));
}


function getProdAttsHelp
Decorated Defs ::= search::String e::[Decorated EnvItem]
{
  local attribute recurse :: Decorated Defs;
  recurse = getProdAttsHelp(search, tail(e));

  return if null(e) then emptyDefs()
         else if (head(e).isProdAttsDeclaration && head(e).itemName == search)
	      then head(e).defs
	      else recurse'';
}


----------------------------------------------------------------------------------------------------
--Generic environment query functions---------------------------------------------------------------
----------------------------------------------------------------------------------------------------

nonterminal EnvMap with {inEnvItem, newEnvItem};
inherited attribute inEnvItem :: Decorated EnvItem;
synthesized attribute newEnvItem :: Decorated EnvItem;

nonterminal EnvFilter with {inEnvItem, keep};
synthesized attribute keep :: Boolean;


--function printEnv
--String ::= e::Decorated Env
--{
--  return "TYPES\n" ++
--	 printEnvItems(e.typeTree) ++ "\nVALUES\n" ++
--	 printEnvItems(e.valueTree) ++ "\nNAMES\n" ++
--	 printEnvItems(e.nameTree) ++ "\nATTRS\n" ++
--	 printEnvItems(e.attrTree) ++ "\nREST\n" ++
--	 printEnvItems(e.restTree) ++ "\n\n";
--}
--

--function mapEnv
--Decorated Env ::= mapper::EnvMap e::Decorated Env
--{
--  return decorate internal_mapEnv(mapper'', e) with {};
--}
--
--abstract production internal_mapEnv
--top::Env ::= mapper::EnvMap e::Decorated Env
--{
--  top.typeTree =  mapEnvItems(mapper'', e.typeTree);
--  top.valueTree = mapEnvItems(mapper'', e.valueTree);
--  top.nameTree =  mapEnvItems(mapper'', e.nameTree);
--  top.attrTree =  mapEnvItems(mapper'', e.attrTree);
--  top.productionTree =  mapEnvItems(mapper'', e.productionTree);
--  top.restTree =  mapEnvItems(mapper'', e.restTree);
--}
--
--function filterEnv
--Decorated Env ::= fil::EnvFilter e::Decorated Env
--{
--  return decorate internal_filterEnv(fil'', e) with {};
--}
--
--abstract production internal_filterEnv
--top::Env ::= fil::EnvFilter e::Decorated Env
--{
--  top.typeTree =  filterEnvItems(fil'', e.typeTree);
--  top.valueTree = filterEnvItems(fil'', e.valueTree);
--  top.nameTree =  filterEnvItems(fil'', e.nameTree);
--  top.attrTree =  filterEnvItems(fil'', e.attrTree);
--  top.productionTree =  filterEnvItems(fil'', e.productionTree);
--  top.restTree =  filterEnvItems(fil'', e.restTree);
--}
--
--
--

abstract production reverseLookUp
top::EnvFilter ::= fName::String
{
  top.keep = top.inEnvItem.isFullNameDeclaration
	  && top.inEnvItem.fullName == fName;
}

----------------------------------------------------------------------------------------------------
--General query functions---------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

function getAllTypeDcls
[Decorated EnvItem] ::= e::Decorated Env
{
  return flattenScope(e.typeTree);
}

function getAllValueDcls
[Decorated EnvItem] ::= e::Decorated Env
{
  return flattenScope(e.valueTree);
}

function getAllAttributeDcls
[Decorated EnvItem] ::= e::Decorated Env
{
  return flattenScope(e.attrTree);
}

function getAllProductionDcls
[Decorated EnvItem] ::= e::Decorated Env
{
  return flattenScope(e.productionTree);
}

function getAllFullNameDcls
[Decorated EnvItem] ::= e::Decorated Env
{
  return flattenScope(e.nameTree);
}

function getOccursDcls
[Decorated EnvItem] ::= e::Decorated Env
{
  return flattenScope(e.occursTree);
}

function getOtherDcls
[Decorated EnvItem] ::= e::Decorated Env
{
  return flattenScope(e.restTree);
}

function getAllOccuringOn
[Decorated EnvItem] ::= dn::String e::Decorated Env
{
  return getAllOccuringOnHelp(dn, flattenScope(e.occursTree), e);
}

function getAllOccuringOnHelp
[Decorated EnvItem] ::= dn::String e::[Decorated EnvItem] fullEnv::Decorated Env
{
  return if null(e)
	 then []
	 else if (head(e).isOccursDeclaration) && (head(e).decoratesName == dn)
     	      then getAttributeDcl(head(e).itemName, fullEnv) +++ recurse
       	      else recurse;

  local attribute recurse :: [Decorated EnvItem];
  recurse = getAllOccuringOnHelp(dn, tail(e), fullEnv);
}
