grammar silver:base:core;

import core;

nonterminal ExportStmt with {pp, grammarName, location, env, errors, compiledGrammars, includedGrammars, exportedDefs} ;

concrete production exportAll
top::ExportStmt ::= i::Export_kwd  pkg::QName  wc::OptWithClause  s::Semi_t
{
  top.pp = "export " ++ pkg.pp ++ wc.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = wc.errors ++ e;

  top.includedGrammars = [pkg.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  local attribute gdefs :: Decorated Defs;
  gdefs = if pkg.name == top.grammarName.name 
	  then toDefs(top.env) 
	  else if mitem.found 
	       then mitem.rootSpec.exportedDefs 
	       else emptyDefs();

  top.exportedDefs = applyMappings(wc.envMapList, gdefs);  
}

concrete production exportOnly
top::ExportStmt ::= i::Export_kwd pkg::QName o::Only_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.pp = "export " ++ pkg.pp ++ " only " ++ ns.pp ++  wc.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = ns.errors ++ wc.errors ++ e;

  top.includedGrammars = [pkg.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  local attribute gdefs :: Decorated Defs;
  gdefs = if pkg.name == top.grammarName.name 
	  then toDefs(top.env) 
	  else if mitem.found 
	       then mitem.rootSpec.exportedDefs 
	       else emptyDefs();

  top.exportedDefs = applyMappings(wc.envMapList, filterDefs(keepFilter(ns.names), gdefs));  
}
concrete production exportHiding
top::ExportStmt ::= i::Export_kwd pkg::QName h::Hiding_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.pp = "export " ++ pkg.pp ++ " hiding " ++ ns.pp ++  wc.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = ns.errors ++ wc.errors ++ e;

  top.includedGrammars = [pkg.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  local attribute gdefs :: Decorated Defs;
  gdefs = if pkg.name == top.grammarName.name 
	  then toDefs(top.env)
	  else if mitem.found 
	       then mitem.rootSpec.exportedDefs 
	       else emptyDefs();

  top.exportedDefs = applyMappings(wc.envMapList, applyFilters(buildRemoveFilters(ns.names), gdefs));  
}
concrete production exportAs
top::ExportStmt ::= i::Export_kwd pkg1::QName a::As_kwd pkg2::QName s::Semi_t
{
  top.pp = "export " ++ pkg1.pp ++ " as " ++ pkg2.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg1.name ++ "' cannot be found.\n";

  top.errors = e;

  top.includedGrammars = [pkg1.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg1.name, top.compiledGrammars);

  local attribute gdefs :: Decorated Defs;
  gdefs = if pkg1.name == top.grammarName.name 
	  then toDefs(top.env)
	  else if mitem.found 
	       then mitem.rootSpec.exportedDefs 
	       else emptyDefs();

  top.exportedDefs = mapDefs(prependMap(pkg2.name), gdefs);  
}
