grammar silver:base:core;

import silver:utils:location hiding Int_t with parse as locParse;
import core;

closed nonterminal Expr with {pp, location, grammarName, env, errors, typerep, localsEnv, signatureEnv, globalEnv};
nonterminal Exprs with {pp, location, grammarName, env, errors, inputTypes, localsEnv, signatureEnv, globalEnv};

--Base Expressions
concrete production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.pp = "( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(l.line) ++ ", column: " ++ toString(l.column) ++ "] ";

  top.typerep = e.typerep;

  top.errors = e.errors;
}

concrete production baseExpr
top::Expr ::= q::QName
{
  top.pp = q.pp; 
  top.location = q.location;

  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName ::String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute vals :: [Decorated EnvItem];
  vals = getValueDclOne(fName, top.env);

  production attribute in_sig ::Boolean;
  in_sig = !null(getValueDclOne(fName, top.signatureEnv));

  production attribute in_locals ::Boolean;
  in_locals = !null(getValueDclOne(fName, top.localsEnv));

  --to define the type
  forwards to if (in_sig || in_locals) && !null(vals) && head(vals).typerep.doDecorate && (top.userFriendly != -1)
	      then decorateExpr(q'')
	      else if !in_sig && !in_locals && !null(vals) 
		&& head(vals).typerep.typeName != "TOP"
		 && head(vals).typerep.isProduction 
	      then productionReference(q'')
	      else dontDecorateExpr(q'') ;
}
terminal IdTick_t / [A-Za-z][A-Za-z0-9\_]*[\'] / lexer precedence = 2 ;
concrete production concreteDecorateExpr
top::Expr ::= q::IdTick_t
{
  forwards to decorateExpr(qNameId(terminal(Id_t, substring(0, length(q.lexeme) -1, q.lexeme), q.line, q.column)));
}

abstract production decorateExpr
top::Expr ::= q::QName
{
  top.pp = q.pp ++ "'"; 
  top.location = q.location;

  --the full name of the identifier
  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName ::String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute vals :: [Decorated EnvItem];
  vals = getValueDclOne(fName, top.env);

  production attribute origTypeRep :: Decorated TypeRep;
  origTypeRep = if !null(fNames) && !null(vals) 
		then head(vals).typerep 
		else topTypeRep();

  production attribute in_sig ::Boolean;
  in_sig = !null(getValueDclOne(fName, top.signatureEnv));

  production attribute in_locals ::Boolean;
  in_locals = !null(getValueDclOne(fName, top.localsEnv));

  local attribute er1 :: String;
  er1 = if null(fNames) || null(vals)
	then top.location ++ "Error: Value '" ++ q.name ++ "' is not declared.\n" 
	else "";

  top.errors = er1;

  top.typerep = if !null(fNames) && !null(vals) 
		then refTypeRep(head(vals).typerep) 
		else topTypeRep();
}

terminal IdTickTick_t / [A-Za-z][A-Za-z0-9\_]*[\'][\'] / lexer precedence = 2 ;
concrete production concreteDontDecorateExpr
top::Expr ::= q::IdTickTick_t
{
  forwards to dontDecorateExpr(qNameId(terminal(Id_t, substring(0, length(q.lexeme) -2, q.lexeme), q.line, q.column)));
}

abstract production dontDecorateExpr
top::Expr ::= q::QName
{
  top.pp = q.pp; 
  top.location = q.location;

  --the full name of the identifier
  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName ::String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute vals :: [Decorated EnvItem];
  vals = getValueDclOne(fName, top.env);

  production attribute in_sig ::Boolean;
  in_sig = !null(getValueDclOne(fName, top.signatureEnv));

  production attribute in_locals ::Boolean;
  in_locals = !null(getValueDclOne(fName, top.localsEnv));

  local attribute er1 :: String;
  er1 = if null(fNames) || null(vals)
	then top.location ++ "Error: Value '" ++ q.name ++ "' is not declared.\n" 
	else "";

  top.errors = er1;
  top.typerep = if !null(fNames) && !null(vals) then head(vals).typerep else topTypeRep();
}

abstract production productionReference
top::Expr ::= q::QName
{
  top.pp = q.pp; 
  top.location = q.location;

  --the full name of the identifier
  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName ::String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute vals :: [Decorated EnvItem];
  vals = getValueDclOne(fName, top.env);

  local attribute er1 :: String;
  er1 = if null(fNames) || null(vals)
	then top.location ++ "Error: Value '" ++ q.name ++ "' is not declared.\n" 
	else "";

  top.errors = er1;


  top.typerep = if !null(fNames) && !null(vals) then head(vals).typerep else topTypeRep();
}



--function buildExprs
--Exprs ::= ts::[Decorated TypeRep] es::[Decorated Expr] {
--
--
--  local attribute dExpr :: Expr;
--  dExpr = exprWith(terminal(Decorate_kwd, "decorate"),
--	     	      new(head(es)), 
--	     	      terminal(With_kwd, "with"),
--	     	      terminal(LCurly_t, "{"),
--	              exprStmtsEmpty(),
--		      terminal(RCurly_t, "}"));
--
--  local attribute nExpr :: Expr;
--  nExpr = newFunction(terminal(New_kwd, "new"),
--		      terminal(LParen_t, "("),
--  		      new(head(es)),
--		      terminal(RParen_t, ")"));
--
--  local attribute top :: Boolean;
--  top = head(ts).typeName == "TOP" && head(es).typerep.typeName == "TOP";
--
--  local attribute noCast :: Boolean;
--  noCast = head(ts).typeEquals(head(ts), head(es).typerep).bValue;
--
--  local attribute castUp :: Boolean;
--  castUp = head(ts).isDecorated && head(es).typerep.isNonTerminal && head(ts).typeEquals(head(ts), refTypeRep(head(es).typerep)).bValue;
--
--  local attribute castDown :: Boolean;
--  castDown = head(ts).isNonTerminal && head(es).typerep.isDecorated && head(ts).typeEquals(head(ts), head(es).typerep.decoratedType).bValue;
--
--  production attribute casted :: Expr;
--  casted = if top || noCast then new(head(es)) else (if castUp then dExpr'' else (if castDown'' then nExpr'' else new(head(es))));
--
--  return if null(es) || length(es) != length(ts) then exprsEmpty()
--	else exprsCons(casted'', terminal(Comma_t, ","), buildExprs(tail(ts), tail(es)));
--
--}
--

concrete production productionApp
top::Expr ::= e::Expr l::LParen_t es::Exprs r::RParen_t
{
--  local attribute newEs:: Exprs;
--  newEs = buildExprs(e.typerep.inputTypes, es.exprList);
  e.userFriendly = 0;
  forwards to e.typerep.applicationDispatcher(e'', es'');
}

concrete production emptyProductionApp
top::Expr ::= e::Expr l::LParen_t r::RParen_t
{
  e.userFriendly = 0;
  forwards to e.typerep.applicationDispatcher(e'', exprsEmpty());
}


abstract production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  top.pp = e.pp ++ "(" ++ es.pp ++ ")";
  top.location = e.location;
  top.errors = e.errors ++ es.errors; 

  top.typerep = if e.typerep.isProduction then e.typerep.outputType else topTypeRep();

  es.expectedInputTypes = e.typerep.inputTypes;
}

concrete production attributeAccess
top::Expr ::= e::Expr d::Dot_t q::QName
{
  --decorate
  e.userFriendly = 1;

  forwards to if e.typerep.isDecorated || e.typerep.isTerminal
       	      then atAccess(e'', terminal(At_t, "@", d.line, d.column), q'')
       	      else hashAccess(e'', terminal(Hash_t, "#", d.line, d.column), q'')
	with {userFriendly = 1;};
}

terminal Hash_t /[\#]/ lexer precedence = 5, precedence = 25, association = left;
concrete production hashAccess
top::Expr ::= e::Expr d::Hash_t q::QName
{
  top.pp = e.pp ++ " # " ++ q.pp;
  top.location = e.location;

    local attribute withPart :: Expr;
  withPart = exprWith(terminal(Decorate_kwd, "decorate", d.line, d.column),
	     	      e'', 
	     	      terminal(With_kwd, "with", d.line, d.column),
	     	      terminal(LCurly_t, "{", d.line, d.column),
	              exprStmtsEmpty(),
		      terminal(RCurly_t, "}", d.line, d.column));

  forwards to atAccess(withPart'', terminal(At_t, "@", d.line, d.column), q'');
}

terminal At_t /[\@]/ lexer precedence = 5, precedence = 25, association = left;
concrete production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  top.pp = e.pp ++ " . " ++ q.pp;
  top.location = e.location;

  top.errors = er ++ e.errors;

  e.userFriendly = 1;

  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName ::String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED" ++ "_" ++ q.name;

  production attribute occursOn :: Boolean;
  occursOn = if !null(fNames)
      	     then doesOccurOn(fName, e.typerep.typeName, top.env)
             else false;	

  production attribute vals :: [Decorated EnvItem];
  vals = getAttributeDclOne(fName, top.env);

  local attribute er :: String;
  er = if null(fNames) || null(vals)
	then top.location ++ "Error: Attribute '" ++ q.name ++ "' is not declared.\n" 
	else "";
  
  top.typerep = if occursOn && !null(vals) then head(vals).typerep else topTypeRep();
}

terminal Decorate_kwd /decorate/ lexer precedence = 5;
concrete production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  top.pp = "eval " ++ e.pp ++ " with { " ++ inh.pp ++ " }";
  top.location = "[line: " ++ toString(ev.line) ++ ", column: " ++ toString(ev.column) ++ "] ";

  top.typerep = refTypeRep(e.typerep);
  top.errors = e.errors ++ inh.errors;

  e.userFriendly = -1;
}


nonterminal ExprInh with {pp, location, grammarName, env, errors, localsEnv, signatureEnv, globalEnv};
concrete production exprStmtsEmpty
top::ExprInh ::= 
{
  top.pp = "";
  top.location = "[line: -1, column: -1, emptyProductionBody()] ";
  top.errors = "";
}

concrete production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 
  top.pp = lhs.pp ++ " = " ++ e.pp ++ ";\n    " ++ inh.pp;
  top.location = lhs.location;
  top.errors = lhs.errors ++ e.errors ++ inh.errors;

  e.userFriendly = lhs.typerep.userFriendlyLHS;
}


nonterminal ExprLHSExpr with {pp, location, grammarName, env, errors, typerep};

-- EVW added from translation:ag:haskell:core:Expr.sv - not sure why it was there.
attribute fullName occurs on ExprLHSExpr;

concrete production exprLhsExpr
top::ExprLHSExpr ::= q::QName
{
  top.pp = q.pp;
  top.location = q.location;

  --the full name of the identifier
  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName ::String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute vals :: [Decorated EnvItem];
  vals = getAttributeDclOne(fName, top.env);

  local attribute er1 :: String;
  er1 = if null(vals)
	then top.location ++ "Error: Value '" ++ q.name ++ "' is not declared.\n" 
	else "";

  top.errors = er1;
  top.typerep = if !null(fNames) && !null(vals) then head(vals).typerep else topTypeRep();

}

nonterminal BOperator with {binaryProduction, pp};
synthesized attribute binaryProduction :: Production(Expr ::= Expr Expr);

concrete production bOperatorPlusPlus
top::BOperator ::= t::PlusPlus_t
{
  top.pp = t.lexeme;
  top.binaryProduction = plusPlusWrapper;
}

abstract production plusPlusWrapper
top::Expr ::= e1::Expr e2::Expr
{
  local attribute loc :: Location;
  loc = locParse(e1.location);

  forwards to plusPlus(e1'', terminal(PlusPlus_t, "++", loc.locLine, loc.locColumn), e2'');
}


--normal expressions
--------------------------------------------------
--------------------------------------------------

terminal True_kwd /true/ lexer precedence = 5;
concrete production trueConst
top::Expr ::= t::True_kwd
{
  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";
  top.pp = t.lexeme;
  top.errors = "";
  top.typerep = booleanTypeRep();
}

terminal False_kwd /false/ lexer precedence = 5;
concrete production falseConst
top::Expr ::= f::False_kwd
{
  top.location = "[line: " ++ toString(f.line) ++ ", column: " ++ toString(f.column) ++ "] ";
  top.pp = f.lexeme;
  top.errors = "";
  top.typerep = booleanTypeRep();
}

terminal And_t /[\&][\&]/ 
  lexer precedence = 5,
  precedence = 5,
  association = left;

concrete production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.pp = e1.pp ++ " && " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}


terminal Or_t /[\|][\|]/ 
  lexer precedence = 5,
  precedence = 5,
  association = left;
concrete production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.pp = e1.pp ++ " || " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}

terminal Not_t /[\!]/
  lexer precedence = 5,
  precedence = 7;

concrete production not
top::Expr ::= n::Not_t e::Expr
{
  top.pp = "! " ++ e.pp;
  top.location = "[line: " ++ toString(n.line) ++ ", column: " ++ toString(n.column) ++ "] ";
  top.typerep = booleanTypeRep();

  top.errors = e.errors;
}
--BREAK

terminal GT_t /[\>]/ 
  lexer precedence = 5,
  precedence = 9,
  association = left;
concrete production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.pp = e1.pp ++ " > " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}

terminal LT_t /[\<]/
  lexer precedence = 5,
  precedence = 9,
  association = left;
concrete production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.pp = e1.pp ++ " < " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}

terminal GTEQ_t /[\>][\=]/
  lexer precedence = 6,
  precedence = 9,
  association = left;

concrete production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.pp = e1.pp ++ " >= " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}


terminal LTEQ_t /[\<][\=]/
  lexer precedence = 6,
  precedence = 9,
  association = left;

concrete production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.pp = e1.pp ++ " <= " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}

terminal EQEQ_t /[\=][\=]/
  lexer precedence = 6,
  precedence = 9,
  association = left;

concrete production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.pp = e1.pp ++ " == " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}

terminal NEQ_t /[\!][\=]/
  lexer precedence = 6,
  precedence = 9,
  association = left;

concrete production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.pp = e1.pp ++ " != " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = booleanTypeRep();
}

terminal If_kwd /if/ lexer precedence = 5;
terminal Then_kwd /then/ lexer precedence = 5;
terminal Else_kwd /else/ lexer precedence = 5;
concrete production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
precedence = 0
{
  top.pp = "if " ++ e1.pp ++ " then " ++ e2.pp ++ " else " ++ e3.pp;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  top.errors = e1.errors ++ e2.errors ++ e3.errors;

  top.typerep = e2.typerep;
}

concrete production intConst
top::Expr ::= i::Int_t
{
  top.pp = i.lexeme;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";
  top.errors = "";
  top.typerep = integerTypeRep();
}

terminal Float_t /[0-9]+[\.][0-9]+/ lexer precedence = 2;
concrete production floatConst
top::Expr ::= f::Float_t
{
  top.pp = f.lexeme;
  top.location = "[line: " ++ toString(f.line) ++ ", column: " ++ toString(f.column) ++ "] ";
  top.errors = "";
  top.typerep = floatTypeRep();
} 

terminal Plus_t /[\+]/
  lexer precedence = 5,
  precedence = 11,
  association = left;

concrete production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.pp = e1.pp ++ " + " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
	       
  top.typerep = e1.typerep;
}

terminal Minus_t /[\-]/
  lexer precedence = 5,
  precedence = 11,
  association = left;

concrete production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.pp = e1.pp ++ " - " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
       
  top.typerep = e1.typerep;
}

terminal Multiply_t /[\*]/
  lexer precedence = 5,
  precedence = 12,
  association = left;

concrete production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.pp = e1.pp ++ " * " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
	       
  top.typerep = e1.typerep;
}

terminal Divide_t /[\/]/
  lexer precedence = 5,
  precedence = 12,
  association = left;

concrete production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.pp = e1.pp ++ " / " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
	       
  top.typerep = e1.typerep;
}

concrete production neg
top::Expr ::= m::Minus_t e::Expr
precedence = 13
{
  top.pp = "-" ++ e.pp;
  top.location = "[line: " ++ toString(m.line) ++ ", column: " ++ toString(m.column) ++ "] ";
  top.errors = e.errors;
  top.typerep = e.typerep;
}


concrete production stringConst
top::Expr ::= s::String_t
{
  top.pp = s.lexeme;
  top.location = "[line: " ++ toString(s.line) ++ ", column: " ++ toString(s.column) ++ "] ";
  top.errors = "";
  top.typerep = stringTypeRep();
}

terminal PlusPlus_t /[\+][\+]/
  lexer precedence = 6,
  precedence = 11,
  association = left;

concrete production plusPlus
top::Expr ::= e1::Expr p::PlusPlus_t e2::Expr
{
  production attribute handler :: [Expr] with epp;
  handler := [];

  forwards to if null(handler) then defaultPlusPlus(e1, e2) else head(handler);
}


aspect production plusPlus
top::Expr ::= e1::Expr p::PlusPlus_t e2::Expr
{
  handler <- if e1.typerep.typeName == "String" && e2.typerep.typeName == "String"
	     then [stringPlusPlus(e1, e2)]
	     else [];
}
abstract production stringPlusPlus
top::Expr ::= e1::Decorated Expr e2::Decorated Expr
{
  top.pp = e1.pp ++ " ++ " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = stringTypeRep();
}

abstract production defaultPlusPlus
top::Expr ::= e1::Decorated Expr e2::Decorated Expr
{
  top.pp = e1.pp ++ " ++ " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.typerep = stringTypeRep();
}

inherited attribute expectedInputTypes :: [Decorated TypeRep];
attribute expectedInputTypes occurs on Exprs;

abstract production exprsEmpty
top::Exprs ::=
{
  top.pp = "";
  top.location =  "[line: -1, column: -1, emptyExprs()] ";
  top.errors = "";
  top.inputTypes = [];
}

concrete production exprsSingle
top::Exprs ::= e::Expr
{
  top.pp = e.pp;
  top.location = e.location;

  top.errors = e.errors;
  top.inputTypes = [e.typerep];

  e.userFriendly = if null(top.expectedInputTypes) then 0 else head(top.expectedInputTypes).userFriendlyLHS;
}

concrete production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.pp = e1.pp ++ ", " ++ e2.pp;
  top.location = e1.location;

  top.errors = e1.errors ++ e2.errors;
  top.inputTypes = cons(e1.typerep, e2.inputTypes);

  e1.userFriendly = if null(top.expectedInputTypes) then 0 else head(top.expectedInputTypes).userFriendlyLHS;
  e2.expectedInputTypes = if null(top.expectedInputTypes) then [] else tail(top.expectedInputTypes);
}
