grammar silver:base:core;
import core;

nonterminal ImportStmt with {pp, grammarName, location, importedEnv, env, errors, includedGrammars, compiledGrammars};

synthesized attribute importedEnv :: Decorated Env;

concrete production importAll
top::ImportStmt ::= i::Import_kwd  pkg::QName  wc::OptWithClause  s::Semi_t
{
  top.pp = "import " ++ pkg.pp ++ wc.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = wc.errors ++ e;

  top.includedGrammars = [pkg.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  top.importedEnv = if mitem.found 
		    then if null(wc.envMapList) 
			 then mitem.rootSpec.exportedEnv 
			 else toEnv(applyMappings(wc.envMapList, mitem.rootSpec.exportedDefs))
		    else emptyEnv();
}

concrete production importOnly
top::ImportStmt ::= i::Import_kwd pkg::QName o::Only_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.pp = "import " ++ pkg.pp ++ " only " ++ ns.pp ++  wc.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = ns.errors ++ wc.errors ++ e;

  top.includedGrammars = [pkg.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  local attribute gdefs :: Decorated Defs;
  gdefs = if mitem.found then mitem.rootSpec.exportedDefs else emptyDefs();

  top.importedEnv = toEnv(applyMappings(wc.envMapList, filterDefs(keepFilter(ns.names), gdefs)));  
}
concrete production importHiding
top::ImportStmt ::= i::Import_kwd pkg::QName h::Hiding_kwd ns::NameList wc::OptWithClause  s::Semi_t 
{
  top.pp = "import " ++ pkg.pp ++ " hiding " ++ ns.pp ++  wc.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg.name ++ "' cannot be found.\n";

  top.errors = ns.errors ++ wc.errors ++ e;

  top.includedGrammars = [pkg.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg.name, top.compiledGrammars);

  local attribute gdefs :: Decorated Defs;
  gdefs = if mitem.found then mitem.rootSpec.exportedDefs else emptyDefs();

  top.importedEnv = toEnv(applyMappings(wc.envMapList, applyFilters(buildRemoveFilters(ns.names), gdefs)));  
}

concrete production importAs
top::ImportStmt ::= i::Import_kwd pkg1::QName a::As_kwd pkg2::QName s::Semi_t
{
  top.pp = "import " ++ pkg1.pp ++ " as " ++ pkg2.pp ++ " ;" ;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  local attribute e :: String;
  e = if mitem.found then "" else top.location ++ "Error: Package '" ++ pkg1.name ++ "' cannot be found.\n";

  top.errors = e;

  top.includedGrammars = [pkg1.name];

  production attribute mitem :: MaybeRootSpec;
  mitem = getGrammar(pkg1.name, top.compiledGrammars);

  local attribute gdefs :: Decorated Defs;
  gdefs = if mitem.found then mitem.rootSpec.exportedDefs else emptyDefs();

  top.importedEnv = toEnv(mapDefs(prependMap(pkg2.name), gdefs));  
}


