grammar silver:base:core;
import core;

nonterminal ProductionBody with {pp, stmtList, location, grammarName, defs, env, errors, signatureEnv, localsEnv, globalEnv, prodDefs};
nonterminal ProductionStmts with {pp, stmtList, location, grammarName, defs, env, errors, signatureEnv, localsEnv, prodDefs, globalEnv};
nonterminal ProductionStmt with {pp, stmtList, location, grammarName, defs, env, errors, signatureEnv, localsEnv, prodDefs, globalEnv};
nonterminal AttributeDef with {pp, stmtList, location, grammarName, defs, env, errors, signatureEnv, localsEnv, globalEnv};
closed nonterminal LHSExpr with {pp, location, grammarName, env, errors, typerep, signatureEnv, localsEnv};

concrete production emptyProductionBodySemi
top::ProductionBody ::= s::Semi_t
{ 
  forwards to emptyProductionBody() ;
}

concrete production emptyProductionBodyCurly
top::ProductionBody ::= l::LCurly_t r::RCurly_t
{ 
  forwards to emptyProductionBody() ;
}

abstract production emptyProductionBody
top::ProductionBody ::=
{ 
  production attribute extraStmts :: ProductionStmts with productionStmtsAppend ;
  extraStmts := productionStmtsNone() ;

  forwards to defaultProductionBody(extraStmts'') ;
}

concrete production productionBody
top::ProductionBody ::= l::LCurly_t stmts::ProductionStmts r::RCurly_t
{
  production attribute extraStmts :: ProductionStmts with productionStmtsAppend ;
  extraStmts := stmts'';

  forwards to defaultProductionBody ( extraStmts'' ) ;
}

abstract production defaultProductionBody
top::ProductionBody ::= stmts::ProductionStmts
{
  top.pp = stmts.pp;
  top.location = stmts.location;
  top.defs = stmts.defs;
  top.prodDefs = stmts.prodDefs;

  top.errors = stmts.errors;
  top.stmtList = stmts.stmtList;
}

abstract production productionStmtsNone
top::ProductionStmts ::= 
{
  top.pp = "";
  top.location = "[line: -1, column: -1, productionStmtsNone()] ";
  top.defs = emptyDefs();
  top.prodDefs = emptyDefs();

  top.errors = "";
  top.stmtList = [];
}

concrete production productionStmtsSingle
top::ProductionStmts ::= stmt::ProductionStmt
{
  top.pp = "  " ++ stmt.pp;
  top.location = stmt.location;
  top.defs = stmt.defs;
  top.prodDefs = stmt.prodDefs;

  top.errors = stmt.errors;
  top.stmtList = stmt.stmtList;
}

concrete production productionStmtsCons
top::ProductionStmts ::= h::ProductionStmt t::ProductionStmts
{
  top.pp = "  " ++ h.pp ++ "\n" ++ t.pp;
  top.location = h.location;
  top.defs = appendDefs(h.defs, t.defs);
  top.prodDefs = appendDefs(h.prodDefs, t.prodDefs);

  top.errors = h.errors ++ t.errors;
  top.stmtList = h.stmtList +++ t.stmtList;
}

abstract production productionStmtsAppend
top::ProductionStmts ::= h::ProductionStmts t::ProductionStmts
{
  top.pp = "  " ++ h.pp ++ "\n" ++ t.pp;
  top.location = h.location;
  top.defs = appendDefs(h.defs, t.defs);
  top.prodDefs = appendDefs(h.prodDefs, t.prodDefs);

  top.errors = h.errors ++ t.errors;
  top.stmtList = h.stmtList +++ t.stmtList;
}

terminal Forwards_kwd /forwards/ lexer precedence = 5;
terminal To_kwd /to/ lexer precedence = 5;

concrete production forwardsDcl
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr s::Semi_t
{
  top.pp = "forwards to " ++ e.pp ++ ";";
  top.location = "[line: " ++ toString(f.line) ++ ", column: " ++ toString(f.column) ++ "] ";
  top.defs = emptyDefs();

  top.prodDefs = emptyDefs();

  top.errors = e.errors;
  top.stmtList = [forwardsDclStmtRep(e.typerep, [], e)];

  e.userFriendly = 1;
}


inherited attribute decorates_type :: TypeRep;
concrete production forwardsDclWith
top::ProductionStmt ::= f::Forwards_kwd t::To_kwd e::Expr w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  top.pp = "forwards to " ++ e.pp ++ "with {\n    " ++ inh.pp ++ "\n  };";
  top.location = "[line: " ++ toString(f.line) ++ ", column: " ++ toString(f.column) ++ "] ";
  top.defs = emptyDefs();

  top.prodDefs = emptyDefs();

  top.errors = e.errors ++ inh.errors;
  top.stmtList = [forwardsDclStmtRep(e.typerep, inh.stmtList, e)];

  e.userFriendly = 1;
}


terminal Forwarding_kwd /forwarding/ lexer precedence = 5;
--used in aspects.
concrete production forwardingDclWith
top::ProductionStmt ::= f::Forwarding_kwd w::With_kwd l::LCurly_t inh::ForwardInh r::RCurly_t s::Semi_t
{
  top.pp = "forwarding with {" ++ inh.pp ++ "};";
  top.location = "[line: " ++ toString(f.line) ++ ", column: " ++ toString(f.column) ++ "] ";
  top.defs = emptyDefs();

  top.prodDefs = emptyDefs();

  top.errors = "";
  top.stmtList = [forwardingDclStmtRep(inh.stmtList)];
}

nonterminal ForwardInh with {pp, stmtList, location, grammarName, env, errors, localsEnv, signatureEnv, globalEnv};
concrete production forwardStmtsEmpty
top::ForwardInh ::= 
{
  top.pp = "";
  top.location = "[line: -1, column: -1, emptyProductionBody()] ";
  top.errors = "";
  top.stmtList = [];
}

concrete production forwardStmtsCons
top::ForwardInh ::= lhs::ForwardLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ForwardInh
{ 
  top.pp = lhs.pp ++ " = " ++ e.pp ++ ";\n    " ++ inh.pp;
  top.location = lhs.location;
  top.errors = lhs.errors ++ e.errors ++ inh.errors;
  top.stmtList = cons(attributeDefStmtRep(lhs.lhsExpr, e), inh.stmtList);

  e.userFriendly = lhs.typerep.userFriendlyLHS;
}

nonterminal ForwardLHSExpr with {pp, lhsExpr, location, grammarName, env, errors, typerep};
concrete production forwardLhsExpr
top::ForwardLHSExpr ::= q::QName
{
  top.pp = q.pp;
  top.location = q.location;

  --the full name of the identifier
  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(q.name, top.env);

  production attribute fName :: String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute attrs :: [Decorated EnvItem];
  attrs = getAttributeDclOne(fName, top.env);

  local attribute er1 :: String;
  er1 = if null(attrs)
	then top.location ++ "Error: Value '" ++ q.name ++ "' is not declared.\n" 
	else "";

  top.errors = er1;
  top.typerep = if !null(fNames) && !null(attrs) then head(attrs).typerep else topTypeRep();

  production attribute lhsE :: LHSExpr;
  lhsE = fakeLHSExpr(q'', top.typerep);
  lhsE.env = top.env;
  lhsE.grammarName = top.grammarName;
  
  top.lhsExpr = lhsE;
}

terminal Local_kwd /local/ lexer precedence = 5 ;
concrete production attributeDclLocal
top::ProductionStmt ::= l::Local_kwd  at::Attribute_kwd a::Id_t 
                        ht::HasType_t te::Type  s::Semi_t
{
  top.pp = "local attribute " ++ a.lexeme ++ " :: " ++ te.pp ++ " ;" ;
  top.location = "[line: " ++ toString(l.line) ++ ", column: " ++ toString(l.column) ++ "] ";

  production attribute fName :: String;
  fName = a.lexeme;

  top.prodDefs = emptyDefs();

  top.defs = addValueDcl(fName, te.typerep, 
	     addFullNameDcl(a.lexeme, fName,  emptyDefs()));

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ a.lexeme ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(fName, top.env)) > 1
       then top.location ++ "Error: Value '" ++ fName ++ "' is already bound.\n"
       else "";	

  top.errors = er1 ++ er2 ++ te.errors;
  top.stmtList = [localDclStmtRep(a.lexeme, fName, te.typerep)];
}

concrete production attributeDclProduction
top::ProductionStmt ::= p::Production_kwd  at::Attribute_kwd a::Id_t 
                        ht::HasType_t te::Type  s::Semi_t
{
  top.pp = "production attribute " ++ a.lexeme ++ " :: " ++ te.pp ++ " ;" ;
  top.location = "[line: " ++ toString(p.line) ++ ", column: " ++ toString(p.column) ++ "] ";

  production attribute fName :: String;
  fName = a.lexeme;

  top.prodDefs = addValueDcl(fName, te.typerep, 
  	         addFullNameDcl(a.lexeme, fName,  emptyDefs()));

  forwards to attributeDclLocal(terminal(Local_kwd, "local", p.line, p.column), at, a, ht, te'', s);
}

concrete production stmtToAttrDef
top::ProductionStmt ::= a::AttributeDef
{
  top.pp = a.pp;
  top.location = a.location;
  top.defs = a.defs;
  top.prodDefs = emptyDefs();
  top.errors = a.errors;
  top.stmtList = a.stmtList;
}

concrete production concreteAttrDef
top::AttributeDef ::= lhs::LHSExpr eq::Equal_t e::Expr s::Semi_t
{
--  --These are defined here so that we can lookup things in the environment without causing a loop.
--  top.defs = emptyDefs();
--  top.stmtList = [attributeDefStmtRep(lhs, e)];
--
--  local attribute dExpr :: Expr;
--  dExpr = exprWith(terminal(Decorate_kwd, "decorate", eq.line, eq.column),
--	     	      e'', 
--	     	      terminal(With_kwd, "with", eq.line, eq.column),
--	     	      terminal(LCurly_t, "{", eq.line, eq.column),
--	              exprStmtsEmpty(),
--		      terminal(RCurly_t, "}", eq.line, eq.column));
--
--  local attribute nExpr :: Expr;
--  nExpr = newFunction(terminal(New_kwd, "new", eq.line, eq.column),
--		      terminal(LParen_t, "(", eq.line, eq.column),
--  		      e'',
--		      terminal(RParen_t, ")", eq.line, eq.column));
--
--
--  local attribute isTop :: Boolean;
--  isTop = lhs.typerep.typeName == "TOP" && e.typerep.typeName == "TOP";
--
--  local attribute noCast :: Boolean;
--  noCast = lhs.typerep.typeEquals(lhs.typerep, e.typerep).bValue;
--
--  local attribute castUp :: Boolean;
--  castUp = lhs.typerep.isDecorated && e.typerep.isNonTerminal && lhs.typerep.typeEquals(lhs.typerep, refTypeRep(e.typerep)).bValue;
--
--  local attribute castDown :: Boolean;
--  castDown = lhs.typerep.isNonTerminal && e.typerep.isDecorated && lhs.typerep.typeEquals(lhs.typerep, e.typerep.decoratedType).bValue;
--
--  production attribute casted :: Expr;
--  casted = if isTop || noCast then e'' else (if castUp then dExpr'' else (if castDown'' then nExpr'' else e''));
--
--  casted.env = top.env;
--  casted.globalEnv = top.globalEnv;
--  casted.localsEnv = top.localsEnv;
--  casted.signatureEnv = top.signatureEnv;  
--
  forwards to attrDef(lhs'', e'');
}

autocopy attribute userFriendly :: Integer;
attribute userFriendly occurs on Expr;
--Values 0=unknown, 1=decorate, -1=new


--The name of this production should really be userFriendlyAttrDef
abstract production attrDef
top::AttributeDef ::= lhs::LHSExpr e::Expr
{
  top.pp = lhs.pp ++ " = " ++ e.pp ++ " ;";
  top.location = lhs.location ;
  top.defs = emptyDefs();
  top.errors = lhs.errors ++ e.errors;
  top.stmtList = [attributeDefStmtRep(lhs, e)];

  e.userFriendly = lhs.typerep.userFriendlyLHS;
}


abstract production fakeLHSExpr
top::LHSExpr ::= c1::QName c2::Decorated TypeRep
{
  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(c1.name, top.env);

  production attribute fName :: String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_FOUND";

  top.pp = c1.name;
  top.location = c1.location ;
  top.typerep = c2;
  top.errors = "";
}

concrete production lhsExprOne
top::LHSExpr ::= id::Id_t
{
  top.pp = id.lexeme;
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";

  --the full name of the identifier
  production attribute fNames :: [Decorated EnvItem];
  fNames = getFullNameDclOne(id.lexeme, top.env);

  production attribute fName :: String;
  fName = if !null(fNames) then head(fNames).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  local attribute vals :: [Decorated EnvItem];
  vals = getValueDclOne(fName, top.env);

  local attribute er1 :: String;
  er1 = if null(fNames) || null(vals)
	then top.location ++ "Error: Value '" ++ id.lexeme ++ "' is not declared.\n" 
	else "";

  top.errors = er1;
  top.typerep = if !null(fNames) && !null(vals) then head(vals).typerep else topTypeRep();
}

concrete production lhsExprTwo
top::LHSExpr ::= id::Id_t d::Dot_t q::QName
{
  top.pp = id.lexeme ++ " . " ++ q.pp;
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";

  production attribute fNames1 :: [Decorated EnvItem];
  fNames1 = getFullNameDclOne(id.lexeme, top.env);

  production attribute fName1 :: String;
  fName1 = if !null(fNames1) then head(fNames1).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute vals1 :: [Decorated EnvItem];
  vals1 = getValueDclOne(fName1, top.env);

  local attribute er1 :: String;
  er1 = if null(fNames1) || null(vals1)
	then top.location ++ "Error: Value '" ++ id.lexeme ++ "' is not declared.\n" 
  	else "";

  production attribute fNames2 :: [Decorated EnvItem];
  fNames2 = getFullNameDclOne(q.name, top.env);

  production attribute fName2 :: String;
  fName2 = if !null(fNames2) then head(fNames2).fullName else "FULL_NAME_NOT_DEFINED";

  --whether it is bound to a value
  production attribute vals2 :: [Decorated EnvItem];
  vals2 = getAttributeDclOne(fName2, top.env);

  local attribute er2 :: String;
  er2 = if null(fNames2) || null(vals2)
	then top.location ++ "Error: Attribute '" ++ q.name ++ "' is not declared.\n" 
        else "";

  top.errors = er1 ++ er2;
  top.typerep = if !null(fNames2) && !null(vals2) then head(vals2).typerep else topTypeRep();
}

