grammar silver:base:core;
import core;

synthesized attribute pp :: String;

synthesized attribute errors :: String;

inherited attribute fileName ::String;

--the location in line and column of this production.
synthesized attribute location :: String ;

--the name of the grammar this abstract syntax tree is working in.
synthesized attribute declaredName :: Decorated QName;
autocopy attribute grammarName :: Decorated QName;

--an attribute to collect definitions up the syntax tree.
synthesized attribute defs :: Decorated Defs;
synthesized attribute globalDefs :: Decorated Defs;

--an attribute to pass down the definitons of the language.
autocopy attribute env :: Decorated Env;
autocopy attribute globalEnv :: Decorated Env;

--the defs that we want other people to use.
synthesized attribute exportedDefs :: Decorated Defs;

start nonterminal Root with {pp, location, defs, declaredName, errors, includedGrammars, compiledGrammars, exportedDefs, env, globalEnv, globalDefs, fileName} ; 
close Root;

abstract production root
top::Root ::= gdcl::GrammarDcl ims::ImportExportStmts  ags::AGDcls
{
  production attribute extraImports :: ImportExportStmts with appendImportsExports;
  extraImports := importStmtsNone ();

  top.pp = gdcl.pp ++ "\n\n" ++ allimports.pp ++ "\n\n" ++ ags.pp;
  top.location = gdcl.location;
  top.declaredName = gdcl.declaredName;
  
  top.globalDefs = ags.globalDefs;
  top.defs = appendDefs(ags.defs, allimports.localDefs);

  local attribute es :: String;
  es = gdcl.errors ++ allimports.errors ++ ags.errors;

  top.errors = if es == "" then "" else ("****** " ++ top.fileName ++ "******\n" ++ es);

  local attribute allimports :: ImportExportStmts;
  allimports = appendImportsExports (ims'', extraImports'');

  top.includedGrammars = allimports.includedGrammars;
  top.exportedDefs = allimports.exportedDefs;

  ims.grammarName = gdcl.declaredName;
  ims.env = top.env;
  ims.compiledGrammars = top.compiledGrammars;

  allimports.grammarName = gdcl.declaredName;
  allimports.env = top.env;
  allimports.compiledGrammars = top.compiledGrammars;

  ags.grammarName = gdcl.declaredName;
  ags.env = appendEnv(allimports.importedEnv, top.env);
  ags.globalEnv = top.globalEnv;
}


concrete production rootImportsDcls
top::Root ::= gdcl::GrammarDcl ims::ImportExportStmts  ags::AGDcls
{
  forwards to root (gdcl'', ims'', ags'');
}

concrete production rootNoImports
top::Root ::= gdcl::GrammarDcl ags::AGDcls
{
  forwards to root (gdcl'', importStmtsNone (), ags'');  
}

concrete production rootNoDcls
top::Root ::= gdcl::GrammarDcl ims::ImportExportStmts
{
  forwards to root (gdcl'', ims'', agDclsOne (agDclNone ()));
}

concrete production rootNoImportsNoDcls
top::Root ::= gdcl::GrammarDcl 
{
  forwards to root (gdcl'', importStmtsNone (), agDclsOne (agDclNone ()));
}

-- GrammarDcl
-------------
nonterminal GrammarDcl with {pp, declaredName, location, errors} ;
terminal Grammar_kwd /grammar/  lexer precedence = 5 ;

concrete production grammarDcl
top::GrammarDcl ::= g::Grammar_kwd  qn::QName  s::Semi_t
{
  top.pp = "grammar " ++ qn.pp ++ " ;";
  top.declaredName = qn;
  top.location = "[line: " ++ toString(g.line) ++ ", column: " ++ toString(g.column) ++ "] ";

  top.errors = "";
  qn.grammarName = top.declaredName;
}

-- AG Dcls
----------
nonterminal AGDcl with {pp, grammarName, location, defs, env, errors, globalEnv, globalDefs} ;
nonterminal AGDcls with {pp, grammarName, location, defs, env, errors, globalEnv, globalDefs};




concrete production agDclsOne
top::AGDcls ::= ag::AGDcl
{
  top.pp = ag.pp ;
  top.location = ag.location;

  top.defs = ag.defs;
  top.globalDefs = ag.globalDefs;
  top.errors = ag.errors;
}

concrete production agDclsCons
top::AGDcls ::= h::AGDcl t::AGDcls
{
  top.pp = h.pp ++ "\n\n" ++ t.pp ;
  top.location = h.location;

  top.defs = appendDefs(h.defs, t.defs);
  top.globalDefs = appendDefs(h.globalDefs, t.globalDefs);
  top.errors = h.errors ++ t.errors;
}

abstract production agDclSeq
top::AGDcl ::= ag1::AGDcl ag2::AGDcl
{
  top.pp = ag1.pp ++ "\n\n" ++ ag2.pp ;
  top.location = ag1.location;

  top.defs = appendDefs(ag1.defs, ag2.defs);
  top.globalDefs = appendDefs(ag1.globalDefs, ag2.globalDefs);
  top.errors = ag1.errors ++ ag2.errors;
}

abstract production agDclNone
top::AGDcl ::=
{
  top.pp = "";
  top.location = "[line: -1, column: -1, agDclNone()] ";
  top.defs = emptyDefs();
  top.globalDefs = emptyDefs ();
  top.errors = "";
}
