grammar silver:base:regex:concretesyntax;
export silver:base:regex:concretesyntax;

import silver:base:regex:abstractsyntax;

synthesized attribute regexpr_abst :: Regex_abst;
synthesized attribute rangeexpr_abst :: Range_abst;

terminal Plus_t /[\+]/
    lexer precedence = 5;
terminal Kleene_t /[\*]/
     lexer precedence = 5;
terminal Optional_t /[\?]/
     lexer precedence = 5;
terminal Choice_t /[\|]/
     lexer precedence = 5;
terminal Range_t /[\-]/
     lexer precedence = 5;
terminal RegexNot_t /[\^]/
     lexer precedence = 5;
terminal RegexLBrack_t /[\[]/
     lexer precedence = 5;
terminal RegexRBrack_t /[\]]/
     lexer precedence = 5;
terminal RegexLParen_t /[\(]/
     lexer precedence = 5;
terminal RegexRParen_t /[\)]/
     lexer precedence = 5;
terminal RegexWildcard_t /[\.]/
     lexer precedence = 4;
terminal RegexChar_t /./
     lexer precedence = 2;
terminal EscapedChar_t /\\./
     lexer precedence = 3;

start nonterminal Regex_R with {regexpr_abst};
nonterminal Regex_DR with {regexpr_abst};
nonterminal Regex_UR with {regexpr_abst};
nonterminal Regex_RR with {regexpr_abst};
nonterminal Regex_G with {rangeexpr_abst};
nonterminal Regex_RG with {rangeexpr_abst};
nonterminal Regex_UG with {rangeexpr_abst};
nonterminal Regex_CHAR with {regexpr_abst};

concrete production RtoDR
r::Regex_R ::= dr::Regex_DR
{
     r.regexpr_abst = dr.regexpr_abst;
}

concrete production RtoDR_bar_R
r::Regex_R ::= first::Regex_DR sep::Choice_t rest::Regex_R
{
     r.regexpr_abst = regex_choice(first.regexpr_abst,rest.regexpr_abst);
}

concrete production DRtoUR_RR
dr::Regex_DR ::= first::Regex_UR rest::Regex_RR
{
     dr.regexpr_abst = regex_concat(first.regexpr_abst,rest.regexpr_abst);
}

concrete production DRtoUR_star_RR
dr::Regex_DR ::= first::Regex_UR sep::Kleene_t rest::Regex_RR
{
     dr.regexpr_abst = regex_kleene(first.regexpr_abst,rest.regexpr_abst);
}

concrete production DRtoUR_plus_RR
dr::Regex_DR ::= first::Regex_UR sep::Plus_t rest::Regex_RR
{
     dr.regexpr_abst = regex_plus(first.regexpr_abst,rest.regexpr_abst);
}

concrete production DRtoUR_question_RR
dr::Regex_DR ::= first::Regex_UR sep::Optional_t rest::Regex_RR
{
     dr.regexpr_abst = regex_optional(first.regexpr_abst,rest.regexpr_abst);
}

concrete production RRtoDR
rr::Regex_RR ::= dr::Regex_DR
{
     rr.regexpr_abst = dr.regexpr_abst;
}

concrete production RRtoeps
rr::Regex_RR ::=
{
     rr.regexpr_abst = regex_empty();
}



concrete production URtoCHAR
ur::Regex_UR ::= char::Regex_CHAR
{
     ur.regexpr_abst = char.regexpr_abst;
}

concrete production URtowildcard
ur::Regex_UR ::= wildcard::RegexWildcard_t
{
     ur.regexpr_abst = regex_wildcard();
}

concrete production URtolb_G_rb
ur::Regex_UR ::= lb::RegexLBrack_t g::Regex_G rb::RegexRBrack_t
{
     ur.regexpr_abst = regex_posCharrange(g.rangeexpr_abst);
}

concrete production URtolb_not_G_rb
ur::Regex_UR ::= lb::RegexLBrack_t sep::RegexNot_t g::Regex_G rb::RegexRBrack_t
{
     ur.regexpr_abst = regex_negCharrange(g.rangeexpr_abst);
}

concrete production URtolp_R_rp
ur::Regex_UR ::= lp::RegexLParen_t r::Regex_R rp::RegexRParen_t
{
     ur.regexpr_abst = regex_parenthesized(r.regexpr_abst);
}

concrete production GtoUG_RG
g::Regex_G ::= ug::Regex_UG rg::Regex_RG
{
     g.rangeexpr_abst = range_rangeunion(ug.rangeexpr_abst,rg.rangeexpr_abst);
}

concrete production UGtoCHAR
ug::Regex_UG ::= char::Regex_CHAR
{
     ug.rangeexpr_abst = range_loosechar(char.regexpr_abst);
}

concrete production UGtoCHAR_dash_CHAR
ug::Regex_UG ::= leastchar::Regex_CHAR sep::Range_t greatestchar::Regex_CHAR
{
     ug.rangeexpr_abst = range_rangechar(leastchar.regexpr_abst,greatestchar.regexpr_abst);
}

concrete production RGtoG
rg::Regex_RG ::= g::Regex_G
{
     rg.rangeexpr_abst = g.rangeexpr_abst;
}

concrete production RGtoeps
rg::Regex_RG ::=
{
     rg.rangeexpr_abst = range_empty();
}

concrete production CHARtochar
charRegex::Regex_CHAR ::= char::RegexChar_t
{
     charRegex.regexpr_abst = regex_char(char.lexeme);
}

concrete production CHARtoescaped
charRegex::Regex_CHAR ::= esc::EscapedChar_t
{
     charRegex.regexpr_abst = regex_escaped_char(esc.lexeme);
}
