grammar silver:base:type:io;
import silver:base:core;

terminal ReadFile_kwd 		/readFile/ 	lexer precedence = 5;
terminal WriteFile_kwd 		/writeFile/	lexer precedence = 5;
terminal AppendFile_kwd 	/appendFile/	lexer precedence = 5;

terminal EnvVar_kwd 		/envVar/ 	lexer precedence = 5;
terminal System_kwd 		/system/ 	lexer precedence = 5;
terminal CWD_kwd 		/cwd/   	lexer precedence = 5;

terminal IsFile_kwd 		/isFile/ 	lexer precedence = 5;
terminal IsDirectory_kwd 	/isDirectory/ 	lexer precedence = 5;

terminal Print_kwd 		/print/ 	lexer precedence = 5;

terminal ListContents_kwd	/listContents/	lexer precedence = 5;

terminal UnsafeIO_kwd 		/unsafeio/ 	lexer precedence = 5;
terminal GenInt_kwd 		/genint/ 	lexer precedence = 5;

concrete production isFileFunction
top::Expr ::= i::IsFile_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "isFile ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.errors = e1.errors ++ e2.errors;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  top.typerep = ntTypeRep("core:IOBoolean");
}

concrete production isDirectoryFunction
top::Expr ::= i::IsDirectory_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "isDirectory ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.errors = e1.errors ++ e2.errors;
  top.location = "[line: " ++ toString(i.line) ++ ", column: " ++ toString(i.column) ++ "] ";

  top.typerep = ntTypeRep("core:IOBoolean");
}

concrete production readFunction
top::Expr ::= r::ReadFile_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr rp::RParen_t
{
  top.pp = "readFile ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.location = "[line: " ++ toString(r.line) ++ ", column: " ++ toString(r.column) ++ "] ";
  top.errors = e1.errors ++ e2.errors;

  top.typerep = ntTypeRep("core:IOString");
}

concrete production cwdFunction
top::Expr ::= c::CWD_kwd l::LParen_t e::Expr r::RParen_t
{
  top.pp = "cwd ( " ++ e.pp ++ " )";
  top.location = "[line: " ++ toString(c.line) ++ ", column: " ++ toString(c.column) ++ "] ";
  top.errors = e.errors;

  top.typerep = ntTypeRep("core:IOString");
}

concrete production envVarFunction
top::Expr ::= e::EnvVar_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "envVar ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.location = "[line: " ++ toString(e.line) ++ ", column: " ++ toString(e.column) ++ "] ";
  top.errors = e1.errors ++ e2.errors;

  top.typerep = ntTypeRep("core:IOString");
}

concrete production systemFunction
top::Expr ::= s::System_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "system ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.location = "[line: " ++ toString(s.line) ++ ", column: " ++ toString(s.column) ++ "] ";
  top.errors = e1.errors ++ e2.errors;

  top.typerep = ntTypeRep("core:IOInteger");
}


concrete production writeFunction
top::Expr ::= w::WriteFile_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{
  top.pp = "writeFile ( " ++ e1.pp ++ ", " ++ e2.pp ++ ", " ++ e3.pp ++ " )";
  top.location = "[line: " ++ toString(w.line) ++ ", column: " ++ toString(w.column) ++ "] ";
  top.errors = e1.errors ++ e2.errors ++ e3.errors;

  top.typerep = ioTypeRep();
}

concrete production appendFunction
top::Expr ::= a::AppendFile_kwd l::LParen_t e1::Expr c1::Comma_t e2::Expr c2::Comma_t e3::Expr r::RParen_t
{
  top.pp = "appendFile ( " ++ e1.pp ++ ", " ++ e2.pp ++ ", " ++ e3.pp ++ " )";
  top.location = "[line: " ++ toString(a.line) ++ ", column: " ++ toString(a.column) ++ "] ";
  top.errors = e1.errors ++ e2.errors ++ e3.errors;

  top.typerep = ioTypeRep();
}

concrete production printFunction
top::Expr ::= p::Print_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "print ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.location = "[line: " ++ toString(p.line) ++ ", column: " ++ toString(p.column) ++ "] ";
  top.errors = e1.errors ++ e2.errors;

  top.typerep = ioTypeRep();
}

concrete production listContentsFunction
top::Expr ::= li::ListContents_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "listContents ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.location = "[line: " ++ toString(li.line) ++ ", column: " ++ toString(li.column) ++ "] ";
  top.errors = e1.errors ++ e2.errors;

  top.typerep = ntTypeRep("core:IOStringList");
}

concrete production genIntFunction
top::Expr ::= g::GenInt_kwd l::LParen_t r::RParen_t
{
  top.pp = "genint (  )";
  top.location = "[line: " ++ toString(g.line) ++ ", column: " ++ toString(g.column) ++ "] ";
  top.errors = "";

  top.typerep = integerTypeRep();
}

concrete production unsafeIOFunction
top::Expr ::= u::UnsafeIO_kwd
{
  top.pp = "unsafeio";
  top.location = "[line: " ++ toString(u.line) ++ ", column: " ++ toString(u.column) ++ "] ";
  top.errors = "";

  top.typerep = ioTypeRep();
}
