grammar silver:extension:autocopy;
import silver:base;

concrete production qNameAutyCopy
top::QName ::= id::AutoCopy_kwd
{
  forwards to qNameId(terminal(Id_t, "autocopy", id.line, id.column));
}

function lhsExprTwo_
LHSExpr ::= id::Id_t q::QName
{
 return lhsExprTwo (id, terminal(Dot_t,"."), q'') ;
}

terminal AutoCopy_kwd /autocopy/ lexer precedence = 5;

concrete production autocopyDcl
top::AGDcl ::= a::AutoCopy_kwd at::Attribute_kwd n::Id_t ht::HasType_t t::Type s::Semi_t
{
  top.pp = "autocopy attribute " ++ n.lexeme ++ " :: " ++ t.pp ++ " ;" ;
  top.location = "[line: " ++ toString(a.line) ++ ", column: " ++ toString(a.column) ++ "] ";

  production attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ n.lexeme;

  top.defs = addAttributeDcl(fName, copyTypeRep(t.typerep), 
             addInheritedDcl(fName, 
	     addFullNameDcl(n.lexeme, fName,  emptyDefs())));


  top.globalDefs = addAttributeDcl(fName, copyTypeRep(t.typerep), 
                   addInheritedDcl(fName, emptyDefs()));

  t.grammarName = top.grammarName;
  t.env = top.env;

  forwards to attributeDclInh(terminal(Inherited_kwd, ""), at, n, ht, t'', s);
}

aspect production emptyProductionBody
top::ProductionBody ::=
{
  local attribute all_stmts :: [StmtRep];
  all_stmts = getAspects(top.grammarName.name ++ ":" ++ top.pname, top.globalEnv);

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(head(top.signatureTypes).typeName, top.globalEnv);

  local attribute allAttrs :: [Decorated EnvItem];
  allAttrs = filterEnvItems(keepAutoCopy(), decorating);

  local attribute fNames :: [Decorated EnvItem];
  fNames = getAllFullNameDcls(top.pureEnv);

  extraStmts <- if top.isAspect then productionStmtsNone()
	            else makeAllCopyDefs(fNames, top.globalEnv, allAttrs, 
					 head(top.signatureNames),
					 head(top.signatureTypes),
					 tail(top.signatureNames),
					 tail(top.signatureTypes),
					 all_stmts); 
}

aspect production productionBody 
top::ProductionBody ::= l::LCurly_t stmts::ProductionStmts r::RCurly_t
{
  local attribute all_stmts :: [StmtRep];
  all_stmts = stmts.stmtList +++ getAspects(top.grammarName.name ++ ":" ++ top.pname, top.globalEnv);

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(head(top.signatureTypes).typeName, top.globalEnv);

  local attribute allAttrs :: [Decorated EnvItem];
  allAttrs = filterEnvItems(keepAutoCopy(), decorating);

  local attribute fNames :: [Decorated EnvItem];
  fNames = getAllFullNameDcls(top.pureEnv);

  extraStmts <- if top.isAspect then productionStmtsNone()
	            else makeAllCopyDefs(fNames, top.globalEnv, allAttrs, 
					 head(top.signatureNames),
					 head(top.signatureTypes),
					 tail(top.signatureNames),
					 tail(top.signatureTypes),
					 all_stmts); 
}

function makeAllCopyDefs
ProductionStmts ::= fNames::[Decorated EnvItem] ge::Decorated Env attrs::[Decorated EnvItem]
		    lhsN::String lhsT::Decorated TypeRep 
	            rhsNs::[String] rhsTs::[Decorated TypeRep]
	            stmts::[StmtRep]
{
  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(head(rhsTs).typeName, ge);

  local attribute allAttrs :: [Decorated EnvItem];
  allAttrs = filterEnvItems(keepAutoCopy(), decorating);

  local attribute sharedAttrs :: [Decorated EnvItem];
  sharedAttrs = filterEnvItems(keepShared(attrs), allAttrs);

  return if null(rhsTs) then productionStmtsNone()
	 else productionStmtsAppend(makeCopyDefs(fNames, sharedAttrs, lhsN, lhsT, head(rhsNs), head(rhsTs), stmts),
	      makeAllCopyDefs(fNames, ge, attrs, lhsN, lhsT, tail(rhsNs), tail(rhsTs), stmts));
}

function makeCopyDefs
ProductionStmts ::= fNames::[Decorated EnvItem] attrs::[Decorated EnvItem] 
		    lhsN::String lhsT::Decorated TypeRep 
	            rhsN::String rhsT::Decorated TypeRep
	            stmts::[StmtRep]
{
  local attribute aNames :: [Decorated EnvItem];
  aNames = filterEnvItems(reverseLookUp(head(attrs).itemName), fNames);

  --the name of the attribute in this grammar, need to do reverse lookup since it could be aliased.
  local attribute aName :: String;
--  aName = if null(aNames) then head(attrs).itemName
--	  else head(aNames).itemName;

  aName = head(attrs).itemName;
	  

  local attribute need :: Boolean;
  need = !containsDef(rhsN, head(attrs).itemName, stmts);

  local attribute attrdef :: ProductionStmt;
  attrdef = stmtToAttrDef(attrDef(lhsExprTwo_(terminal(Id_t, rhsN), qNameId(terminal(Id_t, aName))), 
	                           attributeAccess(baseExpr(qNameId(terminal(Id_t, lhsN))),
					           terminal(Dot_t, "."),
						   qNameId(terminal(Id_t, aName)))));
  
  local attribute result :: ProductionStmts;
--  result = if null(aNames)
--	   then productionStmtsWrap(head(attrs), lhsT, rhsT, attrdef'')	    
--	   else	productionStmtsSingle(attrdef'');
--
  -- just to be safe we wrap the attribute definition wil all of the information needed.
  result = productionStmtsWrap(head(attrs), lhsT, rhsT, attrdef'');
 
  return if !null(attrs) && need 
	 then productionStmtsAppend(result'', makeCopyDefs(fNames, tail(attrs), lhsN, lhsT, rhsN, rhsT, stmts))
	 else if !null(attrs)
	      then makeCopyDefs(fNames, tail(attrs), lhsN, lhsT, rhsN, rhsT, stmts)
	      else productionStmtsNone();
}

abstract production productionStmtsWrap
top::ProductionStmts ::= n::Decorated EnvItem lhsT::Decorated TypeRep rhsT::Decorated TypeRep stmt1::ProductionStmt
{
  forwards to productionStmtsSingle(stmt1'') 
	with {
		env = appendDefsEnv(addAttributeDcl(n.itemName, n.typerep, 
		      addInheritedDcl(n.itemName, 
		      addOccursDcl(n.itemName, lhsT.typeName, 
		      addOccursDcl(n.itemName, rhsT.typeName, 
		      addFullNameDcl(n.itemName, n.itemName, emptyDefs()))))), 
		      top.env);
	     };
}


abstract production keepAutoCopy
top::EnvFilter ::=
{
  top.keep = top.inEnvItem.isAttributeDeclaration
	  && top.inEnvItem.typerep.isAutoCopy;
}

abstract production keepShared
top::EnvFilter ::= eis::[Decorated EnvItem]
{
  top.keep = containsItem(top.inEnvItem.itemName, eis);
}

function containsDef 
Boolean ::= n::String attr::String stmts::[StmtRep]
{
  return !null(stmts) && ((head(stmts).isAttributeDef
	   &&  head(stmts).lhsExpr.firstName == n
           &&  head(stmts).lhsExpr.secondName == attr)
	   ||  containsDef(n, attr, tail(stmts)));
}


synthesized attribute firstName :: String;
synthesized attribute secondName :: String;

attribute firstName occurs on LHSExpr;
attribute secondName occurs on LHSExpr;

aspect production lhsExprTwo
top::LHSExpr ::= id::Id_t d::Dot_t q::QName
{
  local attribute newFNames1 :: [Decorated EnvItem];
  newFNames1 = getFullNameDcl(id.lexeme, top.pureEnv);

  local attribute newFName1 :: String;
  newFName1 = if !null(newFNames1) then head(newFNames1).fullName else "FULL_NAME_NOT_DEFINED";


  local attribute newFNames2 :: [Decorated EnvItem];
  newFNames2 = getFullNameDcl(q.name, top.pureEnv);

  local attribute newFName2 :: String;
  newFName2 = if !null(newFNames2) then head(newFNames2).fullName else "FULL_NAME_NOT_DEFINED";


  top.firstName = newFName1;
  top.secondName = newFName2;
}

aspect production lhsExprOne
top::LHSExpr ::= id::Id_t
{
  local attribute newFNames :: [Decorated EnvItem];
  newFNames = getFullNameDcl(id.lexeme, top.pureEnv);

  local attribute newFName :: String;
  newFName = if !null(newFNames) then head(newFNames).fullName else "FULL_NAME_NOT_DEFINED";


  top.firstName = newFName;
  top.secondName = "";
}

aspect production fakeLHSExpr
top::LHSExpr ::= c1::QName c2::Decorated TypeRep
{
  local attribute newFNames :: [Decorated EnvItem];
  newFNames = getFullNameDcl(c1.name, top.pureEnv);

  local attribute newFName :: String;
  newFName = if !null(newFNames) then head(newFNames).fullName else "FULL_NAME_NOT_DEFINED";



  top.firstName = "";
  top.secondName = newFName;
}

autocopy attribute pureEnv :: Decorated Env;
attribute pureEnv occurs on ProductionBody, ProductionStmt, ProductionStmts, AttributeDef, LHSExpr, AGDcl;

inherited attribute isAspect :: Boolean;
attribute isAspect occurs on ProductionBody;

inherited attribute pname :: String;
attribute pname occurs on ProductionBody;

aspect production productionDcl
 top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{
  body.pname = id.lexeme;
  body.isAspect = false;
  body.pureEnv = appendDefsEnv(ns.defs, top.env);
}

aspect production aspectAGDcl
 top::AGDcl ::=
   a::Aspect_kwd p::Production_kwd id::QName ns::AspectSignature
   body::ProductionBody
{
  body.pname = id.name;
  body.isAspect = true;
  body.pureEnv = appendDefsEnv(ns.defs, top.env);
}

