grammar silver:extension:collection;
import silver:base;

nonterminal NameOrBOperator with {hasOperator, operationName, binaryProduction, pp};

concrete production nameOperator
top::NameOrBOperator ::= q::QName
{
  top.pp = q.pp;
  top.hasOperator = false;
  top.operationName = q;
}

concrete production binaryOperator
top::NameOrBOperator ::= q::BOperator
{
  top.pp = q.pp;
  top.hasOperator = true;
  top.binaryProduction = q.binaryProduction;
}

concrete production collectionAttributeDclSyn
top::AGDcl ::= sn::Synthesized_kwd at::Attribute_kwd a::Id_t ht::HasType_t te::Type w::With_kwd q::NameOrBOperator s::Semi_t
{
  top.pp = "synthesized attribute " ++ a.lexeme ++ " :: " ++ te.pp ++ " with " ++ q.pp ++ " ;" ;

  local attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ a.lexeme;

  local attribute lookup :: [Decorated EnvItem];
  lookup = getFullNameDcl(q.operationName.name, top.env);

  local attribute opN :: String;
  opN = if null(lookup)
        then "OPERATION_NOT_DEFINED"
	else head(lookup).fullName;

  local attribute tr :: Decorated TypeRep;
  tr = if q.hasOperator
       then bOperatorCollectionTypeRep(q.binaryProduction, te.typerep)
       else collectionTypeRep(q.operationName, opN, te.typerep);

  top.defs = addAttributeDcl(fName, tr,
             addSynthesizedDcl(fName, 
	     addFullNameDcl(a.lexeme, fName,  emptyDefs())));

  top.globalDefs = addAttributeDcl(fName, tr,
                   addSynthesizedDcl(fName, emptyDefs()));

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(a.lexeme, top.env)) > 1
        then top.location ++ "Error: Name '" ++ a.lexeme ++ "' is already bound.\n"
        else "";	

  local attribute er2 :: String;
  er2 = if length(getAttributeDcl(fName, top.env)) > 1
        then top.location ++ "Error: Attribute '" ++ fName ++ "' is already bound.\n"
        else "";	

  local attribute er3 :: String;
  er3 = if !q.hasOperator && null(lookup) 
	then top.location ++ "Error: Undefined name for collection operator: '" ++ q.operationName.name ++ "'\n"
	else "";

  top.errors = er1 ++ er2 ++ er3 ++ te.errors;

  forwards to attributeDclSyn(sn, at, a, ht, te'', s);
}

concrete production colleactionAttributeDclInh
top::AGDcl ::= i::Inherited_kwd at::Attribute_kwd a::Id_t ht::HasType_t te::Type w::With_kwd q::NameOrBOperator s::Semi_t
{
  top.pp = "inherited attribute " ++ a.lexeme ++ " :: " ++ te.pp ++ " with " ++ q.pp ++ " ;" ;

  local attribute fName :: String;
  fName = top.grammarName.name ++ ":" ++ a.lexeme;

  local attribute lookup :: [Decorated EnvItem];
  lookup = getFullNameDcl(q.operationName.name, top.env);

  local attribute opN :: String;
  opN = if null(lookup)
        then "OPERATION_NOT_DEFINED"
	else head(lookup).fullName;

  local attribute tr :: Decorated TypeRep;
  tr = if q.hasOperator
       then bOperatorCollectionTypeRep(q.binaryProduction, te.typerep)
       else collectionTypeRep(q.operationName, opN, te.typerep);

  top.defs = addAttributeDcl(fName, tr,
             addInheritedDcl(fName, 
	     addFullNameDcl(a.lexeme, fName,  emptyDefs())));

  top.globalDefs = addAttributeDcl(fName, tr,
                   addInheritedDcl(fName, emptyDefs()));

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(a.lexeme, top.env)) > 1
        then top.location ++ "Error: Name '" ++ a.lexeme ++ "' is already bound.\n"
        else "";	

  local attribute er2 :: String;
  er2 = if length(getAttributeDcl(fName, top.env)) > 1
        then top.location ++ "Error: Attribute '" ++ fName ++ "' is already bound.\n"
        else "";	

  local attribute er3 :: String;
  er3 = if !q.hasOperator && null(lookup) 
	then top.location ++ "Error: Undefined name for collection operator: '" ++ q.operationName.name ++ "'\n"
	else "";

  top.errors = er1 ++ er2 ++ er3 ++ te.errors;

  forwards to attributeDclInh(i, at, a, ht, te'', s);
}

concrete production collectionAttributeDclProd
top::ProductionStmt ::= p::Production_kwd at::Attribute_kwd a::Id_t ht::HasType_t te::Type w::With_kwd q::NameOrBOperator s::Semi_t
{
  top.pp = "production attribute " ++ a.lexeme ++ " :: " ++ te.pp ++ " with " ++ q.pp ++ " ;" ;

  local attribute fName :: String;
  fName = a.lexeme;

  local attribute lookup :: [Decorated EnvItem];
  lookup = getFullNameDcl(q.operationName.name, top.env);

  local attribute opN :: String;
  opN = if null(lookup)
        then "OPERATION_NOT_DEFINED"
	else head(lookup).fullName;

  local attribute tr :: Decorated TypeRep;
  tr = if q.hasOperator
       then bOperatorCollectionTypeRep(q.binaryProduction, te.typerep)
       else collectionTypeRep(q.operationName, opN, te.typerep);

  top.defs = addValueDcl(fName, tr, 
	     addFullNameDcl(a.lexeme, fName,  emptyDefs()));

  top.prodDefs = addValueDcl(fName, tr,
  	         addFullNameDcl(a.lexeme, fName,  emptyDefs()));

  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(a.lexeme, top.env)) > 1
        then top.location ++ "Error: Name '" ++ a.lexeme ++ "' is already bound.\n"
        else "";	

  local attribute er2 :: String;
  er2 = if length(getAttributeDcl(fName, top.env)) > 1
        then top.location ++ "Error: Attribute '" ++ fName ++ "' is already bound.\n"
        else "";	

  local attribute er3 :: String;
  er3 = if !q.hasOperator && null(lookup) 
	then top.location ++ "Error: Undefined name for collection operator: '" ++ q.operationName.name ++ "'\n"
	else "";

  top.errors = er1 ++ er2 ++ er3 ++ te.errors;

  forwards to attributeDclProduction(p, at, a, ht, te'', s);
}

terminal Contains_t   /[\<][\-]/   lexer precedence = 5 ;
concrete production attrContains
top::AttributeDef ::= lhs::LHSExpr c::Contains_t e::Expr s::Semi_t
{
  top.pp = lhs.pp ++ " = " ++ e.pp ++ " ;";
  top.location = lhs.location ;
  top.defs = emptyDefs();
  
  top.errors = lhs.errors ++ e.errors;
  top.stmtList = [collectionStmtRep(lhs, wrapExpr(e'', top.env, top.localsEnv, top.signatureEnv))];

  forwards to attrDef(lhs'', e'');
}

terminal BaseContains_t   /[\:][\=]/   lexer precedence = 5 ;
concrete production attrContainsBase
top::AttributeDef ::= lhs::LHSExpr c::BaseContains_t e::Expr s::Semi_t
{
  local attribute this :: [Decorated EnvItem];
  this = getThisDcl(top.env);

  local attribute stmts :: [StmtRep];
  stmts = getAspects(head(this).itemName, top.globalEnv);

  local attribute myStmts :: [StmtRep];
  myStmts = getCollections(lhs'', stmts);

  local attribute newExpr :: Expr;
  newExpr = if lhs.typerep.hasOperator 
	    then foldCollectionOperator(lhs.typerep.binaryProduction, myStmts, e'')
	    else foldCollection(lhs.typerep.operationName, myStmts, e'');

  forwards to attrDef(lhs'', newExpr'')
	with {
		env = appendDefsEnv(addFullNameDcl(lhs.typerep.operationName.name, lhs.typerep.operationFullName, emptyDefs()), top.env);
	     };
}

abstract production wrapExpr
top::Expr ::= e::Expr bEnv::Decorated Env lEnv::Decorated Env sEnv::Decorated Env
{
  forwards to e''
	with {
		env = bEnv; 
		localsEnv = lEnv; 
		signatureEnv = sEnv;
	     };
}

function getCollections
[StmtRep] ::= lhs::LHSExpr stmts::[StmtRep]
{
  return if null(stmts) then []
	 else if head(stmts).isCollection 
	      && (head(stmts).collectionLHS.pp == lhs.pp)
	      then cons(head(stmts), getCollections(lhs'', tail(stmts)))
	      else getCollections(lhs'', tail(stmts));
}

function foldCollection
Expr ::= app::Decorated QName stmts::[StmtRep] expr::Expr
{
  return if null(stmts)
         then expr'' 
         else productionApp(baseExpr(new(app)), 
		            terminal(LParen_t, "("),
		            exprsCons(
			  	  head(stmts).collectionRHS,
				  terminal(Comma_t, ","),
  				  exprsSingle(foldCollection(app, tail(stmts), expr''))),
	  	            terminal(RParen_t, ")"));
}

function foldCollectionOperator
Expr ::= p::Production(Expr ::= Expr Expr) stmts::[StmtRep] expr::Expr
{
  return if null(stmts)
         then expr'' 
         else p(head(stmts).collectionRHS, foldCollectionOperator(p, tail(stmts), expr''));
}


synthesized attribute isCollection :: Boolean;
synthesized attribute collectionLHS :: Decorated LHSExpr;
synthesized attribute collectionRHS :: Expr;

attribute isCollection occurs on StmtRep;
attribute collectionLHS occurs on StmtRep;
attribute collectionRHS occurs on StmtRep;

abstract production collectionStmtRep
top::StmtRep ::= lhs::Decorated LHSExpr rhs::Expr
{
  top.collectionLHS = lhs;
  top.collectionRHS = rhs'';

  top.isCollection = true;
  
  forwards to defaultStmtRep();
}


aspect production defaultStmtRep
top::StmtRep ::= 
{
  top.collectionLHS = error("Collection: DOES NOT MAKE SENSE");
  top.collectionRHS = error("Collection: DOES NOT MAKE SENSE");

  top.isCollection = false;
}

synthesized attribute operationName :: Decorated QName;
synthesized attribute operationFullName :: String;
synthesized attribute hasOperator :: Boolean; 

attribute typerep occurs on TypeRep;
attribute operationName occurs on TypeRep;
attribute operationFullName occurs on TypeRep;
attribute binaryProduction occurs on TypeRep;
attribute hasOperator occurs on TypeRep;
attribute isCollection occurs on TypeRep;

function collectionTypeRep
Decorated TypeRep ::= opName::Decorated QName fullOpName::String t::Decorated TypeRep
{
  return decorate internal_collectionTypeRep(opName, fullOpName, t) with {}; 
}

abstract production internal_collectionTypeRep
top::TypeRep ::= opName::Decorated QName fullOpName::String t::Decorated TypeRep
{
  top.operationName = opName;
  top.operationFullName = fullOpName;
  top.typerep = t;
  top.isCollection = true;
  top.hasOperator = false;
                             
  forwards to new(t);
}

function bOperatorCollectionTypeRep
Decorated TypeRep ::= p::Production(Expr ::= Expr Expr) t::Decorated TypeRep
{
  return decorate internal_bOperatorCollectionTypeRep(p, t) with {}; 
}
abstract production internal_bOperatorCollectionTypeRep
top::TypeRep ::= p::Production(Expr ::= Expr Expr) t::Decorated TypeRep
{
  top.typerep = t;
  top.isCollection = true;
  top.hasOperator = true;   
  top.binaryProduction = p;

  forwards to new(t);
}

aspect production internal_defaultTypeRep
top::TypeRep ::= 
{
  top.operationName = decorate makeQName("NO_OPERATION_DEFINED") with {};
  top.operationFullName = "NO_OPERATION_DEFINED";
  top.typerep = topTypeRep();
  top.hasOperator = false;
  top.isCollection = false;
}