grammar silver:extension:convenience;
import silver:base;


synthesized attribute qnames :: [QName];
synthesized attribute ids :: [Id_t];


nonterminal QNames2 with {pp, qnames} ;
concrete production qNames2Two
top::QNames2 ::= id1::QName c::Comma_t id2::QName
{
  top.pp = id1.name ++ ", " ++ id2.name ;
  top.qnames = [id1'', id2''];
}

concrete production qNames2Cons
top::QNames2 ::= id1::QName c::Comma_t id2::QNames2
{
  top.pp = id1.name ++ ", " ++ id2.pp ;
  top.qnames = [id1''] +++ id2.qnames;
}

nonterminal QNames with {pp, qnames} ;
concrete production qNamesSingle
top::QNames ::= id::QName
{
  top.pp = id.pp;
  top.qnames = [id''];
}

concrete production qNamesCons
top::QNames ::= id1::QName c::Comma_t id2::QNames
{
  top.pp = id1.pp ++ ", " ++ id2.pp ;
  top.qnames = [id1''] +++ id2.qnames;
}

nonterminal Ids2 with {pp, ids} ;
concrete production id2Single
top::Ids2 ::= id::Id_t c::Comma_t id2::Id_t
{
  top.pp = id.lexeme ++ ", " ++ id2.lexeme;
  top.ids = [id, id2];
}

concrete production id2Cons
top::Ids2 ::= id1::Id_t c::Comma_t id2::Ids2
{
  top.pp = id1.lexeme ++ ", " ++ id2.pp ;
  top.ids = [id1] +++ id2.ids;
}

nonterminal Ids with {pp, ids} ;
concrete production idSingle
top::Ids ::= id::Id_t
{
  top.pp = id.lexeme;
  top.ids = [id];
}

concrete production idCons
top::Ids ::= id1::Id_t c::Comma_t id2::Ids
{
  top.pp = id1.lexeme ++ ", " ++ id2.pp ;
  top.ids = [id1] +++ id2.ids;
}

function qualifyNames
[QName] ::= i::[Id_t]
{
  return if null(i) then [] else [qNameId(head(i))] +++ qualifyNames(tail(i));
}

function makeOccursDcls
AGDcl ::= l::Integer c::Integer ats::[QName] nts::[QName]
{
  return if null(ats) 
	 then agDclNone()
	 else agDclSeq(makeOccursDclsHelp(l, c, head(ats), nts), makeOccursDcls(l, c, tail(ats), nts));
}

function makeOccursDclsHelp
AGDcl ::= l::Integer c::Integer at::QName nts::[QName]
{
  return if null(nts) 
	 then agDclNone()
	 else agDclSeq(attributionDcl(attr_kwd, at'', occurs_kwd, on_kwd, head(nts), semi_t),
		       makeOccursDclsHelp(l, c, at'', tail(nts)));

  local attribute attr_kwd :: Attribute_kwd ;
  attr_kwd = terminal ( Attribute_kwd, "attribute", l, c);

  local attribute occurs_kwd :: Occurs_kwd ;
  occurs_kwd = terminal ( Occurs_kwd, "occurs", l, c);
 
  local attribute on_kwd :: On_kwd ;
  on_kwd = terminal ( On_kwd, "on", l, c);

  local attribute semi_t :: Semi_t ;
  semi_t = terminal ( Semi_t, ";", l, c);
}

function makeNTDcls
AGDcl ::= l::Integer c::Integer nts::[Id_t]
{
  return if null(nts) 
	 then agDclNone()
	 else agDclSeq(nonterminalDcl(terminal(NonTerminal_kwd, "nontermial", l, c), head(nts), terminal(Semi_t, ";", l, c)),
		       makeNTDcls(l, c, tail(nts)));
}

function makeClosedNTDcls
AGDcl ::= l::Integer c::Integer nts::[Id_t]
{
  return if null(nts) 
	 then agDclNone()
	 else agDclSeq(closedNonterminalDcl(terminal(Closed_kwd, "closed", l, c),
					    terminal(NonTerminal_kwd, "nontermial", l, c),
					    head(nts),
					    terminal(Semi_t, ";", l, c)),
		       makeNTDcls(l, c, tail(nts)));
}

function makeSynDcls
AGDcl ::= l::Integer c::Integer t::Type atts::[Id_t]
{
  return if null(atts)
	 then agDclNone()
	 else agDclSeq(attributeDclSyn(terminal(Synthesized_kwd, "synthesized", l, c),
				       terminal(Attribute_kwd, "attribute", l, c),
			      	       head(atts),
			      	       terminal(HasType_t, "::", l, c),
			               t'',
			               terminal(Semi_t, ";", l, c)),
		       makeSynDcls(l, c, t'', tail(atts)));
}

function makeInhDcls
AGDcl ::= l::Integer c::Integer t::Type atts::[Id_t]
{
  return if null(atts)
	 then agDclNone()
	 else agDclSeq(attributeDclInh(terminal(Inherited_kwd, "inherited", l, c),
				       terminal(Attribute_kwd, "attribute", l, c),
			      	       head(atts),
			      	       terminal(HasType_t, "::", l, c),
			               t'',
			               terminal(Semi_t, ";", l, c)),
		       makeInhDcls(l, c, t'', tail(atts)));
}