grammar silver:extension:functions;
import silver:base;

terminal Function_kwd /function/ lexer precedence = 5;

concrete production qNameFunction
top::QName ::= id::Function_kwd
{
  top.pp = "function";
  top.name = "function";
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";
}
concrete production functionDcl
 top::AGDcl ::=
   f::Function_kwd id::Id_t 
   fs::FunctionSignature b::ProductionBody 
{
  top.location = "[line: " ++ toString(f.line) ++ ", column: " ++ toString(f.column) ++ "] ";

  local attribute ntdcl :: AGDcl;
  ntdcl = nonterminalDcl(terminal(NonTerminal_kwd, "nontreminal", f.line, f.column), terminal(Id_t, "GENERATED__" ++ id.lexeme), terminal(Semi_t, ";"));

  local attribute pdcl :: AGDcl;
  pdcl = aFunctionDcl(id, fs'', b'');		

  local attribute adcl :: AGDcl;
  adcl = attributeDclSyn(
			terminal(Synthesized_kwd, "synthesized", f.line, f.column),
			terminal(Attribute_kwd, "attribute"),
			terminal(Id_t, "generated__" ++ id.lexeme),
			terminal(HasType_t, "::"),
			fs.lhstype,
			terminal(Semi_t, ";")
			);

  local attribute odcl :: AGDcl;
  odcl = attributionDcl(
			terminal(Attribute_kwd, "attribute", f.line, f.column),
			qNameId(terminal(Id_t, "generated__" ++ id.lexeme)),
			terminal(Occurs_kwd, "occurs"),
			terminal(On_kwd, "on"),
			qNameId(terminal(Id_t, "GENERATED__" ++ id.lexeme)),
			terminal(Semi_t, ";")
			);

  forwards to agDclSeq(adcl'', agDclSeq(ntdcl'', agDclSeq(odcl'', pdcl'')));
}

abstract production aFunctionDcl
top::AGDcl ::= id::Id_t fs::FunctionSignature b::ProductionBody
{
  top.location = "[line: " ++ toString(id.line) ++ ", column: " ++ toString(id.column) ++ "] ";

  local attribute f_dcl :: Decorated Defs;
  f_dcl = addValueDcl(top.grammarName.name ++ ":" ++ id.lexeme, functionTypeRep(id.lexeme, tail(ns.sigTypes), head(ns.sigTypes)),
          addFullNameDcl(id.lexeme, top.grammarName.name ++ ":" ++ id.lexeme, emptyDefs()));

  top.defs = f_dcl;

  forwards to productionDcl(
			terminal(Abstract_kwd, "abstract", id.line, id.column),
		        terminal(Production_kwd, "production"),
 		        id,
			ns'',
			b'');

  fs.pname = id.lexeme;

  local attribute ns :: ProductionSignature;
  ns = fs.namedsignature;

  ns.env = top.env;
  ns.grammarName = top.grammarName;
}

synthesized attribute namedlhs :: ProductionLHS;
synthesized attribute namedsignature :: ProductionSignature;
synthesized attribute lhstype :: Type;

inherited attribute pname :: String;

nonterminal FunctionLHS with {namedlhs, lhstype, pname};
concrete production functionLHS
top::FunctionLHS ::= c1::Type
{
  top.lhstype = c1'';
  top.namedlhs = productionLHS(productionLHSElem(
			terminal(Id_t, "generated__top"),
			terminal(HasType_t, "::"), 
			nttType(qNameId(terminal(Id_t, "GENERATED__" ++ top.pname)))));
}


nonterminal FunctionSignature with {namedsignature, lhstype, pname};
concrete production functionSignatureEmptyRHS
top::FunctionSignature ::= c1::FunctionLHS c2::CCEQ_t
{
  top.lhstype = c1.lhstype;
  top.namedsignature = productionSignatureEmptyRHS(c1.namedlhs, c2);

  c1.pname = top.pname;
}

concrete production functionSignature
top::FunctionSignature ::= c1::FunctionLHS c2::CCEQ_t c3::ProductionRHS 
{
  top.lhstype = c1.lhstype;
  top.namedsignature = productionSignature(c1.namedlhs, c2, c3'');

  c1.pname = top.pname;
}

terminal Return_kwd /return/ lexer precedence = 5;
concrete production returnDef
top::AttributeDef ::= c1::Return_kwd c3::Expr c4::Semi_t
{
  local attribute this :: String;
  this = head(getThisDcl(top.env)).itemName;

  local attribute pn :: String;
  pn = substring(length(top.grammarName.name)+1, length(this), this);

  forwards to attrDef(	lhsExprTwo(
				terminal(Id_t, "generated__top", c1.line, c1.column),
				terminal(Dot_t, "."),
				qNameId(terminal(Id_t, "generated__" ++ pn))),
			
			 c3'');
}

synthesized attribute functionName :: String;
attribute functionName occurs on TypeRep;


function functionTypeRep
Decorated TypeRep ::= fn::String it::[Decorated TypeRep] ot::Decorated TypeRep
{
  return decorate internal_functionTypeRep(fn, it, ot) with {}; 
}

abstract production functionApplicationDispatcher
top::Expr ::= c1::Expr c2::Exprs
{
  top.location = c1.location;

  local attribute aname :: String;
  aname = "generated__" ++ c1.typerep.functionName;

  forwards to attributeAccess(genericApplicationDispatcher(c1, c2''), terminal(Dot_t, "."), qNameId(terminal(Id_t, aname)));
}


abstract production internal_functionTypeRep
top::TypeRep ::= fn::String it::[Decorated TypeRep] ot::Decorated TypeRep
{
  top.functionName = fn;
  top.applicationDispatcher = functionApplicationDispatcher;

  forwards to internal_prodTypeRep(it, ot);
}

aspect production internal_defaultTypeRep
top::TypeRep ::= 
{
  top.functionName = "";
}