grammar silver:extension:list;
import silver:base;
import silver:analysis:typechecking;
--import silver:translation;
import silver:translation:ag:haskell ;

import core only bValue;

terminal LSqr_t /[\[]/ lexer precedence = 5 ;
terminal RSqr_t /[\]]/ lexer precedence = 5 ;

-- Type Expressions
-------------------
concrete production listType
top::Type ::= lb::LSqr_t te::Type rb::RSqr_t
{
  top.typerep = listTypeRep(te.typerep);
  top.errors = te.errors;

  forwards to  refType(terminal(Decorated_kwd, ""), qNameId (terminal(Id_t,  "AnyTypeList")));
}

-- List Expressions
-------------------
-- Empty List Literal
---------------------
concrete production emptyList
top::Expr ::= lb::LSqr_t rb::RSqr_t
{
  top.haskell = @"[]";

--  top.typerep = emptyListTypeRep();
  top.typerep = top.resolvedType;
  forwards to emptyProductionApp(baseExpr(qNameId(terminal (Id_t, "nil_AnyTypeList"))),
				terminal(LParen_t, "("),
				terminal(RParen_t, ")"));
}


concrete production emptyListWType
top::Expr ::= lb::LSqr_t ht::HasType_t t::Type rb::RSqr_t
{
  top.haskell = @"[]";

  top.typerep = listTypeRep(t.typerep);
  forwards to emptyProductionApp(baseExpr(qNameId(terminal (Id_t, "nil_AnyTypeList"))),
				terminal(LParen_t, "("),
				terminal(RParen_t, ")"));
}

-- List Literal
---------------
concrete production fullList
top::Expr ::= lb::LSqr_t es::Exprs rb::RSqr_t
{ 
  top.haskell = @"[" ++ es.listHaskell ++ @"]";

  top.pp = "[ " ++ es.pp ++ " ]" ;
  top.location = "[line: " ++ toString(lb.line) ++ ", column: " ++ toString(lb.column) ++ "] ";

--  local attribute newExprs :: Exprs;
--  newExprs = buildExprs(e.typerep.inputTypes, es.exprList);
  
  top.typeErrors = es.typeErrors ++
	         if checkListErrors(head(es.inputTypes), es.inputTypes)
	         then ""
		 else top.location ++ "Error: Elements of a list must all be the same type.\n";

  local attribute t :: Decorated TypeRep;
  t = head(es.inputTypes);

  top.typerep = if null(es.inputTypes) then emptyListTypeRep() else listTypeRep(t);

  es.expectedInputTypes = makeListInputs(fakeListTypeRep(top.userFriendly), length(es.inputTypes));

  forwards to es.listtrans ;
}
function makeListInputs
[Decorated TypeRep] ::=  t::Decorated TypeRep c :: Integer
{
  return if c <= 0 then [] else [t] ++ makeListInputs(t, c-1);
}

function checkListErrors
Boolean ::=  t::Decorated TypeRep ts::[Decorated TypeRep]
{
  return null(ts) || (t.typeEquals(t, head(ts)).bValue && checkListErrors(t, tail(ts)));
}

synthesized attribute listtrans :: Expr ;
attribute listtrans occurs on Exprs ;

synthesized attribute listHaskell :: StringAccum;
attribute listHaskell occurs on Exprs;
aspect production exprsEmpty
top::Exprs ::=
{
  top.listHaskell = @"";
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.listHaskell = casted.haskell;

  local attribute casted :: Expr;
  casted = cast_t(terminal(Cast_kwd, ""), e'', anyTypeTypeRep());
  casted.env = top.env;
  casted.signatureEnv = top.signatureEnv;
  casted.localsEnv = top.localsEnv;
  casted.globalEnv = top.globalEnv;
  casted.resolvedType = top.resolvedType;
  casted.userFriendly = if null(top.expectedInputTypes) then 0 else head(top.expectedInputTypes).userFriendlyLHS;

  top.listtrans = 
   productionApp(baseExpr(qNameId( terminal(Id_t, "cons_AnyTypeList" ))),
	          terminal(LParen_t, "("),
                    exprsCons (casted'', terminal(Comma_t, ","),
                               exprsSingle(emptyProductionApp(baseExpr( qNameId (terminal (Id_t, "nil_AnyTypeList"))), 
		 	                                      terminal(LParen_t, "("),
				 	                      terminal(RParen_t, ")")))
                              ),
	          terminal(RParen_t, ")")
                  );
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.listHaskell = casted.haskell ++ @", " ++ e2.listHaskell;

  local attribute casted :: Expr;
  casted = cast_t(terminal(Cast_kwd, ""), e1'', anyTypeTypeRep());
  casted.env = top.env;
  casted.signatureEnv = top.signatureEnv;
  casted.localsEnv = top.localsEnv;
  casted.globalEnv = top.globalEnv;
  casted.resolvedType = top.resolvedType;
  casted.userFriendly = if null(top.expectedInputTypes) then 0 else head(top.expectedInputTypes).userFriendlyLHS;

  top.listtrans = 
   productionApp(baseExpr(qNameId(terminal(Id_t, "cons_AnyTypeList"))),
                   terminal(LParen_t, "("),                   
                   exprsCons (casted'' ,  terminal(Comma_t,","), exprsSingle( e2.listtrans )),
                   terminal(RParen_t, ")")
                  ) ;
}

function makeNew
Expr ::= e::Expr
{
  return newFunction(terminal(New_kwd, ""), terminal(LParen_t, ""), e'', terminal(RParen_t, ""));
}

-- Cons
-------
terminal Cons_t / cons /  lexer precedence = 5 ;

concrete production consList
top::Expr ::= ck::Cons_t lp::LParen_t h::Expr c::Comma_t t::Expr rp::RParen_t 
{ 

  top.haskell = @"(" ++ casted.haskell ++ @" : " ++ t.haskell ++ @")";

  top.pp = "cons (" ++ h.pp ++ ", " ++ t.pp ++ ")" ;
  top.location = "[line: " ++ toString(ck.line) ++ ", column: " ++ toString(ck.column) ++ "] ";
 
  top.typerep = listTypeRep(h.typerep);

  local attribute e1 :: String;
  e1 = if t.typerep.isList then ""
       else top.location ++ "Error: Second argument to cons must be a List.\n";

  local attribute e2 :: String;
  e2 = if t.typerep.isList && !t.typerep.isEmpty && !(h.typerep.typeEquals(h.typerep, t.typerep.listComponent).bValue)
       then top.location ++ "Error: First argument to cons must be of type: " ++ t.typerep.listComponent.typeName ++ ".\n"
       else "";

  top.typeErrors = h.typeErrors ++ t.typeErrors ++ e1 ++ e2;

  local attribute casted :: Expr;
  casted = cast_t(terminal(Cast_kwd, ""), h'', anyTypeTypeRep());
  casted.env = top.env;
  casted.signatureEnv = top.signatureEnv;
  casted.localsEnv = top.localsEnv;
  casted.globalEnv = top.globalEnv;
  casted.resolvedType = top.resolvedType;
  casted.userFriendly = top.userFriendly;

  forwards to
    productionApp(baseExpr(qNameId(terminal(Id_t, "cons_AnyTypeList"))),
                  terminal(LParen_t, "("),
                  exprsCons(casted'', terminal(Comma_t,","), exprsSingle ( t'' )),
                  terminal(RParen_t, ")")
                  );
}


terminal ConsOp_t / [\:][\:][\:] /   
  lexer precedence = 5,
  precedence = 20,
  association = right;

concrete production consListOp
top::Expr ::= h::Expr o::ConsOp_t t::Expr
{ 
  top.haskell = @"(" ++ casted.haskell ++ @" : " ++ t.haskell ++ @")";
  top.pp = "(" ++ h.pp ++ " ::: " ++ t.pp ++ ")" ;

  top.typerep = listTypeRep(h.typerep); 

  local attribute casted :: Expr;
  casted = cast_t(terminal(Cast_kwd, ""), h'', anyTypeTypeRep());
  casted.env = top.env;
  casted.signatureEnv = top.signatureEnv;
  casted.localsEnv = top.localsEnv;
  casted.globalEnv = top.globalEnv;
  casted.resolvedType = top.resolvedType;
  casted.userFriendly = top.userFriendly;

  forwards to  consList(terminal(Cons_t, "", o.line, o.column),
			terminal(LParen_t, ""),
			h'',
			terminal(Comma_t, ""),
			t'',
			terminal(RParen_t, ""));
}


aspect production plusPlus
top::Expr ::= e1::Expr p::PlusPlus_t e2::Expr
{
  handler <- if e1.typerep.isList && e2.typerep.isList 
	     then [appendList(e1, terminal(AppendOp_t, "+++", p.line, p.column), e2)]
	     else [];
}
-- Append
---------
terminal AppendOp_t / [\+][\+][\+] /   
  lexer precedence = 5,
  precedence = 19,
  association = left;


concrete production appendList_c
top::Expr ::= l::Expr o::AppendOp_t r::Expr
{ 
  forwards to appendList(l, o, r);
}


abstract production appendList
top::Expr ::= l::Decorated Expr o::AppendOp_t r::Decorated Expr
{ 
  top.haskell = @"(" ++ l.haskell ++ @" ++ " ++ r.haskell ++ @")";

  top.pp = l.pp ++ " " ++ o.lexeme ++ " " ++ r.pp ;
  top.location = "[line: " ++ toString(o.line) ++ ", column: " ++ toString(o.column) ++ "] ";

  top.typerep = if l.typerep.isEmpty then r.typerep else l.typerep;

  local attribute e1 :: String;
  e1 = if l.typerep.isList then ""
       else top.location ++ "Error: First argument to append must be a List.\n";

  local attribute e2 :: String;
  e2 = if r.typerep.isList then ""
       else top.location ++ "Error: Second argument to append must be a List.\n";

  local attribute e3 :: String;
  e3 = if    l.typerep.isList 
	  && !l.typerep.isEmpty 
	  && r.typerep.isList 
	  && !r.typerep.isEmpty 
	  && !(l.typerep.typeEquals(l.typerep, r.typerep).bValue)
       then top.location ++ "Error: List types differ in application of append.\n"
       else "";

  top.typeErrors = l.typeErrors ++ r.typeErrors ++ e1 ++ e2 ++ e3;

  forwards to
   productionApp(baseExpr(qNameId(terminal(Id_t, "append_AnyTypeList"))),
                 terminal(LParen_t, "("),
                 exprsCons(new(l), terminal(Comma_t,","), exprsSingle ( new(r) ) ), 
                 terminal(RParen_t, ")")                 
	        );

}

concrete production bOperatorAppend
top::BOperator ::= t::AppendOp_t
{
  top.pp = t.lexeme;
  top.binaryProduction = plusPlusWrapper;
}

--abstract production appendWrapper
--top::Expr ::= e1::Expr e2::Expr
--{
--  forwards to appendList(e1'', terminal(AppendOp_t, "+++"), e2'');
--}


aspect production lengthFunction
top::Expr ::= l::Length_kwd lp::LParen_t e::Expr rp::RParen_t
{
  handlers <- if e.typerep.isList
	      then [listLength(e)]
	      else [];
}

-- Length
---------
abstract production listLength
top::Expr ::= e::Decorated Expr
{
  top.haskell = @"(length " ++ e.haskell ++ @")";
  top.pp = "length ( " ++ e.pp ++ " )";
  top.location = e.location;

  top.typerep = integerTypeRep();
 
  top.errors = e.errors;
  top.typeErrors = e.typeErrors;

  forwards to 
   attributeAccess ( new(e) , terminal(Dot_t,"."), 
                     qNameId (terminal (Id_t, "length_AnyTypeList") )
                   );
}

-- Empty
--------
terminal Null_t / null /  lexer precedence = 5 ;

concrete production nullList
top::Expr ::= n::Null_t lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.haskell = @"(null " ++ l.haskell ++ @")";
  top.pp = n.lexeme ++ " (" ++ l.pp ++ ")" ;

  top.location = "[line: " ++ toString(n.line) ++ ", column: " ++ toString(n.column) ++ "] ";

  local attribute e1 :: String;
  e1 = if l.typerep.isList then ""
       else top.location ++ "Error: Argument to null must be a List.\n";

  top.typeErrors = l.typeErrors ++ e1;

  forwards to
   attributeAccess ( l'' , terminal(Dot_t,"."), 
                     qNameId (terminal (Id_t, "empty_AnyTypeList") )
                   );
}



-- Head
--------
terminal Head_t / head /  lexer precedence = 5 ;

concrete production headList
top::Expr ::= h::Head_t lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.location = "[line: " ++ toString(h.line) ++ ", column: " ++ toString(h.column) ++ "] ";

  forwards to
   cast_t(terminal(Cast_kwd, ""), headListNoCast(l''), l.typerep.listComponent);
}

abstract production headListNoCast
top::Expr ::= l::Expr 
{ 
  top.haskell = @"(head " ++ l.haskell ++ @")";
  top.pp = "head(" ++ l.pp ++ ")" ;

  local attribute e1 :: String;
  e1 = if l.typerep.isList then ""
       else top.location ++ "Error: Argument to head must be a List.\n";

  top.typeErrors = l.typeErrors ++ e1;

  forwards to
   attributeAccess ( l'', terminal(Dot_t,"."), qNameId (terminal (Id_t, "head_AnyTypeList")));
	
}


-- Tail
--------
terminal Tail_t / tail /  lexer precedence = 5 ;

concrete production tailList
top::Expr ::= t::Tail_t lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.haskell = @"(tail " ++ l.haskell ++ @")";

  top.pp =  "tail (" ++ l.pp ++ ")" ;
  top.typerep = l.typerep;

  top.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";

  local attribute e1 :: String;
  e1 = if l.typerep.isList then ""
       else top.location ++ "Error: Argument to tail must be a List.\n";

  top.typeErrors = l.typeErrors ++ e1;

  forwards to
   attributeAccess ( l'' , terminal(Dot_t,"."), 
                     qNameId (terminal (Id_t, "tail_AnyTypeList") )
                   );
}

function makeListName
String ::= str::String
{
  local attribute i :: Integer;
  i = indexof( ":", str);

  return  if i == -1
          then str
          else substring(0, i, str) ++ "_" ++ 
               makeListName(substring(i+1, length(str), str));
}


synthesized attribute listComponent :: Decorated TypeRep ;
attribute listComponent occurs on TypeRep ;

synthesized attribute isList :: Boolean;
attribute isList occurs on TypeRep ;
attribute isEmpty occurs on TypeRep ;


aspect production mkHaskell_datatype_and_unwrap_code
top::StringNT ::= cl::[GrammarSpec]
{
  allTerminalsNonTerminalsAndRefsToNTs <- [emptyListTypeRep()];
}


function listTypeRep
Decorated TypeRep ::= tr::Decorated TypeRep
{
  return decorate internal_listTypeRep(tr.userFriendlyLHS, false, tr) with {};
}

function emptyListTypeRep
Decorated TypeRep ::= 
{
  return decorate internal_listTypeRep(0, true, topTypeRep()) with {};
}

function fakeListTypeRep
Decorated TypeRep ::= uf::Integer
{
  return decorate internal_listTypeRep(uf, true, topTypeRep()) with {};
}

abstract production internal_listTypeRep
top::TypeRep ::= uf::Integer e::Boolean tr::Decorated TypeRep
{
  top.type_fullname = "LIST";

  top.haskellType = "[Haskell_AnyType]";

  top.pp = if e then "[]" else "[" ++ tr.pp ++ "]";
  top.isList = true;
  top.listComponent = tr;
  top.doDecorate = false;
  top.isEmpty = e;
  top.typeEquals = listTypeEquals;
  top.decoratedType = decorate internal_listTypeRep(uf,e,tr) with { }  ;
  top.isDecorated = false ;
  top.userFriendlyLHS = uf;

  forwards to internal_refTypeRep(ntTypeRep("lists:anytype:AnyTypeList"));
}

aspect production internal_defaultTypeRep
top::TypeRep ::= 
{
  top.isList = false;
  top.isEmpty = true;
  top.listComponent = top;
}

aspect production internal_topTypeRep
top::TypeRep ::= 
{
  top.isList = true;
  top.isEmpty = true;
  top.listComponent = top;
}
abstract production listTypeEquals
top::TypeEquals ::= t1::Decorated TypeRep t2::Decorated TypeRep
{
    top.bValue =   t1.isList 
		&& t2.isList 
		&& t1.listComponent.typeEquals(t1.listComponent, t2.listComponent).bValue;
}
