grammar silver:extension:patternmatching;
import core ;
import silver:base;
import silver:analysis:typechecking ;
--import silver:translation:haskell ;
import silver:extension:let ;
import silver:extension:list ;
--import silver:base:type:anytype ;
import silver:extension:polymorphism;

terminal Case_kwd /case/ lexer precedence = 5 ;
terminal Of_kwd /of/ lexer precedence = 5 ;
terminal arrow_kwd /=>/ lexer precedence = 7 ;
-- terminal wildc_kwd /\_/ lexer precedence = 3 ;
terminal vbar_kwd /\|/ lexer precedence = 3 ;

inherited attribute typerep_down :: Decorated TypeRep ;
inherited attribute typereps_down :: [ Decorated TypeRep ] ;
inherited attribute base_tree :: Expr ;

autocopy attribute v2_version :: Boolean ;
attribute v2_version occurs on MRuleList, MatchRule, Pattern, PatternList ;

inherited attribute v2_typerep_down :: Decorated TypeRep ;
inherited attribute v2_typereps_down :: [ Decorated TypeRep ] ;
synthesized attribute v2_defs :: Decorated Defs ;
inherited attribute v2_base_tree :: Expr ;
inherited attribute v2_case_expr_type :: Decorated TypeRep ;

synthesized attribute translation_tree :: Expr ;
synthesized attribute cond_tree :: Expr ;
synthesized attribute then_tree :: Expr ;
synthesized attribute letAssigns_tree :: [ AssignExpr ] ;
synthesized attribute fExpr :: Expr ;

nonterminal MRuleList with {pp, grammarName, env, localsEnv, signatureEnv, typerep, 
			    errors, typerep_down, typeErrors, location, --haskell,
			    translation_tree, base_tree, 
                            v2_typerep_down, v2_case_expr_type } ;
nonterminal MatchRule with {pp, grammarName, env, localsEnv, signatureEnv, typerep, 
                            errors, typerep_down, typeErrors, location, --haskell,
			    cond_tree, then_tree, base_tree, 
                            v2_typerep_down, v2_case_expr_type } ;
nonterminal Pattern with {pp, env, localsEnv, signatureEnv, location, errors, defs,
			    typerep_down, typeErrors, cond_tree, --haskell
                            letAssigns_tree, base_tree, 
                            v2_typerep_down, v2_defs, v2_case_expr_type } ;
nonterminal PatternList with {pp, env, localsEnv, signatureEnv, errors, defs, 
			    typereps_down, typeErrors, location, --haskell, 
			    cond_tree, letAssigns_tree, base_tree, 
                            v2_typereps_down, v2_defs, v2_case_expr_type } ;


concrete production caseExpr_c
e::Expr ::= c::Case_kwd e1::Expr o::Of_kwd ml::MRuleList ek::End_kwd
{ 
  e.pp = "case " ++ e1.pp ++ " of " ++ ml.pp ;  
  e.location = "[line: " ++ toString(c.line) ++ ", column: " ++ toString(c.column) ++ "] ";
  e.errors = e1.errors ++ ml.errors ;  

  -- environments
  e1.grammarName = e.grammarName ;
  ml.grammarName = e.grammarName ;
  e1.env = e.env ;
  ml.env = e.env ;
  e1.localsEnv = e.localsEnv ;
  ml.localsEnv = e.localsEnv ;
  e1.signatureEnv = e.signatureEnv ;
  ml.signatureEnv = e.signatureEnv ;

  -- type checking
  e.typerep = ml.typerep ; 


  -- Change this attribute to 'false' to revert to previous version of pattern matching
  -------------------------------------------------------------------------------------
  local attribute v2_version_2_flag :: Boolean ;
  v2_version_2_flag = true ;

  -- If the case expr type is decorated, then pass down the undecorated type.  This
  -- undecorated version of the type will match the output type of any appropriate
  -- productions at the top level of any patterns.
  -- (This is the same for versions 1 and 2.)
  ml.typerep_down = if e1.typerep.isDecorated
 		    then e1.typerep.decoratedType
		    else if e1.typerep.isNonTerminal || e1.typerep.isConstructedType
		    then e1.typerep
		    else topTypeRep() ;


  ml.v2_version = v2_version_2_flag ;
  ml.v2_case_expr_type = e1.typerep ;

  e.typeErrors = if e1.typerep.isDecorated || e1.typerep.isNonTerminal || e1.typerep.isConstructedType
		 then e1.typeErrors ++ ml.typeErrors 
		 else e.location ++ "Error: only nonterminal typed expressions can be pattern matched.\n"
			++ e1.typeErrors ++ ml.typeErrors ;

  -- translation
  ml.base_tree = if e1.typerep.isDecorated 
                 then e1'' 
                 else exprWith(terminal(Decorate_kwd, "decorate", c.line, c.column),
 	     	               e1'', 
	     	               terminal(With_kwd, "with", c.line, c.column),
	     	               terminal(LCurly_t, "{", c.line, c.column),
	                       exprStmtsEmpty(),
		               terminal(RCurly_t, "}", c.line, c.column));

  -- base_tree is always the decorated version of the tree

  e1.userFriendly = 0 ;
  forwards to ml.translation_tree ;
}

concrete production mRuleList_one
ml::MRuleList ::= m::MatchRule
{
  ml.pp = m.pp ;
  ml.errors = m.errors ;
  ml.location = m.location ;

  -- environments
  m.grammarName = ml.grammarName ;
  m.env = ml.env ;
  m.localsEnv = ml.localsEnv ;
  m.signatureEnv = ml.signatureEnv ;

  -- type checking
  ml.typerep = m.typerep;
  m.typerep_down = ml.typerep_down ;

  m.v2_case_expr_type = ml.v2_case_expr_type ;

  ml.typeErrors = m.typeErrors ;

  -- translation
  m.base_tree = ml.base_tree ;
  ml.translation_tree = ifThenElse(terminal(If_kwd, "if"),
			    	   m.cond_tree,
				   terminal(Then_kwd, "then"),
				   m.then_tree,
				   terminal(Else_kwd, "else"),
				   errorFunction(terminal(Error_kwd, "error"),
						 terminal(LParen_t, "("),
						 stringConst(terminal(String_t, "\"Error: pattern match failed.\\n\"")),
						 terminal(RParen_t, ")"))) ;
}

concrete production mRuleList_cons
ml::MRuleList ::= h::MatchRule bar::vbar_kwd t::MRuleList
{
  ml.pp = h.pp ++ " | " ++ t.pp ;
  ml.errors = h.errors ++ t.errors ;
  ml.location = h.location ;  

  -- environments
  h.grammarName = ml.grammarName ;
  t.grammarName = ml.grammarName ;
  h.env = ml.env ;
  t.env = ml.env ;
  h.localsEnv = ml.localsEnv ;
  t.localsEnv = ml.localsEnv ;
  h.signatureEnv = ml.signatureEnv ;
  t.signatureEnv = ml.signatureEnv ;

  -- type checking
  ml.typerep = h.typerep;
  h.typerep_down = ml.typerep_down ;
  t.typerep_down = ml.typerep_down ;

  h.v2_case_expr_type = ml.v2_case_expr_type ;
  t.v2_case_expr_type = ml.v2_case_expr_type ;

  local attribute tr :: String ;
  tr = if h.typerep.typeEquals(h.typerep, t.typerep).bValue
       then "" 
       else h.location ++ "Error: rules of CASE expression do not have the same type.\n" ;  
 
  ml.typeErrors = h.typeErrors ++ t.typeErrors ++ tr ;

  -- translation
  h.base_tree = ml.base_tree ;
  t.base_tree = ml.base_tree ;

  ml.translation_tree = ifThenElse(terminal(If_kwd, "if"),
 				   h.cond_tree,
				   terminal(Then_kwd, "then"),
				   h.then_tree,
 				   terminal(Else_kwd, "else"),
				   t.translation_tree) ;
}

concrete production matchRule
mr::MatchRule ::= pt::Pattern a::arrow_kwd e::Expr
{
  mr.pp = pt.pp ++ " => " ++ e.pp ;
  mr.errors = pt.errors ++ e.errors ;
  mr.location = e.location ;  

  -- environments
  e.grammarName = mr.grammarName ;

  e.env = appendDefsEnv(pt.defs, mr.env) ;

  e.localsEnv = mr.localsEnv ;
  e.signatureEnv = mr.signatureEnv ;

  pt.env = mr.env ;
  pt.localsEnv = mr.localsEnv ;
  pt.signatureEnv = mr.signatureEnv ;

  -- type checking
  e.userFriendly = 0 ; 
  mr.typerep = e.typerep ;
  pt.typerep_down = mr.typerep_down ;

  pt.v2_case_expr_type = mr.v2_case_expr_type ;

  mr.typeErrors = pt.typeErrors ++ e.typeErrors ;

  -- translation
  mr.cond_tree = pt.cond_tree ;
  mr.then_tree = if null(pt.letAssigns_tree) then e''
		 else letp(terminal(Let_kwd, "let"), 
                           toAssigns(pt.letAssigns_tree),
       		           terminal(In_kwd, "in"),
                           e'',
                           terminal(End_kwd, "end")) ;

  pt.base_tree = mr.base_tree;
}


concrete production patternList_one
ps::PatternList ::= p::Pattern
{
  ps.pp = p.pp ;
  ps.errors = p.errors ;
  ps.defs = p.defs ;
  ps.location = p.location ;

  -- environments
  p.env = ps.env ;
  p.localsEnv = ps.localsEnv ;
  p.signatureEnv = ps.signatureEnv ;

  -- type checking
  p.typerep_down = if ! null(ps.typereps_down) 
                   then head(ps.typereps_down) 
                   else topTypeRep() ; 

  p.v2_case_expr_type = ps.v2_case_expr_type ;
  

  local attribute tr :: String ;
  tr = if null(ps.typereps_down)
       then ps.location ++ "Error: production call in pattern has too many arguments.\n" 
       else if length(ps.typereps_down) > 1
	    then ps.location ++ "Error: production call in pattern has too few arguments. \n"
	    else "";

  ps.typeErrors = tr ++ p.typeErrors ;

  -- The type of the element in the child list is the decorated version of the rhs children.
  local attribute cType :: Decorated TypeRep;
  cType = if head(ps.typereps_down).isDecorated || !head(ps.typereps_down).isNonTerminal 
          then head(ps.typereps_down)
          else if head(ps.typereps_down).isConstructedType
	  then refTypeRep(toMonoTypeRep(head(ps.typereps_down)))
	  else refTypeRep(head(ps.typereps_down)) ;

  local attribute orig_base_tree :: Expr ;
  orig_base_tree = if cType.isTypeVar
	then headList(terminal(Head_t, "head"), terminal(LParen_t, "("), ps.base_tree, terminal(RParen_t, ")")) 
	else cast_t(terminal(Cast_kwd, ""), 
                    headList(terminal(Head_t, "head"), terminal(LParen_t, "("), ps.base_tree, terminal(RParen_t, ")")), 
                    cType);
--DB
  p.base_tree = --if ps.v2_version
                --then v2_base_tree2 
                --else orig_base_tree ;
	orig_base_tree'';

  -- base_tree need to wrap a "new" around the cast of the head of the child list.
  -- It does this if: 
  --  case expression is undecorated NT  and  typerep_down  is undecorated NT
  -- if case expression is decorated, then just get the decorated construct out of the list
  -- if typerep_down is decorated, then, just get the decorated construct out of the list

  local attribute v2_base_tree2 :: Expr ; 
  v2_base_tree2 = if p.v2_case_expr_type.isNonTerminal && p.typerep_down.isNonTerminal
                  then newFunction( terminal(New_kwd,"new"), terminal(LParen_t,"("), orig_base_tree'', terminal(RParen_t,")") )
                  else orig_base_tree'' ;

  ps.cond_tree = p.cond_tree ;
  ps.letAssigns_tree = p.letAssigns_tree ;

--  -- previous translation
--  p.base_tree = cast_t(terminal(Cast_kwd, ""), 
--		       attributeAccess (ps.base_tree , 
--					terminal(Dot_t,"."), 
--                           		qNameId(terminal(Id_t, "head_AnyTypeList"))), cType);
}

concrete production patternList_more
ps::PatternList ::= p::Pattern c::Comma_t ps1::PatternList
{
  ps.pp = ps1.pp ++ ", " ++ p.pp ;
  ps.errors = ps1.errors ++ p.errors ;
  ps.defs = appendDefs(ps1.defs, p.defs) ;

  ps.location = p.location ;

  -- environments
  p.env = ps.env ;
  p.localsEnv = ps.localsEnv ;
  p.signatureEnv = ps.signatureEnv ;

  ps1.env = appendDefsEnv(p.defs, ps.env);

  ps1.localsEnv = ps.localsEnv ;
  ps1.signatureEnv = ps.signatureEnv ;

  -- type checking
  p.typerep_down = if ! null(ps.typereps_down)
                   then head(ps.typereps_down) 
                   else topTypeRep();
  ps1.typereps_down = if ! null(ps.typereps_down)
                      then tail(ps.typereps_down) 
                      else [] ;

  p.v2_case_expr_type = ps.v2_case_expr_type ;
  ps1.v2_case_expr_type = ps.v2_case_expr_type ;


  local attribute tr :: String ;
  tr = if null(ps.typereps_down)
       then ps.location ++ "Error: production call in pattern has too many arguments.\n"
       else "";

  ps.typeErrors = tr ++ p.typeErrors ++ ps1.typeErrors ;


  local attribute cType :: Decorated TypeRep;
  cType = if head(ps.typereps_down).isDecorated || !head(ps.typereps_down).isNonTerminal 
          then head(ps.typereps_down)
          else if head(ps.typereps_down).isConstructedType
	  then refTypeRep(toMonoTypeRep(head(ps.typereps_down)))
	  else refTypeRep(head(ps.typereps_down)) ;

  local attribute orig_base_tree :: Expr ;
  orig_base_tree = if cType.isTypeVar
		   then headList(terminal(Head_t, "head"), terminal(LParen_t, "("), ps.base_tree, terminal(RParen_t, ")")) 
		   else cast_t(terminal(Cast_kwd, ""), 
                          headList(terminal(Head_t, "head"), terminal(LParen_t, "("), ps.base_tree, terminal(RParen_t, ")")), 
                          cType);
--DB
  p.base_tree = --if ps.v2_version
--                then v2_base_tree2 
--                else orig_base_tree ;
orig_base_tree'' ;

  -- base_tree need to wrap a "new" around the cast of the head of the child list.
  -- It does this if: 
  --  case expression is undecorated NT  and  typerep_down  is undecorated NT
  -- if case expression is decorated, then just get the decorated construct out of the list
  -- if typerep_down is decorated, then, just get the decorated construct out of the list

  local attribute v2_base_tree2 :: Expr ; 
  v2_base_tree2 = if p.v2_case_expr_type.isNonTerminal && p.typerep_down.isNonTerminal
                  then newFunction( terminal(New_kwd,"new"), terminal(LParen_t,"("), orig_base_tree'', terminal(RParen_t,")") )
                  else orig_base_tree'' ;


--  -- previous translation
--  p.base_tree = cast_t(terminal(Cast_kwd, ""), 
--		       attributeAccess (ps.base_tree , 
--					terminal(Dot_t,"."), 
--                           		qNameId(terminal(Id_t, "head_AnyTypeList"))), cType) ;


  ps1.base_tree = tailList(terminal(Tail_t, "tail"), terminal(LParen_t, "("), ps.base_tree, terminal(RParen_t, ")"));
--  ps1.base_tree = attributeAccess(ps.base_tree , 
--				  terminal(Dot_t,"."), 
--                     		  qNameId(terminal(Id_t, "tail_AnyTypeList")));

  ps.cond_tree = and(p.cond_tree, terminal(And_t, "&&"), ps1.cond_tree) ;
  ps.letAssigns_tree = p.letAssigns_tree +++ ps1.letAssigns_tree ;

}


concrete production PatternList_nil
ps::PatternList ::= 
{
  ps.pp = "" ;
  ps.errors = "" ;
  ps.defs = emptyDefs() ; 
  ps.location = "" ;

  -- type checking
  ps.typeErrors = if null(ps.typereps_down) then "" 
                  else ps.location ++ "Error: expecting more arguments for production call pattern.\n" ;

  -- translation
  ps.cond_tree = trueConst(terminal(True_kwd, "true")) ;
  ps.letAssigns_tree = [] ; 
}

concrete production prodAppPattern
p::Pattern ::= prod::QName l::LParen_t ps::PatternList r::RParen_t
{
  p.pp = prod.pp ++ "(" ++ ps.pp ++ ")" ;
  p.location = "[line: " ++ toString(l.line) ++ ", column: " ++ toString(l.column) ++ "] ";
  p.errors = er1 ++ ps.errors ; 
  p.defs = ps.defs ;

  ps.env = p.env ;
  ps.localsEnv = p.env ;
  ps.signatureEnv = p.signatureEnv ;


  -- type checking
  local attribute n :: [Decorated EnvItem] ;
  n = getFullNameDcl(prod.name, p.env);
  
  local attribute val :: [Decorated EnvItem] ;
  val = getValueDcl(head(n).fullName, p.env) ;

  local attribute er1 :: String ;
  er1 = if length(n) == 0 || length(val) == 0
        then prod.location ++ "Error: '" ++ prod.name ++ "' is not declared.\n" 
        else "";


  local attribute pType :: Decorated TypeRep ;
  pType = if length(n) > 0 && length(val) > 0 
             then head(val).typerep
             else topTypeRep() ; 

  local attribute er2 :: String ;
  er2 = if p.typerep_down.isConstructedType 
	then ( if type.isProduction 
	       then (if type.outputType.typeEquals(type.outputType, p.typerep_down).bValue
	             then ""
	             else prod.location ++ "Error: production '" ++ prod.name ++ "' has incorrect output type.\n" ++ 
		       "(pattern) " ++ type.outputType.pp ++ " <-> " ++ "(expected) " ++ 
		       p.typerep_down.pp ++ "\n")
	       else prod.location ++ "Error: production '" ++ prod.name ++ "' is not polymorphic" ++
		       " or has incorrect arity.\n")
        else if pType.isProduction 
        then (if pType.outputType.typeEquals(pType.outputType, p.typerep_down).bValue
              then  ""
              else prod.location ++ "Error: production '" ++ prod.name ++ "' has incorrect output type.\n" ++ 
		   "(pattern) " ++ pType.outputType.pp ++ " <-> " ++ "(expected) " ++ 
		   p.typerep_down.pp ++ "\n")
        else prod.location ++ "Error: '" ++ prod.name ++ "' is not a production name.\n" ;


  -- type checking for polymorphic types
  local attribute val1 :: [Decorated EnvItem];
  val1 = getPlmProdDcl(head(n).fullName, p.env);
 

  local attribute pType1 :: Decorated TypeRep;
  pType1 = if !null(n) && !null(val1)
	   then head(val1).typerep
	   else topTypeRep();

  local attribute type :: Decorated TypeRep;
  type = if pType1.isQuantifiedType && head(val1).arity == length(p.typerep_down.constructArgs)
	 then typeApplication(pType1, p.typerep_down.constructArgs)
	 else topTypeRep();

  ps.typereps_down  =  if p.typerep_down.isConstructedType && type.isProduction
		       then type.inputTypes 
		       else if pType.isProduction 
		       then pType.inputTypes 
		       else [] ;
  
  ps.v2_case_expr_type = p.v2_case_expr_type ;

  p.typeErrors = ps.typeErrors ++ er2 ;
 
  -- er3 for debugging
  local attribute er3 :: String;
  er3 = prod.location ++ "Production " ++ prod.name ++ " has input types\n  " ++ 
	printTypes(type.inputTypes) ++ "\nOutput type\n  " ++
	printTypes([type.outputType]) ++ "\n\n"; 

  -- translation
  ps.base_tree = attributeAccess(p.base_tree, 
				 terminal(Dot_t, "."),
				 qNameId(terminal(Id_t, "patChildList"))) ;
 
  p.cond_tree = and(eqeq(attributeAccess(p.base_tree, 
				         terminal(Dot_t, "."), 
				         qNameId(terminal(Id_t, "patProdName"))),
		         terminal(EQEQ_t, "=="),
      		         stringConst(terminal(String_t, "\"" ++ prod.name ++ "\""))),
		    terminal(And_t, "&&"),
		    ps.cond_tree) ;

  p.letAssigns_tree = ps.letAssigns_tree ;		     
} 

concrete production intPattern
p::Pattern ::= num::Int_t
{
  p.pp = num.lexeme ;
  p.location = "[line: " ++ toString(num.line) ++ ", column: " ++ toString(num.column) ++ "] ";
  p.errors = "" ;

  p.defs = emptyDefs();

  p.typeErrors = if p.typerep_down.typeEquals(p.typerep_down, integerTypeRep()).bValue
                 then ""
                 else p.location ++ "Error: expecting '" ++ p.typerep_down.typeName ++ "' type.\n" ;

  -- translation
  p.cond_tree = eqeq(p.base_tree, terminal(EQEQ_t, "=="), intConst(num)) ;
  p.letAssigns_tree = [] ; 
}

concrete production strPattern
p::Pattern ::= str::String_t
{
  p.pp = str.lexeme ;
  p.location = "[line: " ++ toString(str.line) ++ ", column: " ++ toString(str.column) ++ "] ";
  p.errors = "" ;

  p.defs = emptyDefs();

  p.typeErrors = if p.typerep_down.typeEquals(p.typerep_down, stringTypeRep()).bValue
                 then ""
                 else p.location ++ "Error: expecting '" ++ p.typerep_down.typeName ++ "' type.\n" ;

  -- translation
  p.cond_tree = eqeq(p.base_tree, terminal(EQEQ_t, "=="), stringConst(str)) ;
  p.letAssigns_tree = [] ;
}

concrete production truePattern
p::Pattern ::= t::True_kwd
{
  p.pp = t.lexeme ;
  p.location = "[line: " ++ toString(t.line) ++ ", column: " ++ toString(t.column) ++ "] ";
  p.errors = "";

  p.defs = emptyDefs() ;

  p.typeErrors = if p.typerep_down.typeEquals(p.typerep_down, booleanTypeRep()).bValue
                 then ""
                 else p.location ++ "Error: expecting '" ++ p.typerep_down.typeName ++ "' type.\n" ;

  -- translation
  p.cond_tree = eqeq(p.base_tree, terminal(EQEQ_t, "=="), trueConst(terminal(True_kwd, "true"))) ;
  p.letAssigns_tree = [] ; 
}

concrete production falsePattern
p::Pattern ::= f::False_kwd
{
  p.pp = f.lexeme ;
  p.location = "[line: " ++ toString(f.line) ++ ", column: " ++ toString(f.column) ++ "] ";
  p.errors = "" ;

  p.defs = emptyDefs();

  p.typeErrors = if p.typerep_down.typeEquals(p.typerep_down, booleanTypeRep()).bValue
                 then ""
                 else p.location ++ "Error: expecting '" ++ p.typerep_down.typeName ++ "' type.\n" ;

  -- translation
  p.cond_tree = eqeq(p.base_tree, terminal(EQEQ_t, "=="), falseConst(terminal(False_kwd, "false"))) ;
  p.letAssigns_tree = [] ;
}

concrete production wildcPattern
p::Pattern ::= w::UnderScore_t
{
  p.pp = w.lexeme ;
  p.location = "[line: " ++ toString(w.line) ++ ", column: " ++ toString(w.column) ++ "] ";
  p.errors = "" ;

  p.defs = emptyDefs();

  p.typeErrors = "";

  -- translation
  p.cond_tree = trueConst(terminal(True_kwd, "true")) ;
  p.letAssigns_tree = [] ;
}

function printTypeRep
String ::= tr::Decorated TypeRep
{
 return ("name: " ++ tr.typeName ++ ", " ++"pp: " ++ tr.pp ++ ", " ++
         (if tr.isNonTerminal then "isNonTerminal, " else "") ++
         (if tr.isDecorated then "isDecorated, " else "") ) ;
}

concrete production varPattern
p::Pattern ::= v::Id_t
{
  p.pp = v.lexeme ;
  p.location = "[line: " ++ toString(v.line) ++ ", column: " ++ toString(v.column) ++ "] ";
  p.errors = er ;

  local attribute var_type :: Decorated TypeRep ;
  var_type = if p.v2_version
             then (if   p.v2_case_expr_type.isNonTerminal
                   then p.typerep_down   -- leave NTs as NTs, DecNTs as DecNTs, primitives as primitives 
                   else -- case expr is decorated - so NTs should be decorated
                        (if p.typerep_down.isDecorated || ! p.typerep_down.isNonTerminal
 	                 then p.typerep_down
	                 else refTypeRep(p.typerep_down) )
                  ) 
             else (if p.typerep_down.isDecorated || ! p.typerep_down.isNonTerminal
	           then p.typerep_down
	           else refTypeRep(p.typerep_down) ) ;


  p.defs = addValueDcl(v.lexeme, 
                       var_type,
                       addFullNameDcl(v.lexeme, v.lexeme, emptyDefs())) ;


  local attribute val :: [Decorated EnvItem] ;
  val = getValueDcl(v.lexeme, p.env) ;

  local attribute er :: String ;
  er = if length(val) != 0
        then p.location ++ "Error: the name for pattern variable '" ++ v.lexeme ++ "' is already used.\n" 
        else "";


  p.typeErrors = "";

  local attribute v2_base_tree2 :: Expr ; 
  v2_base_tree2 = if p.v2_version
                  then
                    (if p.v2_case_expr_type.isNonTerminal && p.typerep_down.isNonTerminal
                     then newFunction( terminal(New_kwd,"new"), terminal(LParen_t,"("), p.base_tree, terminal(RParen_t,")") )
                     else p.base_tree)
                  else
                    p.base_tree ;

  -- translation
  p.cond_tree = trueConst(terminal(True_kwd, "true")) ;
--  p.letAssigns_tree = [ assignExpr(v, terminal(Equal_t, "="), p.base_tree) ] ;
  p.letAssigns_tree = [ assignExpr(v, terminal(Equal_t, "="), v2_base_tree2'') ] ;

--  p.defs = if p.v2_version
--           then (if ! (isDirectory("AA", print("On varPattern:\n      " ++ printTypeRep(p.typerep_down) ++ 
--                                 "\n v2.. " ++ printTypeRep(p.v2_typerep_down)++"\n\n",unsafeio))).bValue 
--                 then
--                   addValueDcl(v.lexeme, 
--                               ------------------ p.v2_typerep_down ,
--                               p.v2_typerep_down ,
--                        addFullNameDcl(v.lexeme, v.lexeme, emptyDefs()))
--               
--                 else error("AA exists..............................")
--                )
--           else (addValueDcl(v.lexeme, 
--		    if p.typerep_down.isDecorated || ! p.typerep_down.isNonTerminal
--			    then p.typerep_down
--			    else refTypeRep(p.typerep_down), 
--	             addFullNameDcl(v.lexeme, v.lexeme, emptyDefs())) 
--                )
--           ;  

}
  
----------------------------------------------------
-- Added Functions
----------------------------------------------------

function toAssigns
LetAssigns ::= ls1::[AssignExpr]
{
  return if length(ls1) == 1 then assignListSingle(head(ls1))
	 else assigns(head(ls1), terminal(Comma_t, ","), toAssigns(tail(ls1))) ;
}


----------------------------------------------------
-- Aspects
----------------------------------------------------

inherited attribute prodName :: Id_t ;
attribute prodName occurs on ProductionBody ;

inherited attribute lhsName_down :: Id_t ;
attribute lhsName_down occurs on ProductionBody ;

inherited attribute rhsListExpr :: Expr ; 
attribute rhsListExpr occurs on ProductionBody ;

synthesized attribute lhsName_up :: Id_t ;
attribute lhsName_up occurs on ProductionSignature, ProductionLHS, ProductionLHSElem ;

synthesized attribute genRhsListExpr :: Expr ;
attribute genRhsListExpr occurs on ProductionSignature, ProductionRHS, ProductionRHSElem ;

synthesized attribute rhsAnyType :: Boolean ;
attribute rhsAnyType occurs on ProductionRHSElem ;

synthesized attribute rhsTypeAllowed :: Boolean ;
attribute rhsTypeAllowed occurs on ProductionSignature, ProductionRHS, ProductionRHSElem ;

inherited attribute isAspect :: Boolean ;
attribute isAspect occurs on ProductionBody ;

aspect production productionDcl
top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{
   body.prodName = id ;
   body.lhsName_down = ns.lhsName_up ;
   body.rhsListExpr = ns.genRhsListExpr ;
   body.isAspect = false ;
}

aspect production aspectAGDcl
top::AGDcl ::=
   a::Aspect_kwd p::Production_kwd id::QName ns::AspectSignature
   body::ProductionBody
{
   body.isAspect = true ;
}

aspect production productionSignatureEmptyRHS
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t
{
   top.lhsName_up = lhs.lhsName_up ;
   top.genRhsListExpr = emptyList(terminal(LSqr_t, "["),
				  terminal(RSqr_t, "]")) ;
}

aspect production productionSignature
top::ProductionSignature ::= lhs::ProductionLHS kwd::CCEQ_t rhs::ProductionRHS 
{
   top.lhsName_up = lhs.lhsName_up ;
   top.genRhsListExpr = if ! rhs.rhsTypeAllowed 
			then errorFunction(terminal(Error_kwd, "error"),
					   terminal(LParen_t, "("),
				 	   stringConst(terminal(String_t, "\"Error: cannot pattern match on production or IO types.\\n\"")),
					   terminal(RParen_t, ")"))
			else rhs.genRhsListExpr ;
}

aspect production productionLHS
top::ProductionLHS ::= lhs::ProductionLHSElem
{
   top.lhsName_up = lhs.lhsName_up ;
}

aspect production productionLHSElem
top::ProductionLHSElem ::= id::Id_t kwd::HasType_t t::Type
{
  top.lhsName_up = id ;
}


aspect production productionRHSSingle
top::ProductionRHS ::= rhs::ProductionRHSElem
{

  top.genRhsListExpr = fullList(terminal(LSqr_t, "["),
			exprsSingle(rhs.genRhsListExpr), 
			terminal(RSqr_t, "]"));


--   top.genRhsListExpr = productionApp(baseExpr(qNameId( terminal(Id_t, "cons_AnyTypeList" ))),
--	          terminal(LParen_t, "("),
--                  exprsCons (
--			     terminal(Comma_t, ","),
--                             exprsSingle(emptyProductionApp(baseExpr( qNameId (terminal (Id_t, "nil_AnyTypeList"))), 
--		 	                                      terminal(LParen_t, "("),
--				 	                      terminal(RParen_t, ")")))
--                              ),
--	          terminal(RParen_t, ")")
--                  );

   top.rhsTypeAllowed = rhs.rhsTypeAllowed ;
}

aspect production productionRHSCons
top::ProductionRHS ::= h::ProductionRHSElem t::ProductionRHS
{

   top.genRhsListExpr =  consList(terminal(Cons_t, "cons"),
                   terminal(LParen_t, "("),                   
                   h.genRhsListExpr,  terminal(Comma_t,","), t.genRhsListExpr,
                   terminal(RParen_t, ")")
                  ) ;

--   top.genRhsListExpr =  productionApp(baseExpr(qNameId(terminal(Id_t, "cons_AnyTypeList"))),
--                   terminal(LParen_t, "("),                   
--                   exprsCons (h.genRhsListExpr,  terminal(Comma_t,","), exprsSingle( t.genRhsListExpr )),
--                   terminal(RParen_t, ")")
--                  ) ;

   top.rhsTypeAllowed = h.rhsTypeAllowed && t.rhsTypeAllowed ;
}

aspect production productionRHSElem
top::ProductionRHSElem ::= id::Id_t kwd::HasType_t t::Type
{
   top.genRhsListExpr = if t.typerep.isAnyType 
			then baseExpr(qNameId(id)) 
			else cast_t(terminal(Cast_kwd, "cast"), baseExpr(qNameId(id)), anyTypeTypeRep());    

   top.rhsAnyType = if t.typerep.isAnyType then true else false ;
   top.rhsTypeAllowed = if t.typerep.isProduction || t.typerep.isIO
			then false
			else true ;   
}


aspect production productionBody
top::ProductionBody ::= l::LCurly_t stmts::ProductionStmts r::RCurly_t
{
   local attribute stmt1 :: ProductionStmt ;
   stmt1 = stmtToAttrDef(attrDef(
	lhsExprTwo(top.lhsName_down, terminal(Dot_t, "."), qNameId(terminal(Id_t, "patProdName"))),
        stringConst(terminal(String_t, "\"" ++ top.prodName.lexeme ++ "\""))));
	

   local attribute stmt2 :: ProductionStmt ;
   stmt2 = stmtToAttrDef(attrDef(
       	lhsExprTwo(top.lhsName_down, terminal(Dot_t, "."), qNameId(terminal(Id_t, "patChildList"))),
        top.rhsListExpr));

   local attribute addedStmts :: ProductionStmts ;
   addedStmts = if top.isAspect then productionStmtsNone() else
		productionStmtsCons(stmt1'', productionStmtsSingle(stmt2'')) ;

   extraStmts <- addedStmts'' ;
}

aspect production emptyProductionBody
top::ProductionBody ::= 
{
   local attribute stmt1 :: ProductionStmt ;
   stmt1 = stmtToAttrDef(attrDef(
	lhsExprTwo(top.lhsName_down, terminal(Dot_t, "."), qNameId(terminal(Id_t, "patProdName"))),
        stringConst(terminal(String_t, "\"" ++ top.prodName.lexeme ++ "\""))));

   local attribute stmt2 :: ProductionStmt ;
   stmt2 = stmtToAttrDef(attrDef(
       	lhsExprTwo(top.lhsName_down, terminal(Dot_t, "."), qNameId(terminal(Id_t, "patChildList"))),
        top.rhsListExpr));

   local attribute addedStmts :: ProductionStmts ;
   addedStmts = if top.isAspect then productionStmtsNone() else 
		productionStmtsCons(stmt1'', productionStmtsSingle(stmt2'')) ;

   extraStmts <- addedStmts'' ;
}

aspect production nonterminalDcl
top::AGDcl ::= n::NonTerminal_kwd id::Id_t s::Semi_t
{
   local attribute occurs1 :: AGDcl ;
   occurs1 = attributionDcl(terminal(Attribute_kwd, "attribute"),
			    qNameId(terminal(Id_t, "patProdName")),
			    terminal(Occurs_kwd, "occurs"),
			    terminal(On_kwd, "on"),
			    qNameId(id),
			    terminal(Semi_t, ";")) ;

   local attribute occurs2 :: AGDcl ;
   occurs2 = attributionDcl(terminal(Attribute_kwd, "attribute"),
			    qNameId(terminal(Id_t, "patChildList")),
			    terminal(Occurs_kwd, "occurs"),
			    terminal(On_kwd, "on"),
			    qNameId(id),
			    terminal(Semi_t, ";")) ;
			    
   extraDcls <- agDclSeq(occurs1'', occurs2'') ;
} 

aspect production closedNonterminalDcl
top::AGDcl ::= c::Closed_kwd n::NonTerminal_kwd  id::Id_t  s::Semi_t
{
   local attribute occurs1 :: AGDcl ;
   occurs1 = attributionDcl(terminal(Attribute_kwd, "attribute"),
			    qNameId(terminal(Id_t, "patProdName")),
			    terminal(Occurs_kwd, "occurs"),
			    terminal(On_kwd, "on"),
			    qNameId(id),
			    terminal(Semi_t, ";")) ;

   local attribute occurs2 :: AGDcl ;
   occurs2 = attributionDcl(terminal(Attribute_kwd, "attribute"),
			    qNameId(terminal(Id_t, "patChildList")),
			    terminal(Occurs_kwd, "occurs"),
			    terminal(On_kwd, "on"),
			    qNameId(id),
			    terminal(Semi_t, ";")) ;
			    
   extraDcls <- agDclSeq(occurs1'', occurs2'') ;
}

aspect production startNonterminalDcl
top::AGDcl ::= stk::Start_kwd n::NonTerminal_kwd id::Id_t  s::Semi_t
{
   local attribute occurs1 :: AGDcl ;
   occurs1 = attributionDcl(terminal(Attribute_kwd, "attribute"),
			    qNameId(terminal(Id_t, "patProdName")),
			    terminal(Occurs_kwd, "occurs"),
			    terminal(On_kwd, "on"),
			    qNameId(id),
			    terminal(Semi_t, ";")) ;

   local attribute occurs2 :: AGDcl ;
   occurs2 = attributionDcl(terminal(Attribute_kwd, "attribute"),
			    qNameId(terminal(Id_t, "patChildList")),
			    terminal(Occurs_kwd, "occurs"),
			    terminal(On_kwd, "on"),
			    qNameId(id),
			    terminal(Semi_t, ";")) ;
			    
   extraDcls <- agDclSeq(occurs1'', occurs2'') ;
}



