grammar silver:extension:string_accum ;


import silver:base ;

import silver:analysis:typechecking only typeErrors ;
import silver:translation:ag:haskell only haskell ;

-- mkStringAccum
----------------
terminal mkStringAccum_kwd  /mkStringAccum/ lexer precedence = 15 ;

concrete production mkStringAccum_p
top::Expr ::= h::mkStringAccum_kwd lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.location = "[line: " ++ toString(h.line) ++ ", column: " ++ toString(h.column) ++ "] ";
  top.pp = "mkStringAccum(" ++ l.pp ++ ")";
  top.errors = l.errors ;
  top.typeErrors = l.typeErrors ++ te ;
  local attribute te :: String ;
  te = if l.typerep.typeName == "String" 
       then ""
       else top.location ++ "Error: Operand to mkStringAccum must be of type String.\n";

  top.typerep = stringAccumTypeRep();

  top.haskell = @"(" ++ l.haskell ++ @" ++ )" ;
  --top.haskell = "(" ++ l.haskell ++ " )" ;

  top.parserUses = l.parserUses;


--  forwards to
--   cast_t(terminal(Cast_kwd, ""), headListNoCast(l''), l.typerep.listComponent);
}

-- mkString
----------------
terminal mkString_kwd  /mkString/ lexer precedence = 15 ;

concrete production mkString_p
top::Expr ::= h::mkString_kwd lp::LParen_t l::Expr rp::RParen_t 
{ 
  top.location = "[line: " ++ toString(h.line) ++ ", column: " ++ toString(h.column) ++ "] ";
  top.pp = "mkString(" ++ l.pp ++ ")";
  top.errors = l.errors ;
  top.typeErrors = l.typeErrors ++ te ;
  local attribute te :: String ;
  te = if l.typerep.typeName == "StringAccum" 
       then ""
       else top.location ++ "Error: Operand to mkString must be of type StringAccum.\n";

  top.typerep = stringTypeRep();

  top.haskell = @"(" ++ l.haskell ++ @" \"\" )" ;
  --top.haskell = "(" ++ l.haskell ++ " )" ;

  top.parserUses = l.parserUses;


--  forwards to
--   cast_t(terminal(Cast_kwd, ""), headListNoCast(l''), l.typerep.listComponent);
}


-- StringAccum literal --
-------------------------
terminal StringAccum_t /([\@]|[\'])[\"]([^\"\\]|[\\][\"]|[\\][\\]|[\\]n|[\\]r|[\\]t)*[\"]/ lexer precedence = 5;

concrete production stringAccumConst
top::Expr ::= s::StringAccum_t
{
  top.pp = s.lexeme;
  top.location = "[line: " ++ toString(s.line) ++ ", column: " ++ toString(s.column) ++ "] ";
  top.errors = "";
  top.typeErrors = "";
  top.typerep = stringAccumTypeRep();

  top.haskell = @"(" ++ mkStringAccum(substring(1,length(s.lexeme),s.lexeme)) ++ @" ++)";
  --top.haskell = "(" ++ s.lexeme ++ " ++)";


  top.parserUses = [];
}


-- ++ --
--------
aspect production plusPlus
top::Expr ::= e1::Expr p::PlusPlus_t e2::Expr
{
  handler <- if e1.typerep.typeName=="StringAccum" && e2.typerep.typeName=="StringAccum"   --&& false  --EVW   
	     then [appendStringAccum(e1'', terminal(AccumPlusPlus_t,"@++",p.line,p.column), e2'')]
	     else [];



--  handler <- if e1.typerep.typeName=="StringAccum" && e2.typerep.typeName == "String"
--	     then [appendStringAccum(e1'', p, 
--                      mkStringAccum_p(terminal(mkStringAccum_kwd,"mkStringAccum"), terminal(LParen_t,"("), e2'', terminal(RParen_t,")")))]
--	     else [];
-- worked with above, now trying with below
-- failed with all three
--  handler <- if e1.typerep.typeName == "String" && e2.typerep.typeName=="StringAccum"
--	     then [appendStringAccum(
--                      mkStringAccum_p(terminal(mkStringAccum_kwd,"mkStringAccum"), terminal(LParen_t,"("), e1, terminal(RParen_t,")")),
--                      p, e2)]
--	     else [];
}


terminal AccumPlusPlus_t /[\@][\+][\+]/  lexer precedence = 6,  precedence = 11, association = left;


concrete production appendStringAccum
top::Expr ::= l::Expr o::AccumPlusPlus_t r::Expr
{ 
  top.haskell = @"(" ++ l.haskell ++ @" . " ++ r.haskell ++ @")";
  --top.haskell = "(" ++ l.haskell ++ " ++ " ++ r.haskell ++ ")";

  top.pp = l.pp ++ " " ++ o.lexeme ++ " " ++ r.pp ;
  top.location = "[line: " ++ toString(o.line) ++ ", column: " ++ toString(o.column) ++ "] ";

  top.typerep = l.typerep;

  local attribute e1 :: String;
  e1 = if l.typerep.typeName=="StringAccum" then ""
       else top.location ++ "Error: First argument to append must be a StringAccum.\n";

  local attribute e2 :: String;
  e2 = if r.typerep.typeName=="StringAccum" then ""
       else top.location ++ "Error: Second argument to append must be a StringAccum.\n";

  top.typeErrors = l.typeErrors ++ r.typeErrors ++ e1 ++ e2 ;
  top.errors = l.errors ++ r.errors ;
  top.parserUses = l.parserUses ++ r.parserUses;


--  forwards to
--   productionApp(baseExpr(qNameId(terminal(Id_t, "append_AnyTypeList"))),
--                 terminal(LParen_t, "("),
--                 exprsCons(l'', terminal(Comma_t,","), exprsSingle ( r'' ) ), 
--                 terminal(RParen_t, ")")                 
--	        );
}


-- printFunction --
-------------------
aspect production printFunction
top::Expr ::= p::Print_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
--  handler <- if e1.typerep.typeName == "StringAccum"   -- && false --EVW
--	     then [stringAccumPrintFunction(p,l,e1,c,e2,r)]
--	     else [];
}


abstract production stringAccumPrintFunction
top::Expr ::= p::Print_kwd l::LParen_t e1::Expr c::Comma_t e2::Expr r::RParen_t
{
  top.pp = "print ( " ++ e1.pp ++ ", " ++ e2.pp ++ " )";
  top.location = "[line: " ++ toString(p.line) ++ ", column: " ++ toString(p.column) ++ "] ";

  top.typerep = ioTypeRep();

  top.errors = e1.errors ++ e2.errors;
  top.typeErrors = e1.typeErrors ++ e2.typeErrors ++ te ;

  local attribute te :: String ;
  te = if e1.typerep.typeName == "StringAccum" 
       then ""
       else top.location ++ "Error: Operand to print (StringAccum version) must be of type StringAccum.\n";


  local attribute task :: StringAccum;
  task = @"((putStr (" ++ e1.haskell ++ @"\"\" )) >>= \\_ -> (return True ))";

  top.haskell = @"( " ++ e2.haskell ++ @">>= \\t -> if t then " ++  task ++ @" else " ++ task ++ @" )";
}

