grammar silver:translation:ag:haskell:core;

import silver:base hiding grammarName;
import silver:utils:command;
import silver:utils;
import core;

function stringPP
String ::= s1::String s2::String
{
  return s1 ++ s2;
}

abstract production genHaskell
top::Unit ::= a::Command specs::[GrammarSpec]
{
  forwards to genHaskellHelp(a'', specs)
	with {
		ioIn = print("Generating Haskell Translation.\n", top.ioIn);
	};
}

abstract production genHaskellHelp
top::Unit ::= a::Command specs::[GrammarSpec]
{
 -- This production defines a number of production attributes that aspects in
 -- translation:haskell(_|c|java):core:BuildProcess.sv will use to generate
 -- ghc_compile and similar files.

 -- These aspects control the io and can thus ignore the attributes and 
 -- helper functions defined here.

  production attribute globalHaskell :: String with stringPP;
  globalHaskell := "";

  production attribute term :: String;
  term = "module Terminal where\n" ++
   	 "data TERMINAL = TERMINAL { syns_TERMINAL :: SYNS_TERMINAL}\n" ++
	 "data SYNS_TERMINAL = SYNS_TERMINAL { attr_TERMINAL__lexeme :: String, " ++ 
					      "attr_TERMINAL__line :: Int, " ++ 
					      "attr_TERMINAL__column :: Int }\n\n" ++
 	 "instance Show TERMINAL where\n" ++
         "  show t = \"[Line: \" ++ (show (attr_TERMINAL__line (syns_TERMINAL t))) ++ " ++
              "\"] [Column: \" ++ (show (attr_TERMINAL__column (syns_TERMINAL t))) ++ " ++
              "\"] [Token: \" ++ (attr_TERMINAL__lexeme (syns_TERMINAL t)) ++ \"]\"\n\n";

-- the possibly future string accumulater version
--         "module Terminal where\n" ++
--   	 "data TERMINAL = TERMINAL { syns_TERMINAL :: SYNS_TERMINAL}\n" ++
--	 "data SYNS_TERMINAL = SYNS_TERMINAL { attr_TERMINAL__lexeme :: String -> String, " ++    ------ change here
--					      "attr_TERMINAL__line :: Int, " ++ 
--					      "attr_TERMINAL__column :: Int }\n\n" ++
-- 	 "instance Show TERMINAL where\n" ++
--         "  show t = \"[Line: \" ++ (show (attr_TERMINAL__line (syns_TERMINAL t))) ++ " ++
--              "\"] [Column: \" ++ (show (attr_TERMINAL__column (syns_TERMINAL t))) ++ " ++
--              "\"] [Token: \" ++ (attr_TERMINAL__lexeme (syns_TERMINAL t)) \"\" ++ \"]\"\n\n"


  production attribute hn :: String;
  hn = makeName(a.grammarName);

  production attribute pUses :: [String];
  pUses = mergeParsers(specs);

  production attribute mod_boot :: String;
  mod_boot = "module Module_" ++ hn ++ " where\n" ++ buildModuleBootSpec(buildStartSet(specs));

--  production attribute globalHaskell :: String with stringPP;
--  globalHaskell := "";


  production attribute header :: String;
  header = "module Module_" ++ hn ++ " where\n" ++
           "import Terminal\nimport System.IO.Unsafe\nimport Directory\nimport List\nimport System\nimport Data.IORef\nimport Char\n" ++ 
            buildImportString(specs, pUses) ++ 
	   "idnum_counter :: IORef Int\n"++
	   "idnum_counter = unsafePerformIO (newIORef (0::Int))\n\n" ++
	   "genint :: () -> Int\n" ++
	   "genint () = unsafePerformIO\n" ++ 
	   "                 ( do\n" ++
	   "                    y <- readIORef idnum_counter\n" ++
	   "                    writeIORef idnum_counter (y + 1)\n" ++ 
	   "                    return y\n" ++                    
	   "                 )\n\n"  ++
           "mkIOStrStr :: IO String -> IO (String -> String) \n" ++
           "mkIOStrStr ios = ios >>= \\ t -> return (t ++)  \n\n" ++
           globalHaskell  ++ "\n\n"
           ;

}



function mergeParsers
[String] ::= l::[GrammarSpec]
{
  return if null(l) then [] else head(l).rootSpec.parserUses ++ mergeParsers(tail(l));
}

function buildModuleBootSpec
String ::= sl::[String]
{
  return if null(sl) then "\n"
         else "data " ++ head(sl) ++ "\n" ++ "data INHS_" ++ head(sl) ++ "\n" ++ buildModuleBootSpec(tail(sl));
}

function buildStartSet
[String] ::= cl::[GrammarSpec]
{
  local attribute spec :: Decorated RootSpec;
  spec = head(cl).rootSpec;

  local attribute has_start :: Boolean;
  has_start = spec.hasStartSymbol;

  local attribute shn :: String;
  shn = "NT_" ++ makeName(spec.startName);

  local attribute rest :: [String];
  rest = buildStartSet(tail(cl));

  return if null(cl) 
	 then []
	 else if has_start && !containsString(shn, rest)
	      then cons(shn, rest)
	      else rest;
}

function buildImportString
String ::= cl::[GrammarSpec] s::[String]
{
  local attribute n :: String;
  n = "Module_" ++ makeName(head(cl).name);

  local attribute has_start :: Boolean;
  has_start = head(cl).rootSpec.hasStartSymbol && containsString(head(cl).name ++ ":parse", s);

  return if null(cl)
	 then "" 
	 else if has_start
	      then "import {-# SOURCE #-} " ++ n ++ "_Parser\n" ++ buildImportString(tail(cl), s)
	      else "\n" ++ buildImportString(tail(cl), s);
}


function buildParserString
String ::= cl::[GrammarSpec] s::[String]
{
  local attribute n :: String;
  n = "Module_" ++ makeName(head(cl).name);

  local attribute has_start :: Boolean;
  has_start = head(cl).rootSpec.hasStartSymbol && containsString(head(cl).name ++ ":parse", s);

  return if null(cl)
	 then "" 
	 else if has_start
	      then 
		   "echo \"ghc -c " ++ n ++ "_Parser.hs\"\n" ++ 
		   "ghc -c " ++ n ++ "_Parser.hs\n" ++ buildParserString(tail(cl), s)
	      else "\n" ++ buildParserString(tail(cl), s);
}



function buildGHCBootString
String ::= cl::[GrammarSpec] s::[String]
{
  local attribute n :: String;
  n = "Module_" ++ makeName(head(cl).name);

  local attribute has_start :: Boolean;
  has_start = head(cl).rootSpec.hasStartSymbol && containsString(head(cl).name ++ ":parse", s);

  return if null(cl)
	 then "" 
	 else if has_start
	      then 
		   "echo \"ghc -c " ++ n ++ "_Parser.hs-boot\"\n" ++ 
		   "ghc -c " ++ n ++ "_Parser.hs-boot\n" ++ buildGHCBootString(tail(cl), s)
	      else "\n" ++ buildGHCBootString(tail(cl), s);
}

function writeMainHS
IO ::= i::IO a::Decorated Command
{
  local attribute hname :: String;
  hname = makeName(a.grammarName);

  local attribute b :: String;
  b = "module Main where\n" ++
	 "import System\n" ++
	 "import Module_" ++ hname ++ "\n" ++ 
	 "main = do\n" ++ 
	 "args <- System.getArgs\n" ++
         "arg  <- return (if (length args) == 0 then \"\" else (tail (foldl (\\old new -> old ++ \" \" ++ new) \"\" args)))\n" ++ 
         "inhs <- return (INHS_NT_core_Main {\n" ++
	 "		   attr_NT_core_Main__core_ioIn = (return 1)\n" ++
	 "	          })\n" ++
--         (if accum()
--          then 
---------------           "syns <- return (syns_NT_core_Main  (production_" ++ hname ++ "_main (arg ++) inhs))\n" 
--          else "syns <- return (syns_NT_core_Main  (production_" ++ hname ++ "_main arg inhs))\n"
--         ) 
           "syns <- return (syns_NT_core_Main  (production_" ++ hname ++ "_main arg inhs))\n"
          ++
         "( attr_NT_core_Main__core_ioOut syns )\n";

  return writeFile(a.generatedPath ++ "Main.hs", b, i);
}

function writeAll
IO ::= i::IO a::Decorated Command l::[GrammarSpec] s::[String]
{
  local attribute now :: IO;
  now = writeSpec(i, head(l).name, head(l).rootSpec, a, s);

  local attribute recurse :: IO;
  recurse = writeAll(now, a, tail(l), s);

  return if null(l) then i else recurse;
}


function writeSpec
IO ::= i::IO gName::String r::Decorated RootSpec a::Decorated Command s::[String]
{
  local attribute thn :: String;
  thn = makeName(a.grammarName);

  -- the haskell name of the grammar.
  local attribute hn :: StringAccum;
  hn = mkStringAccum(makeName(gName));
  
  local attribute has_start :: Boolean;
  has_start = r.hasStartSymbol && containsString(gName ++ ":parse", s);

  local attribute header :: StringAccum;
  header = if has_start then @"production_" ++ hn ++ @"_parse str = Module_" ++ hn ++ @"_Parser.doparse str " else @"\n";

  return appendFile(a.generatedPath ++ "Module_" ++ thn ++ ".hs", mkString(header ++ @"\n\n" ++ r.haskell), i);
}
