--TODO  the copy function must be smart enought to handle any type of data.

grammar silver:translation:c:core;
import silver:base hiding grammarName;
import silver:utils:command;
import silver:utils;

import core;

function stringPP
String ::= s1::String s2::String
{
  return s1 ++ s2;
}

aspect production run
top::RunUnit ::= iIn::IO args::String p::Production(Root ::= String)
{
  ops <- [genC(a'', grammars, gEnv)]; 
}

abstract production genC
top::Unit ::= a::Command specs::[GrammarSpec] e::Decorated Env
{
  forwards to genCHelp(a'', specs, e)
	with {
		ioIn = print("Generating C Translation.\n", top.ioIn);
	};
}

abstract production genCHelp
top::Unit ::= a::Command specs::[GrammarSpec] e::Decorated Env
{
  --The top level name
  local attribute hn :: String;
  hn = makeName(a.grammarName);

  --the parsers that are used
  local attribute pUses :: [String];
  pUses = mergeParsers(specs);

  local attribute theSpecs :: [Decorated ProductionSpec];
  theSpecs = mergeProductions(specs);

  local attribute pSpec :: [Decorated ProductionSpec];
  pSpec = getProductionSpecs(theSpecs);

  local attribute aSpec :: [Decorated ProductionSpec];
  aSpec = getAspectSpecs(theSpecs);

  local attribute i1 :: IO;
  i1 = writeMain(top.ioIn, a);

  local attribute i2 :: IO;
  i2 = writeDefs(i1, a, pSpec, aSpec, e);

  local attribute i3 :: IO;
  i3 = writeInit(i2, a, pSpec, aSpec, e);

  local attribute i4 :: IO;
  i4 = writeAll(i3, a, specs, pUses);

  local attribute c :: IOString;
  c = cwd(i4);

  local attribute ghc_compile_str :: String;
  ghc_compile_str = 
		    "echo \"cc -c init.c\"\n" ++ 
		    "cc  -c init.c\n" ++ 
                     buildCCString(specs, pUses) ++ 
                     buildParserString(specs, pUses) ++ 
		    "echo \"cc -c main.c\"\n" ++
		    "cc  -c main.c\n" ++
		    "echo \"cc -o " ++ c.sValue ++ "/" ++ a.outName ++ " main.o init.o " ++ buildLinkString(specs, pUses) ++ "\"\n" ++
		    "cc -o " ++ c.sValue ++ "/" ++ a.outName ++ " main.o init.o " ++ buildLinkString(specs, pUses) ++ "\n";

  local attribute i5 :: IO;
  i5 = writeFile(a.generatedPath ++ "c-compile", ghc_compile_str, c.io);

  top.io = system("pushd " ++ a.generatedPath ++ ">> /dev/null; chmod +x ./c-compile; ./c-compile; popd >> /dev/null", i5).io;

  top.code = 0;
  top.order = 3;
}

function getProductionSpecs
[Decorated ProductionSpec] ::= l::[Decorated ProductionSpec] {
  return if null(l) 
	 then [] 
	 else ((if head(l).isProductionSpec then [head(l)] else []) +++ getProductionSpecs(tail(l)));
}

function getAspectSpecs
[Decorated ProductionSpec] ::= l::[Decorated ProductionSpec] {
  return if null(l) 
	 then [] 
	 else ((if head(l).isAspectSpec then [head(l)] else []) +++ getAspectSpecs(tail(l)));
}
function writeDefs
IO ::= i::IO a::Decorated Command ps :: [Decorated ProductionSpec] aSpec :: [Decorated ProductionSpec]  e :: Decorated Env
{
  local attribute defs_header :: String;
  defs_header = 
"#ifndef DEFS\n" ++
"#define DEFS\n\n" ++

"struct Value {\n" ++
"  int defined;\n" ++
"  int isNode;\n" ++
"  union{\n" ++
"    int i;\n" ++
"    char * s;\n" ++
"    float f;\n" ++
"    union Node * n;\n" ++
"  };\n" ++
"};\n\n" ++

"struct T {\n" ++
"  int child;\n" ++
"  int value;\n" ++
"};\n\n" ++

"struct NT {\n" ++
"  int production;\n" ++
"  int child;\n" ++
"  union Node * parent;\n" ++
"  struct Value * children;\n" ++
"  struct Value ** locals;\n" ++
"  union Node * forward;\n" ++
"  struct Value * values;\n" ++
"};\n\n" ++

"union Node {\n" ++
"  struct T t;\n" ++
"  struct NT nt;\n" ++
"};\n\n" ++

"struct FUNC_TYPE {\n" ++
"  struct Value (*func)(union Node *);\n" ++
"};\n\n" ++

"struct ASPECT_TYPE {\n" ++
"  union Node * (*func)(union Node *);\n" ++
"};\n\n" ++

"#define TRUE 1\n" ++
"#define FALSE 0\n\n" ++

"extern struct FUNC_TYPE *** func_table;\n" ++
"extern struct ASPECT_TYPE ** aspect_table;\n" ++
"extern void init();\n" ++
"extern struct Value fail(char * );\n" ++
"extern struct Value mkInt(int);\n" ++
"extern struct Value mkFloat(float);\n" ++
"extern struct Value mkStr(char *);\n" ++
"extern struct Value mkNode(union Node *);\n" ++
"extern union Node * copy(union Node *);\n" ++
"extern struct Value getSynAttribute(union Node *, int);\n" ++ 
"extern struct Value getInhAttribute(union Node *, int);\n\n" ++ 

"extern int P;\n\n" ++ 
defineProductions(0, ps, aSpec, true) ++ 
defineAttributes(e, true) ++ 

defineFunctionsForAttributesOnProds(ps, e) ++ "\n" ++

"#endif\n" ;

  return  writeFile(a.generatedPath ++ "defs.h", defs_header, i);
}

function writeInit
IO ::= i::IO a::Decorated Command ps::[Decorated ProductionSpec] aSpec::[Decorated ProductionSpec] e::Decorated Env
{
  local attribute init_header :: String;
  init_header = 
"#include <stdlib.h>\n" ++
"#include <stdio.h>\n" ++
"#include \"defs.h\"\n\n" ++

"int P = " ++ toString(length(ps)) ++ ";\n\n" ++ 

defineProductions(0, ps, aSpec, false) ++ 
defineAttributes(e, false) ++ 

"struct Value getSynAttribute(union Node * n, int attr){\n" ++ 
"  return func_table[n->nt.production][0][attr].func(n);\n" ++  
"}\n\n" ++

"struct Value getInhAttribute(union Node * n, int attr){\n" ++
"  return func_table[n->nt.parent->nt.production][n->nt.child][attr].func(n);\n" ++
"}\n\n" ++

"struct Value fail(char * msg){\n" ++
"  printf(msg);\n" ++
"  exit(1);\n" ++
"  struct Value result;\n" ++
"  return result;\n" ++
"}\n\n" ++ 

"struct Value mkNode(union Node * value){\n" ++
"  struct Value result;\n" ++
"  result.n = value;\n" ++
"  return result;\n" ++
"}\n\n" ++ 

"union Node * copy(union Node * value){\n" ++
"  union Node * result = malloc(sizeof(union Node));\n" ++ 
"  result->nt.production = value->nt.production;\n" ++

"  int num_atts = sizeof(value->nt.values);\n" ++
"  int num_children = sizeof(value->nt.children);\n" ++
"  int num_locals =  sizeof(value->nt.locals);\n\n" ++
  
"  result->nt.values = malloc(num_atts * sizeof(struct Value));\n\n" ++ 

"  int i;\n" ++ 
"  for(i = 0; i < num_atts; i++){\n" ++ 
"    result->nt.values[i].defined = FALSE;\n" ++ 
"  }\n\n" ++ 

"  result->nt.children = malloc(num_children * sizeof(struct Value));\n" ++ 
"  for(i=0; i < num_children; i++){\n" ++
"    result->nt.children[i] = copy(value->nt.children[i]);\n" ++
"  }\n\n" ++

"  result->nt.locals = malloc(num_locals * sizeof(struct Value));\n" ++ 
"  for(i=0; i < num_locals; i++){\n" ++
"    aspect_table[value->nt.production][i].func(result);\n" ++
"  }\n\n" ++

"  result->nt.forward = value->nt.forward;\n\n" ++ 
"  return result;\n" ++
"}\n\n" ++ 

"struct Value mkInt(int value){\n" ++
"  struct Value result;\n" ++
"  result.i = value;\n" ++
"  return result;\n" ++
"}\n\n" ++ 

"struct Value mkFloat(float value){\n" ++
"  struct Value result;\n" ++
"  result.f = value;\n" ++
"  return result;\n" ++
"}\n\n" ++ 

"struct Value mkStr(char * value){\n" ++
"  struct Value result;\n" ++
"  result.s = value;\n" ++
"  return result;\n" ++
"}\n\n" ++ 

"struct FUNC_TYPE *** func_table;\n" ++
"struct ASPECT_TYPE ** aspect_table;\n\n" ++

"void init(){\n" ++
"  func_table = malloc(P * sizeof(struct FUNC_TYPE**));\n" ++ 
"  aspect_table = malloc(P * sizeof(struct ASPECT_TYPE**));\n\n" ++ 
defineInitForAttributesOnProds(ps, aSpec, e) ++ 
"}\n" ;

  return writeFile(a.generatedPath ++ "init.c", init_header, i);
}

function writeMain
IO ::= i::IO a::Decorated Command
{
  local attribute hn :: String;
  hn = makeName(a.grammarName);

  local attribute b :: String;
  b = 
"#include <stdlib.h>\n"++
"#include <stdio.h>\n"++
"#include \"defs.h\"\n"++
"int main(char * argc){\n"++
"  init();\n"++
--"  union Node * n = production_" ++ hn ++ "(argc);\n"++
--"  func_table[n->nonterminal.production][0][0].func(n);\n"++
"  return 0;\n"++
"}\n";

  return writeFile(a.generatedPath ++ "main.c", b, i);
}



function writeAll
IO ::= i::IO a::Decorated Command l::[GrammarSpec] s::[String]
{
  local attribute now :: IO;
  now = writeSpec(i, head(l).name, head(l).rootSpec, a, s);

  local attribute recurse :: IO;
  recurse = writeAll(now, a, tail(l), s);

  return if null(l) then i else recurse;
}


function writeSpec
IO ::= i::IO gName::String r::Decorated RootSpec a::Decorated Command s::[String]
{
  -- the haskell name of the grammar.
  local attribute hn :: String;
  hn = makeName(gName);
  
  local attribute has_start :: Boolean;
  has_start = r.hasStartSymbol && containsString(gName ++ ":parse", s);

  local attribute header :: String;
  header = "#include <stdlib.h>\n"++
	   "#include <stdio.h>\n"++
	   "#include \"defs.h\"\n\n";
--  header = if has_start then "production_" ++ hn ++ "_parse str = Module_" ++ hn ++ "_Parser.doparse str " else "\n";

  return writeFile(a.generatedPath ++ "Module_" ++ hn ++ ".c", header ++ r.translation, i);
}

function buildLinkString
String ::= cl::[GrammarSpec] s::[String]
{
  local attribute n :: String;
  n = "Module_" ++ makeName(head(cl).name);

  return if null(cl) 
	 then ""
	 else n ++ ".o " ++ if head(cl).rootSpec.hasStartSymbol && containsString(head(cl).name ++ ":parse", s)
	      		    then n ++ "_Parser.o " ++ n ++ "_Scanner.o " ++ buildLinkString(tail(cl), s)
	  	            else buildLinkString(tail(cl), s);
}

function buildCCString
String ::= gl::[GrammarSpec] s::[String]
{ 
  local attribute n :: String;
  n = "Module_" ++ makeName(head(gl).name);

  local attribute recurse :: String;
  recurse = buildCCString(tail(gl), s);

  return if null(gl)
	 then "" 
	 else "echo \"cc -c " ++ n ++ ".c\"\n" ++
	      "cc -c " ++ n ++ ".c\n" ++ recurse;	      
}

function buildParserString
String ::= cl::[GrammarSpec] s::[String]
{
  local attribute n :: String;
  n = "Module_" ++ makeName(head(cl).name);

  return if null(cl)
	 then "" 
	 else if head(cl).rootSpec.hasStartSymbol && containsString(head(cl).name ++ ":parse", s)
	      then 
		   "echo \"cc -c " ++ n ++ "_Parser.c " ++ n ++ "_Scanner.c \"\n" ++ 
		   "cc -c " ++ n ++ "_Parser.c " ++ n ++ "_Scanner.c\n" ++ buildParserString(tail(cl), s)
	      else "\n" ++ buildParserString(tail(cl), s);
}

function mergeParsers
[String] ::= l::[GrammarSpec]
{
  return if null(l) then [] else head(l).rootSpec.parserUses ++ mergeParsers(tail(l));
}

function mergeProductions
[Decorated ProductionSpec] ::= l::[GrammarSpec]
{
  return if null(l) then [] else head(l).rootSpec.pSpecs ++ mergeProductions(tail(l));
}


function defineProductions
String ::= c::Integer ps :: [Decorated ProductionSpec] aSpec :: [Decorated ProductionSpec] ext::Boolean{

  local attribute numChild :: Integer;
  numChild = length(head(ps).productionNames);

  local attribute aspects :: [Decorated ProductionSpec];
  aspects = findAspectsForProduction(head(ps).productionName, aSpec);

  local attribute numAspect :: Integer;
  numAspect = length(aspects);

  return if null(ps) then "\n" 
		     else (if ext 
			  then
			   "extern union Node * p_" ++ makeName(head(ps).productionName) ++ 
			     "(" ++ makeProductionSignature(numChild-1, 1) ++ ");\n" ++
			   "extern int P_" ++ makeName(head(ps).productionName) ++ ";\n" ++
			   "extern int P_" ++ makeName(head(ps).productionName) ++ "_SIZE;\n" ++
			   "extern int A_" ++ makeName(head(ps).productionName) ++ "_COUNT;\n"
		          else
			   "int P_" ++ makeName(head(ps).productionName) ++ " = " ++ toString(c) ++ ";\n" ++
			   "int P_" ++ makeName(head(ps).productionName) ++ "_SIZE = " ++ toString(numChild) ++ ";\n" ++
			   "int A_" ++ makeName(head(ps).productionName) ++ "_COUNT = " ++ toString(numAspect) ++ ";\n") ++
			  defineAspects(0, aspects, ext) ++ "\n\n" ++
			  defineProductions(c+1, tail(ps), aSpec, ext);
}

function defineAspects
String ::= c::Integer aspects::[Decorated ProductionSpec] ext::Boolean {

  local attribute an :: String;
  an = makeName(head(aspects).aspectName);

  return if null(aspects) 
	 then "\n" 
	 else if ext
	      then
	      "extern union Node * a_" ++ an ++ "(union Node *);\n" ++
	      "extern int A_" ++ an ++ ";\n" ++ 
	      defineAspects(c + 1, tail(aspects), ext)
	      else
	      "int A_" ++ an ++ " = " ++ toString(c) ++ ";\n" ++ 
	      defineAspects(c + 1, tail(aspects), ext);
}

function findAspectsForProduction
[Decorated ProductionSpec] ::= pn::String aSpec::[Decorated ProductionSpec]{
  return if null(aSpec) 
	 then  []
	 else (if pn == head(aSpec).productionName then [head(aSpec)] else []) +++ findAspectsForProduction(pn, tail(aSpec));
}

function defineAttributes
String ::= e::Decorated Env ext::Boolean{

  local attribute nts :: [Decorated EnvItem];
  nts = toItems(filterEnv(keepNT(), e));

  return defineAttributesHelp(nts, e, ext);
}

function defineAttributesHelp
String ::= nts::[Decorated EnvItem] e::Decorated Env ext::Boolean{

  return if null(nts) 
	 then "\n" 
 	 else (defineAttribute(head(nts).itemName, e, ext) ++ defineAttributesHelp(tail(nts), e, ext));
}

function defineAttribute
String ::= nt::String e::Decorated Env ext::Boolean{

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(nt, e);

  return (if ext then "extern int " ++ "NUM_" ++ makeName(nt) ++ "_ATTS;\n"
 		 else "int " ++ "NUM_" ++ makeName(nt) ++ "_ATTS = " ++ toString(length(decorating)) ++ ";\n") ++
	 defineAttributeHelp(0, nt, e, decorating, ext);
}

function defineAttributeHelp
String ::= c::Integer nt::String e::Decorated Env es::[Decorated EnvItem] ext::Boolean{

  local attribute an :: String;
  an = head(es).itemName;
  
  return if null(es) 
	 then "\n" 
	 else (if ext 
		then ("extern int " ++ (if isSynthesized(an, e) then "SYN_" else "INH_") ++ makeName(nt) ++ "_" ++ 
			makeName(an) ++ ";\n")
		else ("int " ++ (if isSynthesized(an, e) then "SYN_" else "INH_") ++ makeName(nt) ++ "_" ++ 
			makeName(an) ++ " = " ++ toString(c) ++ ";\n")) ++
		defineAttributeHelp(c+1, nt, e, tail(es), ext);
}

abstract production keepNT
top::EnvFilter ::=
{
  top.keep = top.inEnvItem.isTypeDeclaration &&  top.inEnvItem.typerep.isNonTerminal;
}



function defineFunctionsForAttributesOnProds
String ::= ps::[Decorated ProductionSpec] e::Decorated Env {

  return if null(ps) then "\n" 
	 else defineFunctionsForAttributesOnProd(head(ps), e) ++ 
	      defineFunctionsForAttributesOnProds(tail(ps), e);
}

function defineFunctionsForAttributesOnProd
String ::= p::Decorated ProductionSpec e::Decorated Env {

  return defineFunctionsForAttributesOnProdHelp(p.productionName, 0, p.productionTypes, e);
}

function defineFunctionsForAttributesOnProdHelp
String ::= p::String c::Integer cs::[Decorated TypeRep] e::Decorated Env {

  return if null(cs) 
	 then "\n" 
	 else defineFunctionsForAttributesOnChild(p, c, head(cs), e) ++ 
   	      defineFunctionsForAttributesOnProdHelp(p, c+1, tail(cs), e);
}

function defineFunctionsForAttributesOnChild
--production name, child number, child type, env
String ::= p::String c::Integer t::Decorated TypeRep e::Decorated Env {

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(t.typeName, e);

  return defineFunctionsForAttributesOnChildHelp(p, c, decorating, e);
}

function defineFunctionsForAttributesOnChildHelp
--production name, child number, child atts, env
String ::= p::String c::Integer atts::[Decorated EnvItem] e::Decorated Env {
  local attribute isSyn :: Boolean;
  isSyn = isSynthesized(head(atts).itemName, e);

  return if null(atts) 
	 then "\n" 
	 else (if (c == 0 && isSyn) || (c > 0 && !isSyn) then
	       "extern struct Value get_" ++ makeName(p) ++ "_" ++ toString(c) ++ "_" ++ makeName(head(atts).itemName) ++ "(union Node * context);\n"
	       else "" ) ++
	      defineFunctionsForAttributesOnChildHelp(p, c, tail(atts), e);
}






function defineInitForAttributesOnProds
String ::= ps::[Decorated ProductionSpec] aSpec::[Decorated ProductionSpec] e::Decorated Env {

  return if null(ps) then "\n" 
	 else defineInitForAttributesOnProd(head(ps), aSpec, e) ++ 
	      defineInitForAttributesOnProds(tail(ps), aSpec, e);
}

function defineInitForAttributesOnProd
String ::= p::Decorated ProductionSpec aSpec::[Decorated ProductionSpec] e::Decorated Env {

  local attribute aspects :: [Decorated ProductionSpec];
  aspects = findAspectsForProduction(p.productionName, aSpec);

  return "  func_table[P_" ++ makeName(p.productionName) ++ "] = malloc(" ++ "P_" ++ makeName(p.productionName) ++ "_SIZE * sizeof(struct FUNC_TYPE));\n" ++
	 defineInitForAttributesOnProdHelp(p.productionName, 0, p.productionTypes, e) ++
	 "  aspect_table[P_" ++ makeName(p.productionName) ++ "] = malloc(" ++ "A_" ++ makeName(p.productionName) ++ "_COUNT * sizeof(struct ASPECT_TYPE));\n" ++
	 defineInitForAspectsOnProd(aspects);
}

function defineInitForAspectsOnProd
String ::= aspects::[Decorated ProductionSpec] {
  local attribute an :: String;
  an = makeName(head(aspects).aspectName);

  local attribute pn :: String;
  pn = makeName(head(aspects).productionName);

  return if null(aspects) 
	 then "\n" 
	 else "  aspect_table[P_" ++ pn ++ "][A_" ++ an ++ "].func = a_" ++ an ++ ";\n" ++
  	      defineInitForAspectsOnProd(tail(aspects));
}

function defineInitForAttributesOnProdHelp
String ::= p::String c::Integer cs::[Decorated TypeRep] e::Decorated Env {

  return if null(cs) 
	 then "\n" 
	 else defineInitForAttributesOnChild(p, c, head(cs), e) ++ 
   	      defineInitForAttributesOnProdHelp(p, c+1, tail(cs), e);
}

function defineInitForAttributesOnChild
--production name, child number, child type, env
String ::= p::String c::Integer t::Decorated TypeRep e::Decorated Env {

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(t.typeName, e);

  return 
if t.isNonTerminal 
then 
"  func_table[P_" ++ makeName(p) ++ "][" ++ toString(c) ++ "] = malloc(NUM_" ++ makeName(t.typeName) ++ "_ATTS * sizeof(struct FUNC_TYPE));\n" ++
     defineInitForAttributesOnChildHelp(p, c, t.typeName, decorating, e) 
else 
"";
}

function defineInitForAttributesOnChildHelp
--production name, child number, child type, atts, env
String ::= p::String c::Integer nt::String  atts::[Decorated EnvItem] e::Decorated Env {

  local attribute isSyn :: Boolean;
  isSyn = isSynthesized(head(atts).itemName, e);

  local attribute an ::String;
  an = makeName(head(atts).itemName);

  return if null(atts) 
	 then "\n" 
	 else (if (c == 0 && isSyn) || (c > 0 && !isSyn) then
	       "  func_table[P_" ++ makeName(p)++ "][" ++ toString(c)++ "][" ++ (if isSyn then "SYN_" else "INH_") ++ makeName(nt) ++ "_" ++ an ++ "].func = " ++
			"get_" ++ makeName(p) ++ "_" ++ toString(c) ++ "_" ++ an ++ ";\n"
	       else "" ) ++
	      defineInitForAttributesOnChildHelp(p, c, nt, tail(atts), e);
}



