grammar silver:translation:c:core;
import silver:base;
import silver:utils;

attribute translation occurs on Expr, Exprs;

--Base Expressions
aspect production nestedExpr
top::Expr ::= l::LParen_t e::Expr r::RParen_t
{
  top.translation = "(" ++ e.translation ++ ")";
}

aspect production productionReference
top::Expr ::= q::QName
{
  top.translation = "p_" ++ makeName(fName);
}

aspect production dontDecorateExpr
top::Expr ::= q::QName
{
  local attribute i :: Integer;
  i = if in_sig then findIndex(fName, top.signatureNames) else findIndex(fName, top.localNames);
 
  top.translation =  if in_sig 
		     then (if i == 0 then "context" else getValue(top.typerep, "(context->nt.children[" ++ toString(i-1) ++ "])"))
		     else (getValue(top.typerep, "(context->nt.locals[A_" ++ makeName(top.aName) ++ "][" ++ toString(i) ++ "])"));
}

aspect production decorateExpr
top::Expr ::= q::QName
{

  local attribute i :: Integer;
  i = if in_sig then findIndex(fName, top.signatureNames) else findIndex(fName, top.localNames);
 
  top.translation =  if in_sig 
		     then (if i == 0 then "context" else getValue(top.typerep, "(context->nt.children[" ++ toString(i-1) ++ "])"))
		     else (getValue(top.typerep, "(context->nt.locals[A_" ++ makeName(top.aName) ++ "][" ++ toString(i) ++ "])"));
}

function findIndex
Integer ::= s::String ss::[String]{
  return findIndexHelp(0, s, ss);
}

function findIndexHelp
Integer ::= i::Integer s::String ss::[String]{
  return if null(ss) then -1 else if s == head(ss) then i else findIndexHelp(i+1, s, tail(ss));
}

function getValue
String ::= t::Decorated TypeRep s::String {

  local attribute tn :: String ;
  tn = t.typeName;
  return if tn == "Integer"
	 then s ++ ".i"
	 else if tn == "Float"
	 then s ++ ".f"
	 else if tn == "String"
	 then s ++ ".s"
	 else if tn == "Boolean"
	 then s ++ ".i"
	 else if t.isNonTerminal || t.isTerminal || t.isDecorated
	 then s ++ ".n"
	 else s;
}

function makeValue
String ::= e::Decorated Expr {

  local attribute t :: Decorated TypeRep;
  t = e.typerep;

  local attribute tn :: String;
  tn = e.typerep.typeName;

  return if tn == "Integer"
	 then "mkInt(" ++ e.translation ++ ")"
	 else if tn == "Float"
	 then "mkFloat(" ++ e.translation ++ ")"
	 else if tn == "String"
	 then "mkStr(" ++ e.translation ++ ")"
	 else if tn == "Boolean"
	 then "mkInt(" ++ e.translation ++ ")"
	 else if t.isNonTerminal
	 then "mkNode(" ++ e.translation ++ ")"
	 else if t.isDecorated || t.isTerminal
	 then "mkNode(" ++ e.translation ++ ")"
	 else e.translation;
}


aspect production genericApplicationDispatcher
top::Expr ::= e::Expr es::Exprs
{
  top.translation = e.translation ++ "(" ++ es.translation ++ ")";
}

aspect production atAccess
top::Expr ::= e::Expr d::At_t q::QName
{
  local attribute tn :: String;
  tn = makeName(e.typerep.typeName);

  local attribute syn :: Boolean;
  syn = isSynthesized(fName, top.env);

  local attribute attr_str :: String;
  attr_str = (if syn then "SYN" else "INH") ++ "_" ++  tn ++ "_"  ++ makeName(fName);

  top.translation = getValue(top.typerep, (if syn then "getSynAttribute" else "getInhAttribute") ++ "(" ++ e.translation ++ ", " ++ attr_str ++")");
}

aspect production trueConst
top::Expr ::= t::True_kwd
{
  top.translation = "TRUE";
}

aspect production falseConst
top::Expr ::= f::False_kwd
{
  top.translation = "FALSE";
}

aspect production and
top::Expr ::= e1::Expr a::And_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " && " ++ e2.translation ++ ")";
}

aspect production or
top::Expr ::= e1::Expr o::Or_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " || " ++ e2.translation ++ ")";
}

aspect production not
top::Expr ::= n::Not_t e::Expr
{
  top.translation = "(TRUE != " ++ e.translation ++ ")";
}

aspect production gt
top::Expr ::= e1::Expr g::GT_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " > " ++ e2.translation ++ ")";
}

aspect production lt
top::Expr ::= e1::Expr l::LT_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " < " ++ e2.translation ++ ")";
}


aspect production gteq
top::Expr ::= e1::Expr g::GTEQ_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " >= " ++ e2.translation ++ ")";
}


aspect production lteq
top::Expr ::= e1::Expr l::LTEQ_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " <= " ++ e2.translation ++ ")";
}


aspect production eqeq
top::Expr ::= e1::Expr eq::EQEQ_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " == " ++ e2.translation ++ ")";
}


aspect production neq
top::Expr ::= e1::Expr n::NEQ_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " != " ++ e2.translation ++ ")";
}


aspect production ifThenElse
top::Expr ::= i::If_kwd e1::Expr t::Then_kwd e2::Expr e::Else_kwd e3::Expr
{
  top.translation = "fail(\"NOT DEFINED!!\")";
}

aspect production intConst
top::Expr ::= i::Int_t
{
  top.translation = "(" ++ i.lexeme ++ ")";
}

aspect production floatConst
top::Expr ::= f::Float_t
{
  top.translation = "(" ++ f.lexeme ++ ")";
} 

aspect production plus
top::Expr ::= e1::Expr p::Plus_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " + " ++ e2.translation ++ ")";
}
aspect production minus
top::Expr ::= e1::Expr m::Minus_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " - " ++ e2.translation ++ ")";
}
aspect production multiply
top::Expr ::= e1::Expr m::Multiply_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " * " ++ e2.translation ++ ")";
}
aspect production divide
top::Expr ::= e1::Expr d::Divide_t e2::Expr
{
  top.translation = "(" ++ e1.translation ++ " / " ++ e2.translation ++ ")";
}
aspect production neg
top::Expr ::= m::Minus_t e::Expr
{
  top.translation = "(" ++ " - " ++ e.translation ++ ")";
}

aspect production stringConst
top::Expr ::= s::String_t
{
  top.translation = "(" ++ s.lexeme ++ ")";
}

aspect production defaultPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.translation = "fail(\"NOT DEFINED!!\")";
}

aspect production stringPlusPlus
top::Expr ::= e1::Expr e2::Expr
{
  top.translation = "fail(\"NOT DEFINED!!\")";
}

aspect production exprsEmpty
top::Exprs ::=
{
  top.translation = "";
}

aspect production exprsSingle
top::Exprs ::= e::Expr
{
  top.translation = makeValue(e);
}

aspect production exprsCons
top::Exprs ::= e1::Expr c::Comma_t e2::Exprs
{
  top.translation = makeValue(e1) ++ (if e2.translation == "" then "" else ", " ++ e2.translation);
}

aspect production exprWith
top::Expr ::= ev::Decorate_kwd e::Expr w::With_kwd l::LCurly_t inh::ExprInh r::RCurly_t
{
  top.translation = "fail(\"NOT DEFINED!!\")";
}

attribute translation occurs on ExprInh;
aspect production exprStmtsEmpty
top::ExprInh ::= 
{
  top.translation = "fail(\"NOT DEFINED!!\")";
}

aspect production exprStmtsCons
top::ExprInh ::= lhs::ExprLHSExpr ekwd::Equal_t e::Expr s::Semi_t inh::ExprInh
{ 
  top.translation = "fail(\"NOT DEFINED!!\")";
}

attribute translation occurs on ExprLHSExpr;
aspect production exprLhsExpr
top::ExprLHSExpr ::= q::QName
{
  top.translation = "fail(\"NOT DEFINED!!\")";
}


