grammar silver:translation:c:core;
import silver:base;
import silver:utils;

attribute pSpecs occurs on AGDcl;

autocopy attribute localNames :: [String];
attribute localNames occurs on ProductionBody, ProductionStmts, ProductionStmt, AttributeDef, LHSExpr, Expr, Exprs;


autocopy attribute aName :: String;
attribute aName occurs on ProductionBody, ProductionStmts, ProductionStmt, AttributeDef, LHSExpr, Expr, Exprs;
         
aspect production productionDcl
 top::AGDcl ::=
   a::Abstract_kwd p::Production_kwd id::Id_t ns::ProductionSignature
   body::ProductionBody
{
  local attribute an :: String;
  an = fName ++ toString(genint());

  top.pSpecs = [productionSpec(fName, ns.sigNames, ns.sigTypes), aspectSpec(an, fName, lNames, lTypes)];

  local attribute all_stmts :: [StmtRep];
  all_stmts = body.stmtList;

  local attribute l_dcls :: [StmtRep];
  l_dcls = getLocalDeclarations(all_stmts);

  local attribute lNames :: [String];
  lNames =  filterNames(l_dcls);

  body.localNames = lNames;
  body.aName = an;

  local attribute lTypes :: [Decorated TypeRep];
  lTypes =  filterTypes(l_dcls);

  local attribute numChild :: Integer;
  numChild = length(ns.sigNames) - 1;

  local attribute pn :: String;
  pn = makeName(fName);

  local attribute tn :: String;
  tn = makeName(head(ns.sigTypes).typeName);

  local attribute f_dcls :: [StmtRep];
  f_dcls = getForwardDeclarations(all_stmts);

  top.translation = 
"union Node * p_" ++ pn ++ "(" ++ makeProductionSignature(numChild, 1) ++ "){\n" ++ 
"  union Node * this = malloc(sizeof(union Node));\n" ++ 
"  this->nt.production = P_" ++ pn ++ ";\n" ++ 
"  this->nt.values = malloc(NUM_" ++ tn ++ "_ATTS * sizeof(struct Value));\n\n" ++ 

"  int i;\n" ++ 
"  for(i = 0; i < NUM_" ++ tn ++ "_ATTS; i++){\n" ++ 
"    this->nt.values[i].defined = FALSE;\n" ++ 
"  }\n\n" ++ 

"  this->nt.children = malloc(" ++ toString(numChild)++ " * sizeof(struct Value));\n\n" ++ 
"  this->nt.locals = malloc(A_" ++ makeName(fName) ++ "_COUNT * sizeof(struct Value));\n\n" ++ 

(
if null(f_dcls) 
then
"  this->nt.forward = 0;\n\n"
else
"  this->nt.forward = 0;\n\n" --TODO should be the rhsC from the head(f_dcls)
) ++
makeChildrenInit(tail(ns.sigTypes), 1) ++
"\n" ++

"  for(i=0; i < A_" ++ pn ++ "_COUNT; i++){\n" ++
"    aspect_table[P_" ++ pn ++ "][i].func(this);\n" ++
"  }\n\n" ++

"  return this;\n" ++ 
"}\n\n" ++
makeChildrenDefs(0, fName, [head(ns.sigNames)], [head(ns.sigTypes)], all_stmts, ns.defs, top.globalEnv, true, false) ++ 
makeChildrenDefs(1, fName, tail(ns.sigNames), tail(ns.sigTypes), all_stmts, ns.defs, top.globalEnv, false, false) ++
"union Node * a_" ++ makeName(an) ++ "(union Node * context){\n" ++
(
if ! null(l_dcls) then 
"  context->nt.locals[A_" ++ makeName(an) ++ "] = malloc(" ++ toString(length(l_dcls)) ++ " * sizeof(struct Value));\n\n" ++ 
makeLocalInit(an, l_dcls, 0, all_stmts, fName) 
else ""
) ++

"  return context;\n" ++
"}\n\n";

}



function filterNames
[String] ::= s::[StmtRep] {
  return if null(s) then [] else [head(s).name] ++ filterNames(tail(s));
}

function filterTypes
[Decorated TypeRep] ::= s::[StmtRep] {
  return if null(s) then [] else [head(s).typerep] ++ filterTypes(tail(s));
}

function makeForwardDef
String ::= f::StmtRep  {
  return "";
}
function makeChildrenInit
String ::= t::[Decorated TypeRep] c::Integer {
  local attribute cn :: String;
  cn = "c" ++ toString(c);

  return if null(t) then "" else    "  this->nt.children [" ++ toString(c-1) ++ "] = " ++ cn ++ ";\n" ++ 
				  (if head(t).isNonTerminal
				   then
				   "  " ++ cn ++ ".n->nt.parent = this;\n" ++ 
				   "  " ++ cn ++ ".n->nt.child = " ++ toString(c) ++ ";\n\n"
 				   else "") ++
	                          makeChildrenInit(tail(t), c+1);
}

function makeProductionSignature
String ::= m::Integer i::Integer {
  return if i > m then "" else "struct Value c" ++ toString(i) ++ (if i == m then "" else ", " ++ makeProductionSignature(m, i+1));
}

function makeChildrenDefs
String ::= cn::Integer pname::String childrenN::[String] childrenT::[Decorated TypeRep] stmts::[StmtRep] sig::Decorated Env e::Decorated Env syn::Boolean ex::Boolean
{
  local attribute fns :: [Decorated EnvItem];
  fns = getFullNameDcl(head(childrenN), sig);

  local attribute fn :: String;
  fn = if !null(fns) then head(fns).fullName else "FULL_NAME_NOT_FOUND";

  local attribute inhs :: [StmtRep];
  inhs = getDefsForName(fn, stmts);

  local attribute decorating :: [Decorated EnvItem];
  decorating = getAllOccuringOn(head(childrenT).typeName, e);

  local attribute extra :: String;
  extra = makeExtraDefs(pname, head(childrenN), cn, inhs, decorating);

  return if null(childrenT) then "" else 
	 makeChildrenDefsHelp(cn, pname, head(childrenT), inhs, syn) ++ (if ex then extra else "") ++ 
 	 makeChildrenDefs(cn+1, pname, tail(childrenN), tail(childrenT), stmts, sig, e, syn, ex);
}

function makeChildrenDefsHelp
String ::= cn::Integer pname::String childT::Decorated TypeRep stmts::[StmtRep] syn::Boolean
{
  local attribute an :: String;
  an = makeName(head(stmts).lhsExpr.secondAccessFull);

  local attribute child_const :: String;
  child_const = (if syn then "SYN_" else "INH_") ++ makeName(childT.typeName) ++ "_" ++ an;

  local attribute cString :: String ;
  cString = 
	"struct Value get_" ++ makeName(pname) ++ "_" ++ toString(cn) ++ "_" ++ an ++ "(union Node * c){\n" ++
       	"  union Node * context = " ++ (if syn then "c" else "c->nt.parent") ++ ";\n" ++
	"  if(c->nt.values[" ++ child_const ++ "].defined == FALSE){\n" ++
	"    c->nt.values[" ++ child_const ++ "] = " ++ makeValue(head(stmts).rhsExpr) ++ ";\n"++
	"    c->nt.values[" ++ child_const ++ "].defined = TRUE;\n" ++
	"  }\n" ++
	"  return c->nt.values[" ++ child_const ++ "];\n" ++
	"}\n\n";

  return if null(stmts) then "" else 
	cString ++ makeChildrenDefsHelp(cn, pname, childT, tail(stmts), syn);
}

function makeExtraDefs
String ::= pname::String cname::String cn::Integer defined::[StmtRep] all::[Decorated EnvItem]
{
  local attribute an :: String;
  an =  makeName(head(all).itemName);

  local attribute recurse :: String;
  recurse = makeExtraDefs(pname, cname, cn, defined, tail(all));

  local attribute cString :: String ;
  cString = 
	"struct Value get_" ++ makeName(pname) ++ "_" ++ toString(cn) ++ "_" ++ an ++ "(union Node * c){\n" ++
	"  printf( \"Error: Attribute '" ++ cname ++ "." ++ head(all).itemName ++ "' is not defined in production '" ++ pname ++ "'\");" ++
	"  exit(1);\n" ++
	"}\n\n";

  return
	if null(all) then ""
        else if containsAttrDef(head(all).itemName, defined) 
	     then recurse                 
             else (cString ++ recurse);
}

function containsAttrDef
Boolean ::= attr::String stmts::[StmtRep]
{
  local attribute fname :: String;
  fname = head(stmts).lhsExpr.secondAccessFull;

  return
	if null(stmts) then false
	else if head(stmts).isAttributeDef && fname == attr then true
	else containsAttrDef (attr, tail(stmts));
}

aspect production aspectAGDcl
 top::AGDcl ::=
   a::Aspect_kwd p::Production_kwd id::QName ns::AspectSignature
   body::ProductionBody
{
  local attribute an :: String;
  an = fName ++ toString(genint());

  top.pSpecs = [aspectSpec(an, fName, lNames, lTypes)];

  local attribute all_stmts :: [StmtRep];
  all_stmts = body.stmtList;

  local attribute l_dcls :: [StmtRep];
  l_dcls = getLocalDeclarations(all_stmts);

  local attribute lNames :: [String];
  lNames =  filterNames(l_dcls);

  body.localNames = lNames;
  body.aName = an;

  local attribute lTypes :: [Decorated TypeRep];
  lTypes =  filterTypes(l_dcls);

  top.translation = 
	makeChildrenDefs(0, fName, [head(ns.sigNames)], [head(ns.sigTypes)], all_stmts, ns.defs, top.globalEnv, true, false) ++ 
	makeChildrenDefs(1, fName, tail(ns.sigNames), tail(ns.sigTypes), all_stmts, ns.defs, top.globalEnv, false, false) ++ 
"union Node * a_" ++ makeName(an) ++ "(union Node * context){\n" ++
(
if ! null(l_dcls) then 
"  context->nt.locals[A_" ++ makeName(an) ++ "] = malloc(" ++ toString(length(l_dcls)) ++ " * sizeof(struct Value));\n\n" ++ 
makeLocalInit(an, l_dcls, 0, all_stmts, fName) 
else ""
) ++
"  return context;\n" ++
"}\n\n";

}

function makeLocalInit
String ::= an::String t::[StmtRep] c::Integer dcls::[StmtRep] pname::String {

  local attribute localPosition::String;
  localPosition = "  context->nt.locals [A_" ++ makeName(an) ++ "][" ++ toString(c) ++ "]";

  local attribute l_def :: [StmtRep];
  l_def = getLocalDefinition(head(t).fullName, dcls);

  local attribute l_def_str :: String;
  l_def_str = if null(l_def)
	      then " = fail(\"In production '" ++ pname ++ "': Local attribute '" 
			 ++ head(t).name ++ "' is declared but not not defined.\")\n" 
	      else " = " ++ makeValue(head(l_def).rhsExpr) ++ ";\n";

  return if null(t) then "" else    localPosition ++ l_def_str ++ 
				  (if head(t).typerep.isNonTerminal
				   then
				   localPosition ++ ".n->nt.parent = context;\n" ++ 
				   localPosition ++ ".n->nt.child = " ++ toString(c) ++ ";\n\n"
 				   else "") ++
	                          makeLocalInit(an, tail(t), c+1, dcls, pname);
}