grammar silver:translation:haskellc:lexyacc;
import silver:translation:haskellc:core;
import silver:base hiding grammarName;
import silver:translation:ag:haskell:core;
import core;
import silver:utils;
import silver:utils:command;

aspect production run
top::RunUnit ::= iIn::IO args::String p::Production(Root ::= String)
{
  postOps <- if a.noCSGen then [] else [generateCS(a'', grammars)]; 
}

abstract production generateCS
top::Unit ::= a::Command specs::[GrammarSpec]
{
  forwards to generateCSHelp(a'', specs)
	with {
		ioIn = print("Generating Parsers and Scanners.\n", top.ioIn);
	};
}

abstract production generateCSHelp
top::Unit ::= a::Command specs::[GrammarSpec]
{
  top.io = writeAllCS(top.ioIn, a, specs, mergeParsers(specs));
  top.code = 0;
  top.order = 2;
}

function writeAllCS
IO ::= i::IO a::Decorated Command l::[GrammarSpec] s::[String]
{
  local attribute now :: IO;
  now = writeCSSpec(i, head(l).name, head(l).rootSpec, a, s);

  local attribute recurse :: IO;
  recurse = writeAllCS(now, a, tail(l), s);

  return if null(l) then i else recurse;
}


function writeCSSpec
IO ::= i::IO gName::String r::Decorated RootSpec a::Decorated Command s::[String]
{
  --the top level grammar name
  local attribute thn :: String;
  thn = makeName(a.grammarName);

  -- the haskell name of the grammar.
  local attribute hn :: String;
  hn = makeName(gName);
  
  local attribute has_start :: Boolean;
  has_start = r.hasStartSymbol && containsString(gName ++ ":parse", s);

  local attribute shn :: String;
  shn = "NT_" ++ makeName(r.startName);

  local attribute parser_boot :: String;
  parser_boot = 
      "module Module_" ++ hn ++ "_Parser where\n" ++
      "import {-# SOURCE #-} Module_" ++ thn ++ "\n" ++
      "doparse :: String ->  (INHS_" ++ shn ++ " -> " ++ shn ++ ")\n";


  local attribute lexFile :: String;
  lexFile = a.generatedPath ++ "Module_" ++ hn ++ "_Scanner.x";

  local attribute lexFileTemp :: String;
  lexFileTemp = a.generatedPath ++ "Module_" ++ hn ++ "_Scanner.x.temp";

  local attribute existsLex :: IOBoolean;
  existsLex = isFile(lexFile, i);

  local attribute cpLex :: IOInteger;
  cpLex = if existsLex.bValue then system("cp " ++ lexFile ++ " " ++ lexFileTemp, existsLex.io) else ioInteger(existsLex.io, 0);

  local attribute prevLex :: IOString;
  prevLex = if existsLex.bValue then readFile(lexFileTemp, cpLex.io) else ioString(cpLex.io, "");

  local attribute newLex :: String;
  newLex = makeLexSpec(hn, r);

  local attribute needsLex ::Boolean;
  needsLex = prevLex.sValue != newLex;

  local attribute io2 :: IO;
  io2 = if !has_start || !needsLex then prevLex.io
	else writeFile(lexFile, newLex, prevLex.io);

  local attribute io3 :: IO;
  io3 = if !has_start || !needsLex then io2
	else system("flex -o " ++ a.generatedPath ++ "Module_" ++ hn ++ "_Scanner.c " ++ lexFile, io2).io;



  local attribute yaccFile :: String;
  yaccFile = a.generatedPath ++ "Module_" ++ hn ++ "_Parser.y";

  local attribute yaccFileTemp :: String;
  yaccFileTemp = a.generatedPath ++ "Module_" ++ hn ++ "_Parser.y.temp";

  local attribute existsYacc :: IOBoolean;
  existsYacc = isFile(yaccFile, io3);

  local attribute cpYacc :: IOInteger;
  cpYacc = if existsYacc.bValue then system("cp " ++ yaccFile ++ " " ++ yaccFileTemp, existsYacc.io) else ioInteger(existsYacc.io, 0);

  local attribute prevYacc :: IOString;
  prevYacc = if existsYacc.bValue then readFile(yaccFileTemp, cpYacc.io) else ioString(cpYacc.io, "");

  local attribute newYacc :: String;
  newYacc = makeYaccSpec(hn, r, a.grammarName);

  local attribute needsYacc ::Boolean;
  needsYacc = prevYacc.sValue != newYacc;

  local attribute io4 :: IO;
  io4 = if !has_start || !needsYacc then prevYacc.io
	else writeFile(a.generatedPath ++ "Module_" ++ hn ++ "_Parser.hs-boot", parser_boot, prevYacc.io);

  local attribute io5 :: IO;
  io5 = if !has_start || !needsYacc then io4
	else writeFile(yaccFile, newYacc, io4);

  local attribute io6 :: IO;
  io6 = if !has_start || !needsYacc then io5
	else system("bison -d -v -o " ++ a.generatedPath ++ "Module_" ++ hn ++ "_Parser.c " ++ yaccFile, io5).io;

  local attribute io7 :: IO;
  io7 = if !has_start || !needsYacc then io6
	else writeFile(a.generatedPath ++ "Module_" ++ hn ++ "_Parser.hs", makeBridge(gName, r, a.grammarName), io6);

  return io7;
}
