grammar silver:translation:haskelljava:concretesyntax;

import silver:base:cs:default;
import silver:base:core;
import silver:analysis:typechecking;

terminal CppExtract_kwd /extractCPPInfo/ lexer precedence = 10;
terminal CppUpdate_kwd /updateCPPInfo/ lexer precedence = 10;

function checkFormerDcls
String ::= dclNames::[String] topEnv::Decorated Env loc::String
{
  local attribute er1 :: String;
  er1 = if length(getFullNameDcl(head(dclNames), topEnv)) > 1
       then loc ++ "Error: Value '" ++ head(dclNames) ++ "' is already bound.\n"
       else "";	

  local attribute er2 :: String;
  er2 = if length(getValueDcl(head(dclNames), topEnv)) > 1
       then loc ++ "Error: Value '" ++ head(dclNames) ++ "' is already bound.\n"
       else "";	
 
  return if null(dclNames)
         then ""
         else er1 ++ er2 ++ checkFormerDcls(tail(dclNames),topEnv,loc);
}

concrete production cppExtractDef
top::ProductionStmt ::= c1::CppExtract_kwd s::Semi_t
{
   top.hasValidActionCode = top.actionCodeType.isTerminalAction;
   top.actionCode = if top.hasValidActionCode
                    then actionCodeCode("String cpp_filename = \"\";\n" ++
   "int cpp_line = -1;\n" ++
   "int cpp_column = 0;\n" ++
   "boolean failure = false;\n" ++
   "if(lexeme.length() < 1) failure = true;\n" ++
   "java.util.Scanner scanner = new java.util.Scanner(lexeme.substring(1));\n" ++
   "if(!scanner.hasNextInt()) failure = true;\n" ++
   "if(!failure) cpp_line = scanner.nextInt() - 1;\n" ++
   "if(!failure && !scanner.hasNext()) failure = true;\n" ++
   "if(!failure) cpp_filename = scanner.next();\n" ++
   "if(cpp_filename.length() < 2) failure = true;\n" ++
   "if(!failure) cpp_filename = cpp_filename.substring(1,cpp_filename.length() - 1);\n" ++
   "if(failure) error(_pos,\"Malformed CPP directive '\" + lexeme + \"'\");\n")
                   else actionCodeError("Cannot use 'extractCPPInfo' in any but terminal actions");
   top.defs = addFullNameDcl("cpp_filename","cpp_filename",
               addValueDcl("cpp_filename",stringTypeRep(),
                addFullNameDcl("cpp_line","cpp_line",
                 addValueDcl("cpp_line",integerTypeRep(),
                  addFullNameDcl("cpp_column","cpp_column",
                   addValueDcl("cpp_column",integerTypeRep(),
                    emptyDefs()))))));

  top.errors = checkFormerDcls(["cpp_filename","cpp_line","cpp_column"],top.env,top.location);
  top.typeErrors = "";
  top.stmtList = [localDclStmtRep("cpp_filename", "cpp_filename", stringTypeRep()),
                  localDclStmtRep("cpp_line", "cpp_line", stringTypeRep()),
                  localDclStmtRep("cpp_column", "cpp_column", stringTypeRep())];
}

concrete production cppUpdateDef
top::ProductionStmt ::= c1::CppUpdate_kwd s::Semi_t
{
   top.hasValidActionCode = top.actionCodeType.isTerminalAction;
   top.actionCode = if top.hasValidActionCode
                    then actionCodeCode("String cpp_filename = \"\";\n" ++
   "int cpp_line = -1;\n" ++
   "int cpp_column = 0;\n" ++
   "boolean failure = false;\n" ++
   "if(lexeme.length() < 1) failure = true;\n" ++
   "java.util.Scanner scanner = new java.util.Scanner(lexeme.substring(1));\n" ++
   "if(!scanner.hasNextInt()) failure = true;\n" ++
   "if(!failure) cpp_line = scanner.nextInt() - 1;\n" ++
   "if(!failure && !scanner.hasNext()) failure = true;\n" ++
   "if(!failure) cpp_filename = scanner.next();\n" ++
   "if(cpp_filename.length() < 2) failure = true;\n" ++
   "if(!failure) cpp_filename = cpp_filename.substring(1,cpp_filename.length() - 1);\n" ++
   "if(failure) error(_pos,\"Malformed CPP directive '\" + lexeme + \"'\");\n" ++
    "getVirtualLocation().setFileName(cpp_filename);\n" ++
    "getVirtualLocation().setLine(cpp_line);\n" ++
    "getVirtualLocation().setColumn(0);\n")
                   else actionCodeError("Cannot use 'updateCPPInfo' in any but terminal actions");
   top.defs = addFullNameDcl("cpp_filename","cpp_filename",
               addValueDcl("cpp_filename",stringTypeRep(),
                addFullNameDcl("cpp_line","cpp_line",
                 addValueDcl("cpp_line",integerTypeRep(),
                  addFullNameDcl("cpp_column","cpp_column",
                   addValueDcl("cpp_column",integerTypeRep(),
                    emptyDefs()))))));

  top.errors = checkFormerDcls(["cpp_filename","cpp_line","cpp_column"],top.env,top.location);
  top.typeErrors = "";
  top.stmtList = [localDclStmtRep("cpp_filename", "cpp_filename", stringTypeRep()),
                  localDclStmtRep("cpp_line", "cpp_line", stringTypeRep()),
                  localDclStmtRep("cpp_column", "cpp_column", stringTypeRep())];
}
