grammar silver:utils:ale:host ;

start nonterminal Root      with pp, new_lang, new_lang_main_sv, bin_main_sv, bin_silver_compile, bin_build_test ;
nonterminal LangDef         with pp, new_lang_main_sv, new_lang, bin_main_sv, bin_silver_compile, bin_build_test ;
nonterminal Expr            with pp, new_lang_main_sv ;
nonterminal ExtensionExpr   with pp, new_lang_main_sv ;
nonterminal GrammarList     with pp, new_lang_main_sv ;
nonterminal GrammarName     with pp, new_lang_main_sv, bin_main_sv ;
nonterminal HostGrammarName with pp, new_lang_main_sv ;
nonterminal Name            with pp;

synthesized attribute pp :: String ;
synthesized attribute new_lang :: String ;
synthesized attribute new_lang_main_sv :: String ;
synthesized attribute bin_main_sv :: String ;
synthesized attribute bin_silver_compile :: String ;
synthesized attribute bin_build_test :: String ;

autocopy attribute options :: Options ;
attribute options occurs on LangDef, Expr, GrammarName, Name ;

concrete production root
r::Root ::= ld::LangDef opts::Options
{
 r.pp = ld.pp ++ "\n\n" ++ opts.pp ;
 r.new_lang = ld.new_lang ;
 r.new_lang_main_sv = ld.new_lang_main_sv ;
 r.bin_main_sv = ld.bin_main_sv ;
 r.bin_silver_compile = ld.bin_silver_compile ;
 r.bin_build_test = ld.bin_build_test ; 

 ld.options = opts ;
}

-- LangDef --
concrete production lang_def
ld::LangDef ::= n::GrammarName '=' lang::Expr
{
 ld.pp = n.pp ++ "\n =  " ++ lang.pp ;
 ld.new_lang = n.pp ;

 ld.new_lang_main_sv = warning_msg_sv
                    ++ "grammar " ++ n.pp ++ " ; \n" 
                    ++ "export " ++ n.pp ++ " ; \n\n\n" 
                    ++ lang.new_lang_main_sv
                    ++ "\n\n" ;

 ld.bin_main_sv = warning_msg_sv 
               ++ "grammar " ++ n.pp ++ ":bin ; \n" 
               ++ "export " ++ n.pp ++ ":bin ; \n\n\n" 
               ++ n.bin_main_sv
               ++ "\n\n" 
               ++ "import core ; \n\n" 
               ++ "abstract production main\n" 
               ++ "top::Main ::= args::String\n"
               ++ "{ forwards to " ++ the_driver_name ++ "(args,parse,hostParse); }\n\n" ;
 
 local attribute the_driver_name :: String ;  
 the_driver_name = if ld.options.driver_name == "" then "driver" else ld.options.driver_name ;

 ld.bin_silver_compile =  "#!/bin/bash\n"
                       ++ "#----------------------------------------------------\n"
                       ++ "#-- THIS FILE IS AUTOMATICALLY GENERATED.          --\n"
                       ++ "#-- DO NOT EDIT - ANY CHANGES WILL BE OVERWRITTEN. --\n"
                       ++ "#----------------------------------------------------\n\n" 
                       ++ "echo `which " ++ silver_to_use ++ "` \n\n"
                       ++ "time " 
                       ++ silver_to_use ++ -- " --jclasspath $SILVER_HOME/copper/CopperCompiler.jar " ++ 
                                           " --jvmparameters '\\-Xmx512m' "
                       ++ " -o " ++ executable_name 
                       ++ (if   ld.options.separate_ghc
                           then " --xghc"
                           else "")
                       ++ " " ++ n.pp ++ ":bin \n\n"

                       ++ (if   ld.options.separate_ghc
                           then    "if [ \"$?\" == \"0\" ]; then \n"
                                ++ "   echo \"Running ghc-compile.....\"\n" 
                                ++ "   cd .Generated ; chmod +x ghc-compile; time ./ghc-compile; \n"
                                ++ "fi\n\n"
                           else "")

                       ++ "echo -e \"\\a\" ; sleep 0.15 \n"
                       ++ "echo -e \"\\a\" ; sleep 0.15 \n"
                       ++ "echo -e \"\\a\" ; sleep 0.15 \n\n\n" ;

 local attribute silver_to_use :: String ;   
 silver_to_use = if ld.options.silver_name == "" then "silver.copper" else ld.options.silver_name ;

 local attribute executable_name :: String ;   
 executable_name = if ld.options.exe_name == "" then "a.out" else ld.options.exe_name ;


 ld.bin_build_test = warning_msg_sv 
                  ++ "SILVER  " ++ silver_to_use ++ " \n"
                  ++ "PARAMS   --jvmparameters '-Xmx512m' \n" 
                     -- "--jclasspath $CVS_ROOT/Copper/CopperCompiler.jar "
                  ++ "GRAMMAR " ++ n.pp ++ ":bin \n" 
                  ++ "BUILDONLY \n"
                  ++ "SAVEBUILD " ++ executable_name ++ " \n"
                  ++ "OWNER   evw@cs.umn.edu \n" 
                  ++ ld.options.build_lines 
                  ++ "\n\n\n" ;

 local attribute warning_msg_sv :: String ;
 warning_msg_sv = "----------------------------------------------------\n"
               ++ "-- THIS FILE IS AUTOMATICALLY GENERATED.          --\n"
               ++ "-- DO NOT EDIT - ANY CHANGES WILL BE OVERWRITTEN. --\n"
               ++ "----------------------------------------------------\n\n" ;
}

-- Expr --
concrete production grammar_ref
newlang::Expr ::= host::HostGrammarName
{
 newlang.pp = host.pp ;
 newlang.new_lang_main_sv = host.new_lang_main_sv ;
}

concrete production compose
newlang::Expr ::= host::HostGrammarName '+' '{' exts::GrammarList '}'
{
 newlang.pp = host.pp ++ " + { \n    " ++ exts.pp  ++ " } " ;
 newlang.new_lang_main_sv = host.new_lang_main_sv ++ "\n" ++ exts.new_lang_main_sv ;
}

concrete production compose_prev
newlang::Expr ::= host::HostGrammarName '+' exts::ExtensionExpr
{
 newlang.pp = host.pp ++ " + \n    " ++ exts.pp ;
 newlang.new_lang_main_sv = host.new_lang_main_sv ++ "\n" ++ exts.new_lang_main_sv ;
}

-- Grammar List
concrete production grammar_list_one
gl::GrammarList ::= e::GrammarName
{
 gl.pp = e.pp ;
 gl.new_lang_main_sv = e.new_lang_main_sv ;
}

concrete production grammar_list_cons
gl::GrammarList ::= e::GrammarName ',' gl_tail::GrammarList
{
 gl.pp = e.pp ++ " , \n      " ++ gl_tail.pp ;
 gl.new_lang_main_sv = e.new_lang_main_sv ++ "\n" ++ gl_tail.new_lang_main_sv ;
}

-- Sub expressions  (contains no host grammar specifications)
concrete production extension_expr_one
ee::ExtensionExpr ::= e::GrammarName
{
 ee.pp = e.pp ;
 ee.new_lang_main_sv = e.new_lang_main_sv ;
}

concrete production extension_expr_cons
ee::ExtensionExpr ::= e::GrammarName '+' sub_expr::ExtensionExpr
{
 ee.pp = e.pp ++ " + \n      " ++ sub_expr.pp ;
 ee.new_lang_main_sv = e.new_lang_main_sv  ++ "\n" ++ sub_expr.new_lang_main_sv ;
}


-- GrammarName --
concrete production grammar_name
gn::GrammarName ::= n::Name
{
 gn.pp = n.pp ;
 gn.new_lang_main_sv = "import " ++ n.pp ++ " ; \n" ++
                       "export " ++ n.pp ++ " hiding parse ; \n" ++
                       "syntax " ++ n.pp ++ " ; \n" ;

 gn.bin_main_sv 
  = "import " ++ n.pp ++ " ; \n" ++
    "syntax " ++ n.pp ++ " ; \n" ;
}

concrete production grammar_name_host_explicit
gn::HostGrammarName ::= 'host_grammar' '(' n::Name ')'
{
 gn.pp = n.pp ;
 gn.new_lang_main_sv = "import " ++ n.pp ++ " ; \n" ++
                       "export " ++ n.pp ++ " with parse as hostParse; \n" ++
                       "syntax " ++ n.pp ++ " ; \n" ;

-- gn.bin_main_sv = "import " ++ n.pp ++ " ; \n" ++
--                  "syntax " ++ n.pp ++ " ; \n" ;
}

concrete production grammar_name_host_implicit
gn::HostGrammarName ::= n::Name
{
 gn.pp = n.pp ;
 gn.new_lang_main_sv = "import " ++ n.pp ++ " ; \n" ++
                       "export " ++ n.pp ++ " with parse as hostParse; \n" ++
                       "syntax " ++ n.pp ++ " ; \n" ;

-- gn.bin_main_sv = "import " ++ n.pp ++ " ; \n" ++
--                  "syntax " ++ n.pp ++ " ; \n" ;
}

-- Name --
concrete production simple_name
n::Name ::= id::Id_t
{
 n.pp = id.lexeme ;
}

concrete production qualified_name
n::Name ::= g::Name ':' id::Id_t
{
 n.pp = g.pp ++ ":" ++ id.lexeme ;
}


