grammar silver:utils:ale:host ;

import core;
import silver:utils ;

function build_language
IO ::= r::Root  args::String i::IO
{
 local attribute pp_io :: IO ;
 pp_io = print( --"Compiling " ++ args ++ ":" ++
               "\n-----------------------------------\n" ++
               "building language defined by:" ++
               "\n-----------------------------------\n" ++ 
               r.pp ++ "\n\n" , i ) ;

 local attribute grammarPath :: IOString;
 grammarPath = envVar ("GRAMMAR_PATH", pp_io);


 --local attribute pg_io :: IO ;
 --pg_io = print("\n\nGRAMMAR_PATH: " ++ grammarPath.sValue ++ "\n\n", grammarPath.io);
 --local attribute pglist_io :: IO ;
 --pglist_io = print_string_list(separate_paths(grammarPath.sValue),"\n",pg_io) ;

 ------------------------------------------------------------

 local attribute new_lang_dir :: String ; 
 new_lang_dir = mk_absolute_path(r.new_lang, grammarPath.sValue, grammarPath.io).sValue ;

 local attribute start_io :: IO ;
 start_io = grammarPath.io ;

 --  make 'Generated_Main.sv'
 local attribute rm_main_io :: IO ;
 rm_main_io = system ( "rm -f " ++ new_lang_dir ++ "/Main.sv",  start_io ).io ;

 local attribute main_io :: IO ;
 main_io = writeFile ( new_lang_dir ++ "/Main.sv",  r.new_lang_main_sv, rm_main_io ) ;

 --  make the 'bin' directory
 local attribute bin_io :: IO ;
 bin_io = mk_bin_dir( new_lang_dir, main_io ) ;

 --  make 'bin/Main.sv'
 local attribute rm_bin_main_io :: IO ;
 rm_bin_main_io = system ( "rm -f " ++ new_lang_dir ++ "/bin/Main.sv",  bin_io ).io ;

 local attribute bin_main_io :: IO ;
 bin_main_io = writeFile ( new_lang_dir ++ "/bin/Main.sv",  r.bin_main_sv, rm_bin_main_io ) ;

 --  make 'bin/silver-compile'
 local attribute rm_bin_silver_compile_io :: IO ;
 rm_bin_silver_compile_io = system ( "rm -f " ++ new_lang_dir ++ "/bin/silver_compile",  bin_main_io ).io ;

 local attribute bin_silver_compile_io :: IO ;
 bin_silver_compile_io = writeFile ( new_lang_dir ++ "/bin/silver-compile",  r.bin_silver_compile, rm_bin_silver_compile_io ) ;

 local attribute bin_silver_compile_mod_io :: IO ;
 bin_silver_compile_mod_io = system ( "chmod +x " ++ new_lang_dir ++ "/bin/silver-compile", bin_silver_compile_io ).io ;

 
 --  make 'bin/build.test'
 local attribute rm_bin_build_test_io :: IO ;
 rm_bin_build_test_io = system ( "rm -f " ++ new_lang_dir ++ "/bin/build.test",  bin_silver_compile_mod_io ).io ;

 local attribute bin_build_test_io :: IO ;
 bin_build_test_io = writeFile ( new_lang_dir ++ "/bin/build.test",  r.bin_build_test, rm_bin_build_test_io ) ;



 return bin_build_test_io;
}




function mk_absolute_path
IOString ::= lang_grammar::String grammarPath::String i::IO
{
 return mk_absolute_path_helper( makeDirectory(lang_grammar) , separate_paths(grammarPath), i ) ;
}

function mk_absolute_path_helper
IOString ::= lang_dir::String grammarPaths::[String] i::IO
{
 local attribute test_dir :: IOBoolean ;
 test_dir = isDirectory( path , i ) ; --  print("Testing path: " ++ path, i) ) ;

 local attribute path :: String ;
 path = head(grammarPaths) ++ "/" ++ lang_dir ;

 return if null(grammarPaths)
        then ioString(test_dir.io,"")
        else if test_dir.bValue
             then ioString(test_dir.io,path)
             else mk_absolute_path_helper ( lang_dir, tail(grammarPaths), test_dir.io ) ;
}


function mk_bin_dir
IO ::= dir::String  i::IO
{
 return if isDirectory(dir++"/bin", i).bValue then i else system("mkdir " ++ dir ++ "/bin", i).io ;
}


function separate_paths
[String] ::= paths::String
{
 return separate_string_by(paths,":") ;
-- if  indexof(":",paths) == -1  -- paths does not contain a ":"
  --      then [paths]
    --    else [substring(0,indexof(":",paths),paths)] ++
      --       separate_paths(substring(indexof(":",paths)+1,length(paths),paths)) ;
}


function separate_string_by
[String] ::= str::String separator::String
{
 return if  indexof(separator,str) == -1  -- paths does not contain a ":"
        then [str]
        else [substring(0,indexof(separator,str),str)] ++
             separate_string_by(substring(indexof(separator,str)+1,length(str), str), separator) ;
}


function print_string_list
IO ::= strs::[String] separator::String i::IO
{
 return if null(strs)
        then i
        else print_string_list( tail(strs), separator, print(head(strs)++separator,i));
}
